Package: seqcomp
Type: Package
Title: Sequential Comparison of Probabilistic Forecasts
Version: 0.1.0
Authors@R: person(
    "Akbar", "Alasgarli",
    email = "alasgarliakbar@gmail.com",
    role = c("aut", "cre"))
Description: Implements tools for sequential comparison of probabilistic
    forecasts, including binary and categorical scoring rules, anytime-valid
    confidence sequences, e-processes, Winkler-score comparisons, lag handling,
    and predictable-bound e-processes.
License: MIT + file LICENSE
URL: https://github.com/alasgarliakbar/seqcomp
BugReports: https://github.com/alasgarliakbar/seqcomp/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: lamW
Suggests: scoringRules, VGAM, testthat (>= 3.0.0), knitr, rmarkdown
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-06-24 15:35:04 UTC; hp
Author: Akbar Alasgarli [aut, cre]
Maintainer: Akbar Alasgarli <alasgarliakbar@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-30 12:20:02 UTC

# Generated by roxygen2: do not edit by hand

export(brier_score)
export(calibrate_p_to_e)
export(cm_boundary)
export(compare_forecasts)
export(crps_empirical)
export(crps_normal)
export(crps_std)
export(cs_asymptotic)
export(cs_bernstein)
export(cs_hoeffding)
export(eprocess)
export(eprocess_lag)
export(eprocess_predictable)
export(eprocess_rejections)
export(ge_boundary)
export(log_score)
export(predictable_rejections)
export(ps_boundary)
export(qlike_score)
export(rho_from_vopt)
export(score_bounds)
export(score_diff_scales)
export(spherical_score)
export(split_streams)
export(tick_loss)
export(unroll_stream)
export(winkler_compare)
export(winkler_cs)
export(winkler_etest)
export(winkler_score)
importFrom(stats,pgamma)
importFrom(stats,uniroot)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# seqcomp

<!-- badges: start -->

[![R-CMD-check](https://github.com/alasgarliakbar/seqcomp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/alasgarliakbar/seqcomp/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

`seqcomp` implements anytime-valid tools for the sequential comparison
of probabilistic forecasters, following the framework of Choe and Ramdas
(2024). Given two competing forecasters and a sequence of binary or
categorical outcomes, the package constructs confidence sequences and
e-processes for the running mean score difference that are valid
simultaneously at every point in time, without requiring a pre-specified
sample size or adjustment for repeated monitoring.

The package provides:

- Positively oriented proper scoring rules (Brier, logarithmic,
  spherical, CRPS, tick loss, QLIKE, Winkler).
- Anytime-valid confidence sequences: Hoeffding-style (Theorem 1, Choe
  and Ramdas 2024) and empirical Bernstein (Theorem 2).
- An asymptotic confidence sequence for unbounded scoring rules
  (Appendix C, Choe and Ramdas 2024).
- Sub-exponential e-processes for sequential hypothesis testing (Theorem
  3).
- Winkler-score tools for one-sided sequential inference with unbounded
  binary scoring rules.
- Lag-handling utilities for multi-step-ahead forecast evaluation via
  stream splitting.
- Predictable-bounds tools for settings where score differences are not
  globally bounded in advance.

All boundary computations (normal mixture, gamma-exponential mixture,
polynomial stitching) are implemented from scratch, with no dependency
on the `confseq` package.

## Installation

The development version can be installed from GitHub:

``` r
# install.packages("pak")
pak::pak("alasgarliakbar/seqcomp")
```

After CRAN acceptance:

``` r
install.packages("seqcomp")
```

## Basic example

`compare_forecasts()` is the main entry point. It computes pointwise
scores, the running mean score difference, a confidence sequence, and
two one-sided e-processes in a single call.

``` r
library(seqcomp)

set.seed(1)
n <- 300
y <- rbinom(n, size = 1, prob = 0.55)

# Forecaster p has some signal; forecaster q always predicts 0.5
p <- ifelse(y == 1, 0.62, 0.38)
q <- rep(0.50, n)

out <- compare_forecasts(
  p            = p,
  q            = q,
  y            = y,
  scoring_rule = "brier"
)

tail(out[, c("t", "estimate", "lower", "upper", "e_pq", "e_qp")])
#>       t estimate      lower     upper    e_pq         e_qp
#> 295 295   0.1056 0.07129320 0.1399068 2681618 2.220446e-16
#> 296 296   0.1056 0.07140910 0.1397909 2824574 2.220446e-16
#> 297 297   0.1056 0.07152422 0.1396758 2975161 2.220446e-16
#> 298 298   0.1056 0.07163857 0.1395614 3133784 2.220446e-16
#> 299 299   0.1056 0.07175215 0.1394478 3300873 2.220446e-16
#> 300 300   0.1056 0.07186498 0.1393350 3476882 2.220446e-16
```

The column `estimate` is the running mean score difference
$\hat{\Delta}_t = t^{-1}\sum_{i=1}^t (S(p_i, y_i) - S(q_i, y_i))$.
Positive values favour `p`; negative values favour `q`. The columns
`lower` and `upper` are the empirical Bernstein confidence sequence
bounds. The columns `e_pq` and `e_qp` are the two one-sided e-process
values; the two-sided rejection threshold at level `alpha = 0.05` is
`2 / 0.05 = 40`.

``` r
plot(
  out$t, out$estimate,
  type = "l",
  ylim = range(c(out$lower, out$upper, 0), finite = TRUE),
  xlab = "Time",
  ylab = "Running mean score difference"
)
lines(out$t, out$lower, lty = 2, col = "steelblue")
lines(out$t, out$upper, lty = 2, col = "steelblue")
abline(h = 0, col = "gray50")
legend(
  "topleft",
  legend = c("Estimate", "95% EB confidence sequence"),
  lty    = c(1, 2),
  col    = c("black", "steelblue"),
  bty    = "n"
)
```

<img src="man/figures/README-plot-1.png" alt="" width="100%" />

## Scale parameter conventions

Two scale conventions are used throughout, following Choe and Ramdas
(2024) exactly. Theorem 1 (Hoeffding CS) requires
$|\hat{\delta}_i| \leq c$, so `c = 1` is used for Brier or spherical
score differences in $[-1, 1]$. Theorems 2 and 3 (empirical Bernstein CS
and e-process) require $|\hat{\delta}_i| \leq c/2$, so `c = 2` is used
for the same score differences. `compare_forecasts()` applies these
conventions automatically. They differ from the Python `comparecast`
package, which applies the Theorem 2/3 convention throughout.

## Lower-level interface

`compare_forecasts()` is a convenience wrapper. The underlying functions
can be called directly for finer control:

``` r
scores_p <- brier_score(p, y)
scores_q <- brier_score(q, y)

cs  <- cs_bernstein(scores_p, scores_q, alpha = 0.05, c = 2)
ep  <- eprocess(scores_p, scores_q, alpha = 0.05, c = 2)
```

## Background

The statistical methods in `seqcomp` are based on:

- Confidence sequences and time-uniform probability bounds (Howard,
  Ramdas, McAuliffe and Sekhon, 2021).
- E-values and game-theoretic statistics (Ramdas, Grünwald, Vovk and
  Shafer, 2023).
- Sequential forecaster comparison under the weak null hypothesis (Choe
  and Ramdas, 2024).

The package was developed as part of a bachelor’s thesis at the Vienna
University of Economics and Business (WU Vienna).

## Citation

If this package is used in published work, please cite the package
itself and the following papers:

``` r
citation("seqcomp")
```

## References

Choe, Y. J. and Ramdas, A. (2024). Comparing Sequential Forecasters.
*Operations Research*, 72(4), 1368–1387.
<https://doi.org/10.1287/opre.2021.0792>

Howard, S. R., Ramdas, A., McAuliffe, J. and Sekhon, J. (2021).
Time-uniform, nonparametric, nonasymptotic confidence sequences. *The
Annals of Statistics*, 49(2). <https://doi.org/10.1214/20-AOS1991>

Howard, S. R., Ramdas, A., McAuliffe, J. and Sekhon, J. (2020).
Time-uniform Chernoff bounds via nonnegative supermartingales.
*Probability Surveys*, 17, 257–317. <https://doi.org/10.1214/18-PS321>

Ramdas, A., Grünwald, P., Vovk, V. and Shafer, G. (2023). Game-theoretic
statistics and safe anytime-valid inference. *Statistical Science*,
38(4), 576–601. <https://doi.org/10.1214/23-STS894>

Waudby-Smith, I., Arbour, D., Sinha, R., Kennedy, E. H. and Ramdas, A.
(2024). Time-uniform central limit theory and asymptotic confidence
sequences. *The Annals of Statistics*, 52(6).
<https://doi.org/10.1214/24-AOS2408>

# =============================================================================
# scores.R
# Scoring rules for probabilistic forecast evaluation
#
# All scoring rules are POSITIVELY ORIENTED: higher values = better forecast.
# This matches the CR23 convention:
#   hat_delta_t = S(p_t, y_t) - S(q_t, y_t)
# is positive when forecaster p outperforms forecaster q.
#
# Binary/categorical convention:
#   - Vector probability input is treated as binary and uses y in {0, 1}.
#   - Matrix probability input is treated as categorical and uses y in {1, ..., K}
#     with R-native one-based class labels.
#
# Scoring rules implemented:
#   1. brier_score()      — Brier/quadratic score, binary or categorical,
#                           bounded in [-1, 0]
#   2. log_score()        — Logarithmic score, binary or categorical,
#                           unbounded below
#   3. spherical_score()  — Spherical score, binary or categorical,
#                           bounded in [0, 1]
#   4. tick_loss()        — Quantile/tick loss, negated
#   5. qlike_score()      — QLIKE for variance forecasts, negated
#   6. winkler_score()    — Winkler normalised score for binary outcomes
#   7. CRPS wrappers      — CRPS wrappers over scoringRules, negated
#
# References:
#   GR07  Gneiting & Raftery (2007), JASA 102(477)
#   CR23  Choe & Ramdas (2023), Operations Research 72(4)
#   JKL19 Jordan, Krueger & Lerch (2019), Journal of Statistical Software, 90(12)
# =============================================================================


#' Preprocess categorical probability forecasts
#'
#' Internal helper used by matrix-input branches of `brier_score()`,
#' `spherical_score()`, and `log_score()`.
#'
#' Vector input is converted to a two-column probability matrix using the
#' convention that the vector gives `P(Y = class 2)`. Matrix input is treated
#' as categorical probability forecasts with rows on the probability simplex.
#'
#' @param p Numeric vector or matrix of forecast probabilities.
#' @param y Integer/numeric vector of one-based class labels.
#'
#' @return A list with elements `p_mat`, `y_int`, `T`, and `K`.
#'
#' @keywords internal
#' @noRd
.preprocess_cat <- function(p, y) {

  # ── 1. Shape detection ──────────────────────────────────────────────────
  if (is.vector(p) && !is.matrix(p)) {
    # Binary shorthand: p is a length-T vector of P(Y = class 2).
    # NOTE: when called from brier_score / spherical_score / log_score,
    # the vector path is handled by the binary code branch in those
    # functions and .preprocess_cat() is NOT called.  This branch exists
    # for standalone / pipeline use where matrix form is always desired.
    if (!is.numeric(p))
      stop("p must be numeric.")
    T_ <- length(p)
    K  <- 2L
    p_mat <- cbind(1 - p, p)
  } else if (is.matrix(p)) {
    if (!is.numeric(p))
      stop("p must be a numeric matrix.")
    T_ <- nrow(p)
    K  <- ncol(p)
    p_mat <- p
  } else {
    stop("p must be a numeric vector (binary shorthand) or a numeric matrix (categorical).")
  }

  # ── 2. Dimension sanity ─────────────────────────────────────────────────
  if (as.integer(K) < 2L)
    stop("p must have at least 2 columns (K >= 2).")

  # ── 3. Probability validity ─────────────────────────────────────────────
  if (any(!is.finite(p_mat)))
    stop("p contains non-finite values (NA, NaN, Inf).")
  if (any(p_mat < -1e-9 | p_mat > 1 + 1e-9))
    stop("All entries of p must be in [0, 1].")
  p_mat <- pmin(pmax(p_mat, 0), 1)   # clip negligible floating-point deviations

  row_sums <- rowSums(p_mat)
  if (any(abs(row_sums - 1) > 1e-6)) {
    stop(sprintf(
      "Rows of p must sum to 1; max deviation is %.2e.",
      max(abs(row_sums - 1))
    ))
  }

  # ── 4. Strict integer validation for y ─────────────────────────────────
  if (!is.numeric(y) && !is.integer(y))
    stop("y must be an integer/numeric vector of class labels.")
  if (any(!is.finite(y)))
    stop("y contains non-finite values.")
  if (any(abs(y - round(y)) > 0))
    stop("y must contain integer-valued class labels (e.g. y = 1.9 is not accepted).")
  y <- as.integer(y)

  if (length(y) != T_)
    stop(sprintf(
      "length(y) = %d but nrow(p) = %d; they must agree.",
      length(y), T_
    ))
  if (any(y < 1L | y > as.integer(K)))
    stop(sprintf(
      "y values must be integers in 1:%d (R 1-based labels).",
      K
    ))

  list(p_mat = p_mat, y_int = y, T = T_, K = as.integer(K))
}

#' Brier score for binary and categorical forecasts
#'
#' Computes the positively oriented Brier/quadratic score. Vector probability
#' input is treated as binary; matrix probability input is treated as
#' categorical.
#'
#' @param p Numeric vector in `[0, 1]` for binary forecasts, or a numeric
#'   matrix whose rows are probability vectors for categorical forecasts.
#' @param y For binary vector input, numeric vector in `{0, 1}`. For
#'   categorical matrix input, integer vector in `{1, ..., K}`, where
#'   `K = ncol(p)`.
#'
#' @return Numeric vector of scores in `[-1, 0]`. Higher is better.
#'
#' @details
#' For binary forecasts, this computes
#' \deqn{S(p, y) = -(p-y)^2.}
#'
#' For categorical forecasts, this computes
#' \deqn{S(\mathbf{p}, y) = -\frac{1}{2}\|\mathbf{p} - e_y\|_2^2,}
#' where `e_y` is the one-hot vector of the realised category.
#'
#' With the convention that category 2 corresponds to the binary event
#' `y = 1`, the categorical formula recovers the binary formula exactly
#' when `K = 2`.
#'
#' @section Bounds:
#' Score differences lie in `[-1, 1]`, so use `c = 1` for Theorem 1 and
#' `c = 2` for Theorems 2 and 3.
#'
#' @examples
#' p <- c(0.2, 0.7, 0.9)
#' y <- c(0, 1, 1)
#' brier_score(p, y)
#'
#' @export
brier_score <- function(p, y) {
  if (is.matrix(p)) {
    prep <- .preprocess_cat(p, y)
    idx  <- cbind(seq_len(prep$T), prep$y_int)
    p_y  <- prep$p_mat[idx]

    return(p_y - 0.5 * rowSums(prep$p_mat^2) - 0.5)
  }

  stopifnot(
    length(p) == length(y),
    all(p >= 0 & p <= 1),
    all(y %in% c(0, 1))
  )

  -(p - y)^2
}

#' Logarithmic score for binary and categorical forecasts
#'
#' Computes the positively oriented logarithmic score. Vector probability input
#' is treated as binary; matrix probability input is treated as categorical.
#'
#' @param p Numeric vector in `[0, 1]` for binary forecasts, or a numeric
#'   matrix whose rows are probability vectors for categorical forecasts.
#' @param y For binary vector input, numeric vector in `{0, 1}`. For
#'   categorical matrix input, integer vector in `{1, ..., K}`, where
#'   `K = ncol(p)`.
#' @param eps Numeric. Probability floor used before taking logarithms.
#'   Default is `1e-15`. Set to `0` to disable clipping.
#'
#' @return Numeric vector of scores in `(-Inf, 0]`. Higher is better.
#'
#' @details
#' For binary forecasts, this computes
#' \deqn{S(p, y) = y\log(p) + (1-y)\log(1-p).}
#'
#' For categorical forecasts, this computes
#' \deqn{S(\mathbf{p}, y) = \log(p_y),}
#' where `p_y` is the forecast probability assigned to the realised category.
#'
#' @section Use with seqcomp:
#' The logarithmic score is unbounded below. It should not be used directly
#' with the finite-sample bounded-difference confidence sequences or
#' e-processes. For binary outcomes, use `winkler_score()` and `winkler_cs()`
#' when the Winkler construction is appropriate. For unbounded score
#' differences, use `cs_asymptotic()` or supply genuine predictable bounds
#' to `eprocess_predictable()`.
#'
#' @examples
#' p <- c(0.2, 0.7, 0.9)
#' y <- c(0, 1, 1)
#' log_score(p, y)
#'
#' @export
log_score <- function(p, y, eps = 1e-15) {
  if (is.matrix(p)) {
    prep <- .preprocess_cat(p, y)
    idx  <- cbind(seq_len(prep$T), prep$y_int)
    p_y  <- prep$p_mat[idx]

    return(log(pmax(p_y, eps)))
  }

  stopifnot(
    length(p) == length(y),
    all(is.finite(p)),
    all(y %in% c(0, 1)),
    eps >= 0
  )

  p <- pmax(pmin(p, 1 - eps), eps)
  y * log(p) + (1 - y) * log(1 - p)
}

#' Negated tick loss for quantile forecasts
#'
#' Computes the positively oriented (negated) tick/quantile loss
#' (Koenker & Bassett, 1978).
#'
#' @param q     Numeric vector. Quantile forecasts at level alpha.
#' @param y     Numeric vector. Realised outcomes.
#' @param alpha Numeric in (0,1). Quantile level.
#'
#' @return Numeric vector of negated tick loss scores. Higher = better.
#'
#' @details
#' The standard tick loss is
#' \deqn{\rho_\alpha(u) = u \left(\alpha - \mathbb{1}(u < 0)\right),}
#' where \eqn{u = y - q_\alpha} is the forecast error. This is loss-oriented
#' (lower = better), so the function negates it:
#' \deqn{S_T(q, y; \alpha) = -(y - q)\left(\alpha - \mathbb{1}(y < q)\right).}
#'
#' Tick loss is unbounded on general real-valued outcomes. Bounds derived from
#' an empirical data range are ex-post and do not provide theorem-valid
#' constants for finite-sample Hoeffding/Bernstein confidence sequences or
#' e-processes.
#'
#' Sign convention: the negation means `hat_delta_t > 0` when forecaster `p`
#' has smaller tick loss, hence a better quantile forecast, than forecaster `q`.
#'
#' @examples
#' q <- c(1.0, 1.5, 2.0)
#' y <- c(1.2, 1.4, 2.3)
#' tick_loss(q, y, alpha = 0.5)
#'
#' @export
tick_loss <- function(q, y, alpha) {
  stopifnot(alpha > 0, alpha < 1, length(q) == length(y))
  u <- y - q
  -(u * (alpha - as.numeric(u < 0)))
}

#' Negated QLIKE score for variance forecasts
#'
#' Computes the positively oriented (negated) QLIKE quasi-likelihood loss for
#' variance forecasts.
#'
#' @param sigma2_hat Numeric vector. Forecast variance (strictly positive).
#' @param sigma2     Numeric vector. Realised variance (strictly positive).
#'
#' @return Numeric vector of negated QLIKE scores. Higher is better.
#'   Maximum value is 0, achieved at a perfect forecast `sigma2_hat = sigma2`.
#'   Unbounded below.
#'
#' @details
#' Standard QLIKE loss is
#' \deqn{L_{QL}(\hat\sigma^2, \sigma^2) = \frac{\sigma^2}{\hat\sigma^2} -
#'   \log\frac{\sigma^2}{\hat\sigma^2} - 1.}
#' This is loss-oriented (lower = better, minimum 0 at a perfect forecast), so
#' the function negates it: \eqn{S_{QL} = -L_{QL}}.
#'
#' Literature note: some sources define QLIKE as
#' `log(sigma2_hat) + sigma2 / sigma2_hat`, which differs by constants from
#' the form above. Here the loss is normalised to have minimum 0 and is then
#' negated for positive orientation.
#'
#' @section Unbounded below:
#' QLIKE is unbounded below. It should not be used directly with the
#' finite-sample bounded-difference confidence sequences or e-processes.
#' Use `cs_asymptotic()` for QLIKE-based confidence sequences, or use
#' `eprocess_predictable()` only when genuine ex ante predictable bounds are
#' available. QLIKE is not compatible with the Winkler construction because
#' Winkler scores are restricted to binary outcomes and probability forecasts.
#'
#' @examples
#' sigma2_hat <- c(1.0, 1.5, 2.0)
#' sigma2 <- c(1.1, 1.4, 2.2)
#' qlike_score(sigma2_hat, sigma2)
#'
#' @export
qlike_score <- function(sigma2_hat, sigma2) {
  stopifnot(all(sigma2_hat > 0), all(sigma2 > 0))
  ratio <- sigma2 / sigma2_hat
  -(ratio - log(ratio) - 1)
}

#' Winkler-normalized binary score
#'
#' Normalises the score difference S(p,y) - S(q,y) by the maximum possible
#' score difference given the forecaster ordering, mapping the result to
#' `(-Inf, 1]` (Proposition 4, Choe & Ramdas 2023). Used to apply Theorems 2 & 3
#' to unbounded scoring rules on binary outcomes.
#'
#' @param p         Numeric vector in (0,1). Forecasts from model 1.
#' @param q         Numeric vector in (0,1). Forecasts from model 2.
#' @param y         Numeric vector containing only 0 and 1. Binary outcomes.
#' @param base_score Function. The underlying scoring rule S(p, y).
#'                   Must accept two arguments: forecast probability and outcome.
#'                   Default: log_score (with eps clipping).
#' @param eps       Numeric. Zero-protection for the normaliser denominator.
#'                  Default: 1e-8 (matches Python comparecast convention).
#'
#' @return Numeric vector. Winkler scores in `(-Inf, 1]`.
#'         Upper bound of 1 is tight: w = 1 when y = 1(p > q).
#'
#' @details
#' \deqn{w(p, q, y) = \frac{S(p,y) - S(q,y)}{S(p, \mathbb{1}(p>q)) - S(q, \mathbb{1}(p>q))}}
#' with the convention 0/0 := 0.
#'
#' The lower bound is problem-dependent (depends on how extreme p and q can
#' be). For a two-sided CS via Corollary 2, the user must establish a finite
#' lower bound analytically. If no finite lower bound can be guaranteed, use
#' the one-sided (upper) CS only, as in the CR23 MLB experiments.
#'
#' @section When to use:
#' Strictly limited to binary outcomes `y` in `{0, 1}` and probability
#' forecasts `p`, `q` in `(0, 1)`. Not applicable to QLIKE or other
#' continuous-outcome scoring rules. See CR23 Section G for discussion.
#'
#' For use in Theorems 2 & 3: upper bound = 1 implies c/2 = 1, so use `c = 2`
#' in all GE boundary and e-process calls.
#'
#' @examples
#' p <- c(0.7, 0.6, 0.8, 0.65)
#' q <- c(0.5, 0.7, 0.6, 0.55)
#' y <- c(1, 1, 0, 1)
#' winkler_score(p, q, y)
#'
#' @export
winkler_score <- function(p, q, y,
                          base_score = log_score,
                          eps = 1e-8) {
  stopifnot(
    all(p > 0 & p < 1),
    all(q > 0 & q < 1),
    all(y %in% c(0, 1)),
    length(p) == length(q),
    length(p) == length(y)
  )

  # Numerator: actual score difference
  num <- base_score(p, y) - base_score(q, y)

  # Denominator: score difference at the "favourable" outcome
  # If p > q: favourable outcome is y=1 (both forecasters assigned more
  #           probability to 1, but p assigned more, so p wins when y=1)
  # If p < q: favourable outcome is y=0
  y_fav <- as.numeric(p > q)          # 1 where p > q, 0 where p <= q

  denom <- base_score(p, y_fav) - base_score(q, y_fav)

  # Zero-protection: replace |denom| < eps with eps (preserving sign)
  # When p == q exactly, both scores are identical and the ratio is 0/0 := 0.
  # The eps guard handles near-ties numerically.
  denom_safe <- ifelse(abs(denom) < eps, eps, denom)

  result <- num / denom_safe

  # Enforce 0/0 := 0 convention from CR23
  result[abs(num) < eps & abs(denom) < eps] <- 0

  return(result)
}


# -----------------------------------------------------------------------------
# CRPS wrappers
#
# scoringRules convention: lower score = better forecast (loss-oriented).
# seqcomp convention:      higher score = better forecast (reward-oriented).
#
# Each wrapper checks inputs and negates the scoringRules output.
# -----------------------------------------------------------------------------


#' Negated CRPS for normal predictive distributions
#'
#' Computes the Continuous Ranked Probability Score for a normal predictive
#' distribution using `scoringRules::crps_norm()` and negates it so that higher
#' values are better.
#'
#' @param mu    Numeric vector. Location parameters (conditional means).
#' @param sigma Numeric vector. Scale parameters (conditional SDs, > 0).
#' @param x     Numeric vector. Realised observations.
#'
#' @return Numeric vector of CRPS values in `(-Inf, 0]` (negated loss).
#'
#' @details
#'   Calls \code{scoringRules::crps_norm(y = x, mean = mu, sd = sigma)} and
#'   negates. Use for GARCH(1,1)-norm forecasts where mu is the conditional
#'   mean and sigma is the conditional standard deviation.
#'
#' @examples
#' if (requireNamespace("scoringRules", quietly = TRUE)) {
#'   crps_normal(mu = c(0, 1), sigma = c(1, 2), x = c(0.2, 1.3))
#' }
#'
#' @export
crps_normal <- function(mu, sigma, x) {
  if (!requireNamespace("scoringRules", quietly = TRUE))
    stop("Package 'scoringRules' required. Install with: ",
         "install.packages('scoringRules')")
  stopifnot(
    length(mu)    == length(x),
    length(sigma) == length(x),
    all(sigma > 0)
  )
  -scoringRules::crps_norm(y = x, mean = mu, sd = sigma)
}


#' Negated CRPS for empirical predictive distributions
#'
#' Wrapper over \code{scoringRules::crps_sample} using method = "edf"
#' (empirical distribution function, O(n log n) via quantile decomposition
#' of Laio & Tamea, 2007). Positively oriented: higher = better.
#'
#' @param ensemble Matrix. T x n matrix of forecast draws. Each row
#'                 corresponds to one observation in y and comprises
#'                 n simulation draws from the predictive distribution.
#'                 For Historical Simulation: each row is the past
#'                 WINDOW returns.
#' @param y        Numeric vector of length T. Realised observations.
#'
#' @return Numeric vector of length T of CRPS values in `(-Inf, 0]`
#'         (negated loss).
#'
#' @details
#'   Requires \code{nrow(ensemble) == length(y)}. Passes
#'   \code{dat = ensemble} directly to \code{crps_sample} which handles
#'   vectorisation over rows natively. \code{show_messages} is suppressed
#'   as the "edf" method requires no bandwidth selection messages.
#'
#' @examples
#' if (requireNamespace("scoringRules", quietly = TRUE)) {
#'   ensemble <- matrix(c(0.1, 0.2, 0.3, 1.0, 1.1, 1.2), nrow = 2, byrow = TRUE)
#'   crps_empirical(ensemble, y = c(0.25, 1.05))
#' }
#'
#' @export
crps_empirical <- function(ensemble, y) {
  if (!requireNamespace("scoringRules", quietly = TRUE))
    stop("Package 'scoringRules' required. Install with: ",
         "install.packages('scoringRules')")
  stopifnot(
    is.matrix(ensemble),
    nrow(ensemble) == length(y)
  )
  -scoringRules::crps_sample(y   = y,
                             dat = ensemble,
                             method        = "edf",
                             show_messages = FALSE)
}


#' Negated CRPS for Student-t predictive distributions
#'
#' Wrapper over \code{scoringRules::crps_t}. Positively oriented:
#' higher = better. The dof > 2 constraint ensures finite variance,
#' which is required for the CRPS to be well-defined for the t-distribution.
#'
#' @param mu    Numeric vector. Location parameters (conditional means).
#' @param sigma Numeric vector. Scale parameters (conditional SDs, > 0).
#' @param dof   Numeric vector or scalar. Degrees of freedom (> 2).
#'              May be scalar if constant across all observations (e.g.
#'              estimated once per rolling window).
#' @param x     Numeric vector. Realised observations.
#'
#' @return Numeric vector of CRPS values in `(-Inf, 0]` (negated loss).
#'
#' @details
#'   Calls \code{scoringRules::crps_t(y = x, df = dof, location = mu,
#'   scale = sigma)} and negates. Use for GARCH(1,1)-std forecasts where
#'   dof is the estimated degrees-of-freedom parameter from ugarchroll.
#'
#' @examples
#' if (requireNamespace("scoringRules", quietly = TRUE)) {
#'   crps_std(mu = c(0, 1), sigma = c(1, 2), dof = 5, x = c(0.2, 1.3))
#' }
#'
#' @export
crps_std <- function(mu, sigma, dof, x) {
  if (!requireNamespace("scoringRules", quietly = TRUE))
    stop("Package 'scoringRules' required. Install with: ",
         "install.packages('scoringRules')")
  dof <- rep_len(dof, length(x))
  stopifnot(
    length(mu)    == length(x),
    length(sigma) == length(x),
    all(sigma > 0),
    all(dof   > 2)
  )
  -scoringRules::crps_t(y        = x,
                        df       = dof,
                        location = mu,
                        scale    = sigma)
}

#' Spherical score for binary and categorical forecasts
#'
#' Computes the positively oriented spherical score. Vector probability input
#' is treated as binary; matrix probability input is treated as categorical.
#'
#' @param p Numeric vector in `[0, 1]` for binary forecasts, or a numeric
#'   matrix whose rows are probability vectors for categorical forecasts.
#' @param y For binary vector input, numeric vector in `{0, 1}`. For
#'   categorical matrix input, integer vector in `{1, ..., K}`, where
#'   `K = ncol(p)`.
#'
#' @return Numeric vector of scores in `[0, 1]`. Higher is better.
#'
#' @details For binary forecasts, this computes \deqn{S(p, y) =
#' \frac{py + (1-p)(1-y)}{\sqrt{p^2 + (1-p)^2}}.}
#' For categorical forecasts, this computes \deqn{S(\mathbf{p}, y) =
#' \frac{p_y}{\|\mathbf{p}\|_2},}
#' where `p_y` is the forecast probability assigned to the realised category.
#'
#' Score differences lie in `[-1, 1]`, so use `c = 1` for Theorem 1 and
#' `c = 2` for Theorems 2 and 3.
#'
#' @examples
#' p <- c(0.2, 0.7, 0.9)
#' y <- c(0, 1, 1)
#' spherical_score(p, y)
#'
#' @export
spherical_score <- function(p, y) {
  if (is.matrix(p)) {
    prep <- .preprocess_cat(p, y)
    idx  <- cbind(seq_len(prep$T), prep$y_int)
    p_y  <- prep$p_mat[idx]

    return(p_y / sqrt(rowSums(prep$p_mat^2)))
  }

  stopifnot(
    length(p) == length(y),
    all(p >= 0 & p <= 1),
    all(y %in% c(0, 1))
  )

  (p * y + (1 - p) * (1 - y)) / sqrt(p^2 + (1 - p)^2)
}

#' Score difference bounds for a named scoring rule
#'
#' Returns lo, hi and the derived scale parameters c_thm1, c_thm23 for the
#' score difference process
#'   hat_delta_t = S(p, y) - S(q, y),
#' in those cases where a genuine, theorem-valid bound is available.
#'
#' Convention (utils.R::score_diff_scales):
#'   c_thm1  = (hi - lo) / 2   # Theorem 1:      |delta_i| <= c
#'   c_thm23 = hi - lo        # Theorems 2 & 3: |delta_i| <= c/2
#'
#' @param scoring_rule Character. One of:
#' * `"brier"`, `"spherical"` — bounded, exact finite-sample `c`.
#' * `"winkler"` — descriptive helper for the one-sided CS on the log score.
#' * `"tick"` — unbounded; returns `NULL` with guidance.
#' * `"crps"`, `"crps_normal"`, `"crps_empirical"`, `"crps_std"` — unbounded;
#'   returns `NULL` with guidance.
#' * `"log"`, `"qlike"` — unbounded; returns `NULL` with guidance.
#'
#' @return Named list with elements lo, hi, c_thm1, c_thm23 for bounded
#'         rules, or NULL for unbounded rules (with an informative message).
#'
#' @section Per-rule notes:
#' * **Brier / Spherical** — individual scores lie in `[-1, 0]` (Brier) or
#'   `[0, 1]` (Spherical), so score differences lie in `[-1, 1]` either way.
#'   This bound is exact and yields finite-sample anytime-valid CS via
#'   Hoeffding/Bernstein.
#' * **Winkler** — bounded above by 1; the lower bound is problem-dependent,
#'   so `lo = -Inf` and only `hi = 1` is used, as a descriptive helper for
#'   the one-sided CS wrapper `winkler_cs()`. Not intended for generic
#'   Hoeffding/Bernstein use (Theorem 1 requires a finite symmetric interval).
#' * **Tick loss** — unbounded on general financial returns. Any bound
#'   derived from an empirical data range is ex-post and not
#'   filtration-respecting, so it cannot justify finite-sample anytime
#'   validity. Use `cs_asymptotic()` for tick comparisons.
#' * **CRPS** (normal, t, empirical) — unbounded, since both the predictive
#'   distributions and the realised outcomes are unbounded. A historical
#'   data range is again an ex-post surrogate and does not provide a
#'   theorem-valid `c` for Hoeffding/Bernstein. Use `cs_asymptotic()`, or
#'   supply genuine ex ante bounds in problem-specific code if available.
#' * **Log / QLIKE** — both unbounded. For binary log-score comparisons, use
#'   `winkler_score()` + `winkler_cs()` when the Winkler construction is
#'   appropriate. For categorical log-score, QLIKE, and other unbounded
#'   score differences, use `cs_asymptotic()`, or `eprocess_predictable()`
#'   only with genuine ex ante predictable bounds.
#'
#' @examples
#' score_bounds("brier")
#' score_bounds("winkler")
#'
#' @export
score_bounds <- function(scoring_rule) {

  # Helper: add lo, hi to the scales from score_diff_scales()
  full_bounds <- function(lo, hi) {
    c(list(lo = lo, hi = hi), score_diff_scales(lo, hi))
  }

  switch(
    scoring_rule,

    "spherical" = ,
    "brier" = {
      # Individual Brier score in [-1, 0], Spherical in [0, 1].
      # Both have score differences bounded in [-1, 1].
      full_bounds(lo = -1, hi = 1)
    },

    "winkler" = {
      # Descriptive only: used by winkler_cs() for one-sided log-score CS.
      # lo = -Inf, hi = 1; c_thm1 is not meaningful here.
      list(lo = -Inf, hi = 1, c_thm1 = NA, c_thm23 = 2)
    },

    "tick" = {
      message(
        "Tick loss is unbounded on general outcomes. ",
        "There is no theorem-valid fixed c for Hoeffding/Bernstein. ",
        "Use cs_asymptotic() for tick-based confidence sequences."
      )
      invisible(NULL)
    },

    "crps"           = ,
    "crps_normal"    = ,
    "crps_empirical" = ,
    "crps_std"       = {
      message(
        "CRPS differences are unbounded when both forecasts and outcomes ",
        "have unbounded support. ",
        "Empirical data ranges cannot be used to justify theorem-valid ",
        "Hoeffding/Bernstein bounds. ",
        "Use cs_asymptotic() for CRPS-based confidence sequences."
      )
      invisible(NULL)
    },

    "log" = {
      message(
        "Logarithmic score is unbounded. ",
        "For binary outcomes use winkler_score() + winkler_cs() when appropriate. ",
        "For categorical log-score or other unbounded score differences, use ",
        "cs_asymptotic(), or eprocess_predictable() only with genuine predictable bounds."
      )
      invisible(NULL)
    },

    "qlike" = {
      message(
        "QLIKE is unbounded. ",
        "Use cs_asymptotic() for QLIKE-based confidence sequences."
      )
      invisible(NULL)
    },

    {
      stop(
        "Unknown scoring_rule: '", scoring_rule, "'. ",
        "Supported: 'brier', 'spherical', 'winkler', 'tick', 'crps', 'crps_normal', ",
        "'crps_empirical', 'crps_std', 'log', 'qlike'."
      )
    }
  )
}

# =============================================================================
# tests/testthat/test_scores.R
# =============================================================================

# --- 1. Brier score ---

test_that("brier_score: perfect forecasts score 0", {
  expect_equal(brier_score(1, 1), 0)
  expect_equal(brier_score(0, 0), 0)
})

test_that("brier_score: worst forecasts score -1", {
  expect_equal(brier_score(0, 1), -1)
  expect_equal(brier_score(1, 0), -1)
})

test_that("brier_score: p=0.5 gives -0.25", {
  expect_equal(brier_score(0.5, 1), -0.25)
})

test_that("brier_score is positively oriented", {
  expect_gt(brier_score(0.8, 1), brier_score(0.2, 1))
})

test_that("brier_score values are bounded in [-1, 0]", {
  set.seed(101)
  p <- seq(0, 1, by = 0.01)
  y <- rbinom(length(p), 1, 0.5)
  scores <- brier_score(p, y)
  expect_true(all(scores >= -1 & scores <= 0))
})

# --- 2. Log score ---

test_that("log_score: y=1 at p=1 is near 0", {
  expect_gt(log_score(1, 1), -1e-10)
})

test_that("log_score matches log(p) at p=0.5 for both outcomes", {
  expect_equal(log_score(0.5, 1), log(0.5), tolerance = 1e-10)
  expect_equal(log_score(0.5, 0), log(0.5), tolerance = 1e-10)
})

test_that("log_score is positively oriented", {
  expect_gt(log_score(0.8, 1), log_score(0.3, 1))
})

test_that("log_score eps clipping prevents -Inf", {
  expect_true(is.finite(log_score(0, 1, eps = 1e-15)))
})

# --- 3. Tick loss ---

test_that("tick_loss is finite for exceedance and non-exceedance", {
  q_test <- -0.02
  alpha_test <- 0.05
  expect_true(is.finite(tick_loss(q_test, -0.05, alpha_test)))
  expect_true(is.finite(tick_loss(q_test,  0.01, alpha_test)))
})

test_that("tick_loss is 0 when forecast equals outcome exactly", {
  expect_equal(tick_loss(0.0, 0.0, 0.05), 0, tolerance = 1e-12)
})

test_that("tick_loss propriety: correct quantile scores better than a biased one", {
  set.seed(1)
  n_prop <- 5000
  y_prop <- rnorm(n_prop)
  q_correct <- qnorm(0.05)
  q_wrong   <- qnorm(0.05) + 0.5

  mean_correct <- mean(tick_loss(rep(q_correct, n_prop), y_prop, 0.05))
  mean_wrong   <- mean(tick_loss(rep(q_wrong,   n_prop), y_prop, 0.05))

  expect_gt(mean_correct, mean_wrong)
})

# --- 4. QLIKE score ---

test_that("qlike_score: perfect forecast scores 0", {
  expect_equal(qlike_score(1.0, 1.0), 0, tolerance = 1e-12)
})

test_that("qlike_score is bounded above by 0", {
  expect_lte(qlike_score(2.0, 1.0), 0)
  expect_lte(qlike_score(0.5, 1.0), 0)
})

test_that("qlike_score is positively oriented", {
  sigma2_true <- 0.04
  expect_gt(qlike_score(sigma2_true, sigma2_true),
            qlike_score(sigma2_true * 2, sigma2_true))
})

test_that("qlike_score propriety: correct variance beats biased forecast on average", {
  set.seed(2)
  n_q <- 5000
  sigma2_true_val <- 0.04
  eps_q <- rnorm(n_q, sd = sqrt(sigma2_true_val))
  rv    <- eps_q^2

  mean_q_correct <- mean(qlike_score(rep(sigma2_true_val, n_q), rv))
  mean_q_wrong   <- mean(qlike_score(rep(sigma2_true_val * 1.5, n_q), rv))

  expect_gt(mean_q_correct, mean_q_wrong)
})

# --- 5. Winkler score ---

test_that("winkler_score equals 1 at the favourable outcome", {
  p_w <- c(0.8, 0.3, 0.6)
  q_w <- c(0.5, 0.6, 0.4)
  y_fav <- as.numeric(p_w > q_w)
  w_fav <- winkler_score(p_w, q_w, y_fav)
  expect_equal(w_fav, rep(1, 3), tolerance = 1e-8)
})

test_that("winkler_score is <= 0 at the unfavourable outcome", {
  p_w <- c(0.8, 0.3, 0.6)
  q_w <- c(0.5, 0.6, 0.4)
  y_fav   <- as.numeric(p_w > q_w)
  y_unfav <- 1 - y_fav
  w_unfav <- winkler_score(p_w, q_w, y_unfav)
  expect_true(all(w_unfav <= 1e-10))
})

test_that("winkler_score is bounded above by 1", {
  set.seed(3)
  p_rand <- runif(1000, 0.05, 0.95)
  q_rand <- runif(1000, 0.05, 0.95)
  y_rand <- rbinom(1000, 1, 0.5)
  w_rand <- winkler_score(p_rand, q_rand, y_rand)
  expect_true(all(w_rand <= 1 + 1e-10))
})

test_that("winkler_score follows the 0/0 convention when p == q", {
  p_eq <- rep(0.5, 5); q_eq <- rep(0.5, 5); y_eq <- c(0, 1, 0, 1, 0)
  w_eq <- winkler_score(p_eq, q_eq, y_eq)
  expect_equal(w_eq, rep(0, 5), tolerance = 1e-6)
})

# --- 6. crps_normal ---

test_that("crps_normal: perfect deterministic forecast is near 0", {
  skip_if_not_installed("scoringRules")
  expect_equal(crps_normal(mu = 0, sigma = 1e-6, x = 0), 0, tolerance = 1e-4)
})

test_that("crps_normal: all values are <= 0", {
  skip_if_not_installed("scoringRules")
  set.seed(5)
  mu_t  <- rnorm(100)
  sig_t <- abs(rnorm(100)) + 0.1
  x_t   <- rnorm(100)
  expect_true(all(crps_normal(mu_t, sig_t, x_t) <= 1e-10))
})

test_that("crps_normal: correct sigma scores better than overdispersed", {
  skip_if_not_installed("scoringRules")
  set.seed(6)
  y_t       <- rnorm(1000)
  crps_good <- mean(crps_normal(rep(0, 1000), rep(1, 1000), y_t))
  crps_bad  <- mean(crps_normal(rep(0, 1000), rep(3, 1000), y_t))
  expect_gt(crps_good, crps_bad)
})

test_that("crps_normal: unbiased forecast scores better than biased", {
  skip_if_not_installed("scoringRules")
  set.seed(6)
  y_t          <- rnorm(1000)
  crps_centred <- mean(crps_normal(y_t,     rep(1, 1000), y_t))
  crps_biased  <- mean(crps_normal(y_t + 1, rep(1, 1000), y_t))
  expect_gt(crps_centred, crps_biased)
})

test_that("crps_normal: sigma <= 0 raises an error", {
  skip_if_not_installed("scoringRules")
  expect_error(crps_normal(0, -1, 0))
})

test_that("crps_normal: length mismatch raises an error", {
  skip_if_not_installed("scoringRules")
  expect_error(crps_normal(c(0, 0), c(1, 1), c(0, 0, 0)))
})

# --- 7. crps_empirical ---

test_that("crps_empirical agrees with crps_normal for a large N(0,1) ensemble", {
  skip_if_not_installed("scoringRules")
  set.seed(7)
  n_ens   <- 10000
  y_obs   <- rnorm(20)
  ens_mat <- matrix(rnorm(20 * n_ens, mean = 0, sd = 1), nrow = 20, ncol = n_ens)

  crps_norm_vals <- crps_normal(rep(0, 20), rep(1, 20), y_obs)
  crps_emp_vals  <- crps_empirical(ens_mat, y_obs)

  expect_lt(max(abs(crps_emp_vals - crps_norm_vals)), 0.05)
})

test_that("crps_empirical: all values are <= 0", {
  skip_if_not_installed("scoringRules")
  set.seed(8)
  small_ens <- matrix(rnorm(50 * 250), nrow = 50, ncol = 250)
  y_small   <- rnorm(50)
  expect_true(all(crps_empirical(small_ens, y_small) <= 1e-10))
})

test_that("crps_empirical: centred ensemble scores better than shifted ensemble", {
  skip_if_not_installed("scoringRules")
  set.seed(9)
  y_ref     <- rnorm(100)
  ens_good  <- matrix(rnorm(100 * 2000, mean = 0, sd = 1), nrow = 100)
  ens_shift <- matrix(rnorm(100 * 2000, mean = 2, sd = 1), nrow = 100)

  expect_gt(mean(crps_empirical(ens_good,  y_ref)),
            mean(crps_empirical(ens_shift, y_ref)))
})

test_that("crps_empirical: non-matrix ensemble raises an error", {
  skip_if_not_installed("scoringRules")
  set.seed(9)
  y_ref    <- rnorm(100)
  ens_good <- matrix(rnorm(100 * 2000, mean = 0, sd = 1), nrow = 100)
  expect_error(crps_empirical(as.data.frame(ens_good), y_ref))
})

test_that("crps_empirical: nrow/length mismatch raises an error", {
  skip_if_not_installed("scoringRules")
  set.seed(9)
  y_ref    <- rnorm(100)
  ens_good <- matrix(rnorm(100 * 2000, mean = 0, sd = 1), nrow = 100)
  expect_error(crps_empirical(ens_good, y_ref[1:50]))
})

# --- 8. crps_std ---

test_that("crps_std: perfect deterministic forecast is near 0 at large df", {
  skip_if_not_installed("scoringRules")
  expect_equal(crps_std(mu = 0, sigma = 1e-6, dof = 30, x = 0), 0, tolerance = 1e-4)
})

test_that("crps_std: all values are <= 0", {
  skip_if_not_installed("scoringRules")
  set.seed(10)
  mu_s  <- rnorm(100)
  sig_s <- abs(rnorm(100)) + 0.1
  dof_s <- runif(100, min = 3, max = 30)
  x_s   <- rnorm(100)
  expect_true(all(crps_std(mu_s, sig_s, dof_s, x_s) <= 1e-10))
})

test_that("crps_std converges to crps_normal as dof grows large", {
  skip_if_not_installed("scoringRules")
  set.seed(11)
  y_conv   <- rnorm(200)
  crps_n   <- crps_normal(rep(0, 200), rep(1, 200), y_conv)
  crps_t30 <- crps_std(rep(0, 200), rep(1, 200), dof = 1000, y_conv)
  expect_lt(max(abs(crps_t30 - crps_n)), 1e-3)
})

test_that("crps_std: higher df scores better than lower df on normal data", {
  skip_if_not_installed("scoringRules")
  set.seed(12)
  y_norm  <- rnorm(2000)
  crps_hi <- mean(crps_std(rep(0, 2000), rep(1, 2000), dof = 100, y_norm))
  crps_lo <- mean(crps_std(rep(0, 2000), rep(1, 2000), dof = 3,   y_norm))
  expect_gt(crps_hi, crps_lo)
})

test_that("crps_std recycles a scalar dof to the input vector length", {
  skip_if_not_installed("scoringRules")
  set.seed(10)
  mu_s  <- rnorm(100)
  sig_s <- abs(rnorm(100)) + 0.1
  x_s   <- rnorm(100)
  expect_equal(length(crps_std(mu_s, sig_s, dof = 5, x_s)), 100)
})

test_that("crps_std: dof <= 2 raises an error", {
  skip_if_not_installed("scoringRules")
  expect_error(crps_std(0, 1, dof = 2, x = 0))
})

test_that("crps_std: sigma <= 0 raises an error", {
  skip_if_not_installed("scoringRules")
  expect_error(crps_std(0, -1, dof = 5, x = 0))
})

test_that("crps_std: length mismatch raises an error", {
  skip_if_not_installed("scoringRules")
  expect_error(crps_std(c(0, 0), c(1, 1), dof = 5, x = c(0, 0, 0)))
})

# --- 9. Spherical score ---

test_that("spherical_score: perfect forecasts score 1", {
  expect_equal(spherical_score(1, 1), 1)
  expect_equal(spherical_score(0, 0), 1)
})

test_that("spherical_score: worst forecasts score 0", {
  expect_equal(spherical_score(0, 1), 0)
  expect_equal(spherical_score(1, 0), 0)
})

test_that("spherical_score at p=0.5 equals sqrt(0.5)", {
  expect_equal(spherical_score(0.5, 1), sqrt(0.5), tolerance = 1e-10)
})

test_that("spherical_score is positively oriented", {
  expect_gt(spherical_score(0.8, 1), spherical_score(0.2, 1))
})

test_that("spherical_score values are bounded in [0, 1]", {
  set.seed(101)
  p_test <- seq(0, 1, by = 0.01)
  y_test <- rbinom(length(p_test), 1, 0.5)
  scores <- spherical_score(p_test, y_test)
  expect_true(all(scores >= 0 & scores <= 1))
})

# --- 10. Categorical checks ---

test_that("categorical brier_score matches hand-computed values", {
  P3 <- matrix(c(0.7, 0.1, 0.2,
                 0.2, 0.5, 0.3),
               nrow = 2, byrow = TRUE)
  y3 <- c(1L, 2L)
  expect_equal(brier_score(P3, y3), c(-0.07, -0.19), tolerance = 1e-12)
})

test_that("categorical spherical_score matches hand-computed values", {
  P3 <- matrix(c(0.7, 0.1, 0.2,
                 0.2, 0.5, 0.3),
               nrow = 2, byrow = TRUE)
  y3 <- c(1L, 2L)
  expect_equal(spherical_score(P3, y3),
               c(0.7 / sqrt(0.54), 0.5 / sqrt(0.38)), tolerance = 1e-12)
})

test_that("categorical log_score matches hand-computed values", {
  P3 <- matrix(c(0.7, 0.1, 0.2,
                 0.2, 0.5, 0.3),
               nrow = 2, byrow = TRUE)
  y3 <- c(1L, 2L)
  expect_equal(log_score(P3, y3), c(log(0.7), log(0.5)), tolerance = 1e-12)
})

test_that("binary vector path agrees with the K=2 categorical matrix path", {
  p_vec <- c(0.7, 0.2)
  y_01  <- c(1,   0)
  P2    <- cbind(1 - p_vec, p_vec)
  y_12  <- y_01 + 1L

  expect_equal(brier_score(p_vec, y_01),     brier_score(P2, y_12),     tolerance = 1e-12)
  expect_equal(spherical_score(p_vec, y_01), spherical_score(P2, y_12), tolerance = 1e-12)
  expect_equal(log_score(p_vec, y_01),       log_score(P2, y_12),       tolerance = 1e-12)
})

test_that("categorical preprocessing rejects non-integer y", {
  P3 <- matrix(c(0.7, 0.1, 0.2, 0.2, 0.5, 0.3), nrow = 2, byrow = TRUE)
  expect_error(brier_score(P3, c(1, 1.9)))
})

test_that("categorical preprocessing rejects NA y", {
  P3 <- matrix(c(0.7, 0.1, 0.2, 0.2, 0.5, 0.3), nrow = 2, byrow = TRUE)
  expect_error(brier_score(P3, c(1L, NA_integer_)))
})

test_that("categorical preprocessing rejects rows not summing to 1", {
  P3 <- matrix(c(0.7, 0.1, 0.2, 0.2, 0.5, 0.3), nrow = 2, byrow = TRUE)
  y3 <- c(1L, 2L)
  P_bad <- P3
  P_bad[1, 1] <- 0.8
  expect_error(brier_score(P_bad, y3))
})

test_that("categorical preprocessing rejects negative probabilities", {
  P3 <- matrix(c(0.7, 0.1, 0.2, 0.2, 0.5, 0.3), nrow = 2, byrow = TRUE)
  y3 <- c(1L, 2L)
  P_bad <- P3
  P_bad[1, 2] <- -0.01
  expect_error(brier_score(P_bad, y3))
})

test_that("categorical preprocessing rejects zero class labels", {
  P3 <- matrix(c(0.7, 0.1, 0.2, 0.2, 0.5, 0.3), nrow = 2, byrow = TRUE)
  expect_error(brier_score(P3, c(0L, 2L)))
})

test_that("categorical preprocessing rejects K=1 matrices", {
  P_bad <- matrix(1, nrow = 2, ncol = 1)
  expect_error(brier_score(P_bad, c(1L, 1L)))
})

# --- 11. score_bounds ---

test_that("score_bounds: Brier has exact bounds and correct constants", {
  sb_brier <- score_bounds("brier")
  expect_equal(sb_brier$lo, -1)
  expect_equal(sb_brier$hi,  1)
  expect_equal(sb_brier$c_thm1,  1)
  expect_equal(sb_brier$c_thm23, 2)
})

test_that("score_bounds: Brier constants are arithmetically consistent with lo/hi", {
  sb_brier <- score_bounds("brier")
  expect_equal(sb_brier$c_thm1,  (sb_brier$hi - sb_brier$lo) / 2, tolerance = 1e-12)
  expect_equal(sb_brier$c_thm23, (sb_brier$hi - sb_brier$lo),     tolerance = 1e-12)
})

test_that("score_bounds: spherical has the same score-difference interval as Brier", {
  sb_spherical <- score_bounds("spherical")
  expect_equal(sb_spherical$lo, -1)
  expect_equal(sb_spherical$hi,  1)
  expect_equal(sb_spherical$c_thm1,  1)
  expect_equal(sb_spherical$c_thm23, 2)
})

test_that("score_bounds: Winkler is a special case with lo = -Inf and c_thm1 = NA", {
  sb_wink <- score_bounds("winkler")
  expect_true(is.infinite(sb_wink$lo) && sb_wink$lo < 0)
  expect_equal(sb_wink$hi, 1)
  expect_true(is.na(sb_wink$c_thm1))
  expect_equal(sb_wink$c_thm23, 2)
})

unbounded_rules <- c("tick", "crps", "crps_normal", "crps_empirical", "crps_std",
                     "log", "qlike")
for (rule in unbounded_rules) {
  test_that(sprintf("score_bounds('%s') returns NULL with an 'unbounded' message", rule), {
    expect_message(out <- score_bounds(rule), "unbounded")
    expect_null(out)
  })
}

test_that("score_bounds raises an error for an unknown rule", {
  expect_error(score_bounds("rmse"))
})

# =============================================================================
# tests/testthat/test_compare_forecasts.R
# =============================================================================

test_that("compare_forecasts returns a well-formed data.frame for Brier scores", {
  set.seed(101)
  T_ <- 100
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(p, q, y, scoring_rule = "brier")

  expect_true(is.data.frame(out))
  expect_equal(nrow(out), T_)
  expect_equal(
    names(out),
    c("t", "score_p", "score_q", "delta", "estimate", "lower", "upper", "e_pq", "e_qp")
  )
  expect_equal(out$t, seq_len(T_))
  expect_true(all(is.finite(out$score_p)))
  expect_true(all(is.finite(out$score_q)))
  expect_true(all(is.finite(out$delta)))
  expect_true(all(is.finite(out$estimate)))
  expect_true(all(is.finite(out$lower)))
  expect_true(all(is.finite(out$upper)))
  expect_true(all(out$e_pq > 0))
  expect_true(all(out$e_qp > 0))
})

test_that("compare_forecasts computes the same Brier scores as brier_score", {
  set.seed(102)
  T_ <- 80
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(p, q, y, scoring_rule = "brier")

  expect_equal(out$score_p, brier_score(p, y))
  expect_equal(out$score_q, brier_score(q, y))
  expect_equal(out$delta, brier_score(p, y) - brier_score(q, y))
  expect_equal(out$estimate, cumsum(out$delta) / seq_len(T_))
})

test_that("compare_forecasts agrees with direct cs_bernstein and eprocess calls", {
  set.seed(103)
  T_ <- 120
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(
    p, q, y,
    scoring_rule = "brier",
    cs_type = "bernstein",
    compute_cs = TRUE,
    compute_e = TRUE,
    alpha = 0.05
  )

  sp <- brier_score(p, y)
  sq <- brier_score(q, y)

  cs_direct <- cs_bernstein(sp, sq, alpha = 0.05, c = 2)
  ep_direct <- eprocess(sp, sq, alpha = 0.05, c = 2)

  expect_equal(out$lower, cs_direct$lower, tolerance = 1e-10)
  expect_equal(out$upper, cs_direct$upper, tolerance = 1e-10)
  expect_equal(out$e_pq, ep_direct$e_pq, tolerance = 1e-10)
  expect_equal(out$e_qp, ep_direct$e_qp, tolerance = 1e-10)
})

test_that("compare_forecasts supports Hoeffding confidence sequences", {
  set.seed(104)
  T_ <- 100
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(
    p, q, y,
    scoring_rule = "brier",
    cs_type = "hoeffding",
    compute_e = FALSE
  )

  sp <- brier_score(p, y)
  sq <- brier_score(q, y)
  cs_direct <- cs_hoeffding(sp, sq, alpha = 0.05, c = 1)

  expect_equal(out$lower, cs_direct$lower, tolerance = 1e-10)
  expect_equal(out$upper, cs_direct$upper, tolerance = 1e-10)
  expect_true(all(is.na(out$e_pq)))
  expect_true(all(is.na(out$e_qp)))
})

test_that("compare_forecasts supports spherical scores", {
  set.seed(105)
  T_ <- 100
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(p, q, y, scoring_rule = "spherical")

  expect_equal(out$score_p, spherical_score(p, y))
  expect_equal(out$score_q, spherical_score(q, y))
  expect_true(all(is.finite(out$lower)))
  expect_true(all(is.finite(out$upper)))
  expect_true(all(out$e_pq > 0))
  expect_true(all(out$e_qp > 0))
})

test_that("compare_forecasts supports categorical matrix forecasts", {
  set.seed(106)
  T_ <- 60
  K <- 3

  raw_p <- matrix(runif(T_ * K), nrow = T_, ncol = K)
  raw_q <- matrix(runif(T_ * K), nrow = T_, ncol = K)

  p <- raw_p / rowSums(raw_p)
  q <- raw_q / rowSums(raw_q)
  y <- sample(seq_len(K), T_, replace = TRUE)

  out <- compare_forecasts(p, q, y, scoring_rule = "brier")

  expect_equal(out$score_p, brier_score(p, y))
  expect_equal(out$score_q, brier_score(q, y))
  expect_equal(nrow(out), T_)
  expect_true(all(is.finite(out$lower)))
  expect_true(all(is.finite(out$upper)))
})

test_that("compare_forecasts defaults to asymptotic CS for log scores", {
  set.seed(107)
  T_ <- 100
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_, 0.05, 0.95)
  q <- runif(T_, 0.05, 0.95)

  out <- compare_forecasts(
    p, q, y,
    scoring_rule = "log",
    compute_e = FALSE
  )

  sp <- log_score(p, y)
  sq <- log_score(q, y)
  cs_direct <- cs_asymptotic(sp, sq, alpha = 0.05)

  expect_equal(attr(out, "cs_type"), "asymptotic")
  expect_equal(out$score_p, sp)
  expect_equal(out$score_q, sq)
  expect_equal(out$lower, cs_direct$lower, tolerance = 1e-10)
  expect_equal(out$upper, cs_direct$upper, tolerance = 1e-10)
  expect_true(all(is.na(out$e_pq)))
  expect_true(all(is.na(out$e_qp)))
})

test_that("compare_forecasts refuses e-processes for unbounded log scores", {
  set.seed(108)
  T_ <- 50
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_, 0.05, 0.95)
  q <- runif(T_, 0.05, 0.95)

  expect_error(
    compare_forecasts(p, q, y, scoring_rule = "log", compute_e = TRUE),
    regexp = "bounded score differences"
  )
})

test_that("compare_forecasts refuses finite-sample CS for unbounded log scores", {
  set.seed(109)
  T_ <- 50
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_, 0.05, 0.95)
  q <- runif(T_, 0.05, 0.95)

  expect_error(
    compare_forecasts(
      p, q, y,
      scoring_rule = "log",
      cs_type = "bernstein",
      compute_e = FALSE
    ),
    regexp = "bounded score differences"
  )
})

test_that("compare_forecasts can suppress CS and e-process output", {
  set.seed(110)
  T_ <- 70
  y <- rbinom(T_, 1, 0.5)
  p <- runif(T_)
  q <- runif(T_)

  out <- compare_forecasts(
    p, q, y,
    scoring_rule = "brier",
    compute_cs = FALSE,
    compute_e = FALSE
  )

  expect_true(all(is.na(out$lower)))
  expect_true(all(is.na(out$upper)))
  expect_true(all(is.na(out$e_pq)))
  expect_true(all(is.na(out$e_qp)))
  expect_true(all(is.finite(out$estimate)))
})

test_that("compare_forecasts detects shape and length mismatches", {
  y <- c(0, 1, 0)
  p <- c(0.2, 0.7, 0.4)
  q <- c(0.3, 0.6)

  expect_error(
    compare_forecasts(p, q, y, scoring_rule = "brier"),
    regexp = "same number"
  )

  p_mat <- cbind(1 - p, p)

  expect_error(
    compare_forecasts(p_mat, p, y, scoring_rule = "brier"),
    regexp = "same shape"
  )
})

#' seqcomp: Sequential Comparison of Probabilistic Forecasts
#'
#' `seqcomp` provides tools for comparing probabilistic forecasters
#' sequentially, following the anytime-valid framework of Choe and Ramdas
#' (2024).
#'
#' The package is built around the score difference
#'
#' \deqn{\hat{\delta}_t = S(p_t, y_t) - S(q_t, y_t),}
#'
#' where scores are positively oriented, so larger values are better. Positive
#' score differences favour forecaster `p`; negative score differences favour
#' forecaster `q`.
#'
#' @section Main workflow:
#' For most applications, start with [compare_forecasts()]. It computes
#' pointwise scores, running mean score differences, confidence sequences, and
#' e-processes in one call.
#'
#' @section Scoring rules:
#' The package includes positively oriented scoring rules such as
#' [brier_score()], [log_score()], [spherical_score()], [tick_loss()],
#' [qlike_score()], [winkler_score()], [crps_normal()], [crps_empirical()],
#' and [crps_std()].
#'
#' @section Confidence sequences:
#' Use [cs_hoeffding()] for Hoeffding-style confidence sequences,
#' [cs_bernstein()] for empirical Bernstein confidence sequences, and
#' [cs_asymptotic()] for asymptotic confidence sequences when finite-sample
#' boundedness is not available.
#'
#' @section E-processes:
#' Use [eprocess()] for the main sub-exponential mixture e-process and
#' [eprocess_rejections()] to extract first rejection times. For multi-step
#' forecasts, see [eprocess_lag()]. For predictable time-varying bounds, see
#' [eprocess_predictable()].
#'
#' @section Winkler scores:
#' For binary probability forecasts with unbounded base scores, use
#' [winkler_score()], [winkler_cs()], [winkler_etest()], or
#' [winkler_compare()].
#'
#' @references
#' Choe, Y. J. and Ramdas, A. (2024). Comparing Sequential Forecasters.
#' Operations Research, 72(4), 1368-1387.
#'
#' Howard, S. R., Ramdas, A., McAuliffe, J. and Sekhon, J. (2021).
#' Time-uniform, nonparametric, nonasymptotic confidence sequences.
#' The Annals of Statistics, 49(2).
#'
#' @name seqcomp-package
#' @aliases seqcomp seqcomp-package
NULL

# =============================================================================
# compare_forecasts.R
# High-level wrapper for sequential comparison of two probabilistic forecasters
#
# This file provides a user-facing pipeline around the lower-level seqcomp
# building blocks:
#   1. compute pointwise scores,
#   2. compute score differences,
#   3. optionally construct a confidence sequence,
#   4. optionally construct two one-sided e-processes.
#
# The wrapper is intentionally restricted to the Choe & Ramdas framework
# implemented elsewhere in the package.
# =============================================================================

#' Compare Two Sequential Forecasters
#'
#' Computes pointwise scores for two probabilistic forecasters and compares
#' them sequentially using confidence sequences and, when valid finite-sample
#' bounds are available, e-processes.
#'
#' This is a convenience wrapper around [brier_score()], [spherical_score()],
#' [log_score()], [cs_hoeffding()], [cs_bernstein()], [cs_asymptotic()], and
#' [eprocess()]. It is designed for the common workflow where the user has two
#' forecast streams `p` and `q`, an outcome stream `y`, and wants a single tidy
#' output object.
#'
#' @param p Forecasts from forecaster 1. For binary outcomes, a numeric vector
#'   of probabilities for event `y = 1`. For categorical outcomes, a numeric
#'   matrix whose rows are probability vectors.
#' @param q Forecasts from forecaster 2, in the same format as `p`.
#' @param y Outcomes. For binary vector forecasts, a numeric vector in `{0, 1}`.
#'   For categorical matrix forecasts, integer class labels in `{1, ..., K}`.
#' @param scoring_rule Character. Scoring rule used to compare forecasts.
#'   Currently supports `"brier"`, `"spherical"`, and `"log"`.
#' @param alpha Numeric in `(0, 1)`. Significance level. Default is `0.05`.
#' @param cs_type Character or `NULL`. Confidence sequence type:
#'   `"bernstein"`, `"hoeffding"`, `"asymptotic"`, or `"none"`.
#'   If `NULL`, the wrapper uses `"bernstein"` for bounded scoring rules
#'   (`"brier"` and `"spherical"`) and `"asymptotic"` for `"log"`.
#' @param compute_cs Logical. If `TRUE`, compute a confidence sequence.
#'   Default is `TRUE`.
#' @param compute_e Logical. If `TRUE`, compute two one-sided e-processes.
#'   Default is `TRUE`. This is only allowed for bounded score differences
#'   under the current wrapper, namely `"brier"` and `"spherical"`.
#' @param v_opt Numeric > 0. Intrinsic time at which the mixture boundary or
#'   e-process is tuned to be tightest. Default is `10`.
#' @param boundary Character. Boundary type passed to [cs_hoeffding()] or
#'   [cs_bernstein()]. Default is `"mixture"`.
#' @param lcb_only Logical. If `TRUE`, compute a lower one-sided empirical
#'   Bernstein CS. Only used when `cs_type = "bernstein"`.
#' @param ucb_only Logical. If `TRUE`, compute an upper one-sided empirical
#'   Bernstein CS. Only used when `cs_type = "bernstein"`.
#' @param eps Numeric. Probability floor passed to [log_score()] when
#'   `scoring_rule = "log"`. Default is `1e-15`.
#' @param clip_max Numeric. Maximum e-process value before clipping. Passed to
#'   [eprocess()]. Default is `1e7`.
#'
#' @return A `data.frame` with one row per time point and columns:
#' \describe{
#'   \item{`t`}{Time index.}
#'   \item{`score_p`}{Pointwise score of forecaster `p`.}
#'   \item{`score_q`}{Pointwise score of forecaster `q`.}
#'   \item{`delta`}{Pointwise score difference, `score_p - score_q`.}
#'   \item{`estimate`}{Running mean score difference. Positive values favour
#'     forecaster `p`; negative values favour forecaster `q`.}
#'   \item{`lower`, `upper`}{Confidence sequence bounds. These are `NA` if
#'     `compute_cs = FALSE` or `cs_type = "none"`.}
#'   \item{`e_pq`, `e_qp`}{One-sided e-processes. `e_pq` tests whether
#'     forecaster `p` outperforms `q`; `e_qp` tests the reverse direction.
#'     These are `NA` if `compute_e = FALSE`.}
#' }
#'
#' @details
#' All scoring rules in `seqcomp` are positively oriented: higher scores are
#' better. Therefore
#' \deqn{\hat{\delta}_t = S(p_t, y_t) - S(q_t, y_t)}
#' is positive when forecaster `p` performs better than forecaster `q` at time
#' `t`.
#'
#' For `"brier"` and `"spherical"`, score differences are bounded in `[-1, 1]`.
#' The wrapper therefore uses `c = 1` for Hoeffding-style confidence sequences
#' and `c = 2` for empirical Bernstein confidence sequences and e-processes.
#'
#' For `"log"`, score differences are unbounded. The wrapper therefore defaults
#' to [cs_asymptotic()] and refuses to compute finite-sample e-processes. For
#' binary log-score comparisons where the Winkler construction is appropriate,
#' use [winkler_compare()] instead.
#'
#' @section Interpretation:
#' The confidence sequence estimates the running average score advantage of
#' `p` over `q`. If the whole interval lies above zero, the data favour `p`;
#' if the whole interval lies below zero, the data favour `q`.
#'
#' The e-processes are evidence processes for one-sided null hypotheses. At
#' level `alpha`, the two-sided rejection threshold used by [eprocess()] is
#' `2 / alpha`.
#'
#' @examples
#' set.seed(1)
#' y <- rbinom(200, 1, 0.5)
#' p <- rep(0.5, 200)
#' q <- runif(200)
#'
#' out <- compare_forecasts(p, q, y, scoring_rule = "brier")
#' tail(out)
#'
#' @export
compare_forecasts <- function(p, q, y,
                              scoring_rule = c("brier", "spherical", "log"),
                              alpha      = 0.05,
                              cs_type    = NULL,
                              compute_cs = TRUE,
                              compute_e  = TRUE,
                              v_opt      = 10,
                              boundary   = "mixture",
                              lcb_only   = FALSE,
                              ucb_only   = FALSE,
                              eps        = 1e-15,
                              clip_max   = 1e7) {

  scoring_rule <- match.arg(scoring_rule)

  stopifnot(
    alpha > 0, alpha < 1,
    is.logical(compute_cs), length(compute_cs) == 1,
    is.logical(compute_e),  length(compute_e)  == 1,
    v_opt > 0,
    eps >= 0,
    clip_max > 0
  )

  forecast_length <- function(x) {
    if (is.matrix(x)) nrow(x) else length(x)
  }

  T_ <- forecast_length(p)

  if (forecast_length(q) != T_ || length(y) != T_) {
    stop("p, q, and y must have the same number of forecast/outcome rows.")
  }

  if (is.matrix(p) != is.matrix(q)) {
    stop("p and q must have the same shape: both vectors or both matrices.")
  }

  if (is.matrix(p) && ncol(p) != ncol(q)) {
    stop("p and q must have the same number of columns.")
  }

  score_p <- switch(
    scoring_rule,
    "brier"     = brier_score(p, y),
    "spherical" = spherical_score(p, y),
    "log"       = log_score(p, y, eps = eps)
  )

  score_q <- switch(
    scoring_rule,
    "brier"     = brier_score(q, y),
    "spherical" = spherical_score(q, y),
    "log"       = log_score(q, y, eps = eps)
  )

  delta <- score_p - score_q
  estimate <- cumsum(delta) / seq_along(delta)

  bounded_rule <- scoring_rule %in% c("brier", "spherical")

  if (is.null(cs_type)) {
    cs_type <- if (bounded_rule) "bernstein" else "asymptotic"
  } else {
    cs_type <- match.arg(cs_type, c("bernstein", "hoeffding", "asymptotic", "none"))
  }

  lower <- rep(NA_real_, T_)
  upper <- rep(NA_real_, T_)

  if (compute_cs && cs_type != "none") {
    if (cs_type %in% c("bernstein", "hoeffding") && !bounded_rule) {
      stop(
        "Finite-sample Hoeffding/Bernstein confidence sequences require ",
        "bounded score differences. Use cs_type = 'asymptotic' for scoring_rule = 'log'."
      )
    }

    cs <- switch(
      cs_type,

      "bernstein" = cs_bernstein(
        scores1  = score_p,
        scores2  = score_q,
        alpha    = alpha,
        c        = 2,
        v_opt    = v_opt,
        boundary = boundary,
        lcb_only = lcb_only,
        ucb_only = ucb_only
      ),

      "hoeffding" = cs_hoeffding(
        scores1  = score_p,
        scores2  = score_q,
        alpha    = alpha,
        c        = 1,
        v_opt    = v_opt,
        boundary = boundary
      ),

      "asymptotic" = cs_asymptotic(
        scores1 = score_p,
        scores2 = score_q,
        alpha   = alpha
      )
    )

    lower <- cs$lower
    upper <- cs$upper
  }

  e_pq <- rep(NA_real_, T_)
  e_qp <- rep(NA_real_, T_)

  if (compute_e) {
    if (!bounded_rule) {
      stop(
        "e-processes in compare_forecasts() require bounded score differences. ",
        "Use compute_e = FALSE for scoring_rule = 'log', or use winkler_compare() ",
        "for binary log-score comparisons when the Winkler construction applies."
      )
    }

    ep <- eprocess(
      scores1  = score_p,
      scores2  = score_q,
      alpha    = alpha,
      c        = 2,
      v_opt    = v_opt,
      clip_max = clip_max
    )

    e_pq <- ep$e_pq
    e_qp <- ep$e_qp
  }

  out <- data.frame(
    t        = seq_len(T_),
    score_p  = score_p,
    score_q  = score_q,
    delta    = delta,
    estimate = estimate,
    lower    = lower,
    upper    = upper,
    e_pq     = e_pq,
    e_qp     = e_qp
  )

  attr(out, "scoring_rule") <- scoring_rule
  attr(out, "cs_type")      <- cs_type
  attr(out, "alpha")        <- alpha

  out
}

# =============================================================================
# confidence_sequences.R
# Anytime-valid confidence sequences for comparing two forecasters
#
# Implements:
#   cs_hoeffding()  — Theorem 1 CR23: sub-Gaussian (Hoeffding-style) CS
#   cs_bernstein()  — Theorem 2 CR23: empirical Bernstein (variance-adaptive) CS
#   cs_asymptotic() — Appendix C, Eq. 55 CR23: asymptotic CS for unbounded differences
#
# All three return a data.frame with columns:
#   t        — time index
#   estimate — running mean hat_Delta_t
#   lower    — lower confidence bound
#   upper    — upper confidence bound
#
# Coverage guarantee for cs_hoeffding and cs_bernstein:
#   P(forall t >= 1: Delta_t in [lower_t, upper_t]) >= 1 - alpha
#
# By design, cs_asymptotic does not have a finite-sample coverage guarantee.
# It is valid in the limit as m -> infinity under the assumption of finite
# variance. Waudby-Smith et al. (2021) define:
# lim inf_{m -> infinity} P(forall t >= m: mu_t in C^(m)_t) >= 1 - alpha
# m being the "peeking time".
#
# References:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4), 1368-1387
#   H21   Howard et al. (2021), Annals of Statistics 49(2), 1055-1080
#   WS21  Waudby-Smith et al. (2021), Annals of Statistics 52(6), 2613-2640
# =============================================================================

#' Hoeffding-style confidence sequence (Theorem 1, Choe & Ramdas 2023)
#'
#' Constructs a time-uniform confidence sequence for the mean score difference
#' \eqn{\Delta_t = \frac{1}{t} \sum_{i=1}^t E[\hat{\delta_i} \mid \mathcal{F}_{i-1}]}.
#'
#' @param scores1 Numeric vector. Scores S(p_t, y_t) for forecaster 1.
#' @param scores2 Numeric vector. Scores S(q_t, y_t) for forecaster 2.
#' @param alpha   Numeric in (0,1). Significance level. The CS has coverage
#'                1 - alpha uniformly over all t. Default: 0.05.
#' @param c       Numeric > 0. Sub-Gaussian scale. The process must satisfy
#'                |hat_delta_i| <= c for all i. For scores in `[a,b]`, the
#'                difference is in `[a-b, b-a]`, so c = b - a.
#'                Default: 1 (appropriate for Brier score differences in `[-1,1]`).
#' @param v_opt   Numeric > 0. Intrinsic time at which the CS is tightest.
#'                Default: 10 (recommended by CR23).
#' @param boundary Character. "mixture" (default, recommended) or "stitching".
#'
#' @return data.frame with columns t, estimate, lower, upper.
#'
#' @section Assumption:
#' Requires `hat_delta_i` to be c-sub-Gaussian given \eqn{\mathcal{F}_{i-1}},
#' i.e. `|hat_delta_i| <= c` for all `i`.
#'
#' @section Boundary:
#' \deqn{C_t^H = \hat\Delta_t \pm u^{CM}_{\alpha/2}(c^2 t; \rho) / t}
#' where \eqn{u^{CM}} is the normal mixture boundary and \eqn{c^2 t} is the
#' intrinsic time for a c-sub-Gaussian process with deterministic variance
#' proxy. The intrinsic time for Theorem 1 is `v_t = c^2 * t`, not `v_t = t`:
#' the CM boundary implicitly assumes 1-sub-Gaussian inputs, so the `c^2`
#' scaling must be applied explicitly. This matches the H21 convention,
#' where the boundary absorbs the sub-Gaussian parameter via the variance
#' process definition.
#'
#' Relation to Python comparecast: Python uses `v_t = sigma * t` where
#' `sigma = (hi - lo)/2 = c`. This is equivalent to our `c^2 * t` only when
#' `c = 1`. For `c != 1` the parametrisations differ; we follow the paper.
#'
#' @section Output:
#' Returns a `data.frame` with one row per `t` and columns `t`, `estimate`
#' (the running mean `hat_Delta_t`), `lower`, and `upper`, with coverage
#' guarantee \eqn{P(\forall t \geq 1 : \Delta_t \in [\text{lower}_t,
#' \text{upper}_t]) \geq 1 - \alpha}.
#'
#' @examples
#' scores1 <- c(-0.04, -0.09, -0.01, -0.16)
#' scores2 <- c(-0.09, -0.16, -0.04, -0.25)
#' cs_hoeffding(scores1, scores2, alpha = 0.05)
#'
#' @export
cs_hoeffding <- function(scores1, scores2,
                         alpha    = 0.05,
                         c        = 1,
                         v_opt    = 10,
                         boundary = "mixture") {

  stopifnot(
    length(scores1) == length(scores2),
    length(scores1) >= 1,
    alpha > 0, alpha < 1,
    c > 0,
    v_opt > 0,
    boundary %in% c("mixture", "stitching")
  )

  xs  <- scores1 - scores2          # hat_delta_t: score differences
  T_  <- length(xs)
  ts  <- seq_len(T_)                # time indices 1, 2, ..., T

  # Running mean: hat_Delta_t = (1/t) * sum_{i=1}^t hat_delta_i
  mus <- cumsum(xs) / ts

  # Intrinsic time for Theorem 1: deterministic, v_t = c^2 * t
  # For a c-sub-Gaussian process, the variance proxy is c^2 per observation.
  vs  <- c^2 * ts

  # One-sided alpha for two-sided CS: each side has crossing prob alpha/2
  alpha_os <- alpha / 2

  # Tuning parameter rho from v_opt
  rho <- rho_from_vopt(v_opt = v_opt, alpha = alpha_os)

  # Boundary values (cumulative-sum scale)
  if (boundary == "mixture") {
    u_t <- cm_boundary(v = vs, alpha = alpha_os, rho = rho)
  } else {
    # Polynomial stitching: c parameter enters boundary explicitly
    u_t <- ps_boundary(v = vs, alpha = alpha_os, v_opt = v_opt * c^2,
                       c = c, s = 1.4, eta = 2)
  }

  # CS radius: divide boundary by t to convert from cumulative-sum to mean scale
  radii <- u_t / ts

  data.frame(
    t        = ts,
    estimate = mus,
    lower    = mus - radii,
    upper    = mus + radii
  )
}

#' Empirical Bernstein confidence sequence (Theorem 2, Choe & Ramdas 2023)
#'
#' Constructs a variance-adaptive time-uniform CS using empirical intrinsic
#' time \eqn{\hat{V}_t = \sum_{i=1}^t (\hat{\delta}_i - \gamma_i)^2}.
#' Tighter than the Hoeffding CS when score differences have low variance.
#'
#' The CS is:
#' \deqn{C_t^{EB} = \hat{\Delta}_t \pm u_{\alpha/2}^{GE}(\hat{V}_t;\, \rho, c) \;/\; t}
#'
#' @param scores1  Numeric vector. Scores for forecaster 1.
#' @param scores2  Numeric vector. Scores for forecaster 2.
#' @param alpha    Numeric in (0,1). Significance level. Default: 0.05.
#' @param c        Numeric > 0. Sub-exponential scale. The process must satisfy
#'                 |hat_delta_i| <= c/2. For score differences in `[a-b, b-a]`,
#'                 c = b - a (e.g. c = 2 for Brier score differences in `[-1,1]`).
#'                 Default: 2.
#' @param v_opt    Numeric > 0. Optimal intrinsic time. Default: 10.
#' @param boundary Character. "mixture" (default, GE mixture) or "stitching"
#'                 (polynomial stitched) or "hardcoded" (CR23 example formula,
#'                 only valid for alpha=0.05, c=1).
#' @param gammas   Numeric vector or NULL. Predictable centering sequence.
#'                 If NULL, constructed as lagged running mean (default).
#' @param lcb_only Logical. If TRUE, return lower CS only: `[lower, +Inf)`.
#'                 Requires finite lower bound on hat_delta_i; provide c.
#' @param ucb_only Logical. If TRUE, return upper CS only: `(-Inf, upper]`.
#'
#' @return data.frame with columns t, estimate, lower, upper.
#'         lower = -Inf if ucb_only = TRUE; upper = Inf if lcb_only = TRUE.
#'
#' @examples
#' scores1 <- c(-0.04, -0.09, -0.01, -0.16)
#' scores2 <- c(-0.09, -0.16, -0.04, -0.25)
#' cs_bernstein(scores1, scores2, alpha = 0.05)
#'
#' @export
cs_bernstein <- function(scores1, scores2,
                         alpha    = 0.05,
                         c        = 2,
                         v_opt    = 10,
                         boundary = "mixture",
                         gammas   = NULL,
                         lcb_only = FALSE,
                         ucb_only = FALSE) {

  stopifnot(
    length(scores1) == length(scores2),
    length(scores1) >= 1,
    alpha > 0, alpha < 1,
    c > 0,
    v_opt > 0,
    boundary %in% c("mixture", "stitching", "hardcoded"),
    !(lcb_only && ucb_only)   # cannot request both simultaneously
  )

  if (boundary == "hardcoded" && (abs(alpha - 0.05) > 1e-10 || abs(c - 1) > 1e-10)) {
    warning(
      "boundary='hardcoded' is only valid for alpha=0.05 and c=1. ",
      "Supplied alpha=", alpha, ", c=", c, ". ",
      "Switching to boundary='mixture'."
    )
    boundary <- "mixture"
  }

  xs  <- scores1 - scores2
  T_  <- length(xs)
  ts  <- seq_len(T_)

  # Running mean
  mus <- cumsum(xs) / ts

  # Predictable centering sequence gamma_i
  if (is.null(gammas)) {
    gammas <- make_gammas(xs, lag = 1)
  } else {
    stopifnot(length(gammas) == T_)
  }

  # One-sided gamma adjustment for one-sided CS (mirrors Python comparecast)
  # lcb_only: center is clipped from above to prevent upward drift of gammas
  # ucb_only: center is clipped from below
  # These adjustments ensure the variance process remains valid for one-sided CSs.
  if (lcb_only) {
    # For lower CS: gammas clipped to [-c/2, c/2] from above
    gammas <- pmin(c / 2, gammas)
  }
  if (ucb_only) {
    # For upper CS: gammas clipped from below
    gammas <- pmax(-c / 2, gammas)
  }

  # Empirical intrinsic time: V_hat_t = sum_{i=1}^t (hat_delta_i - gamma_i)^2
  # Floor at 1 prevents log(0) in boundary formulas at t=1.
  vs <- intrinsic_time(xs, gammas, floor = TRUE)

  # Alpha for one-sided vs two-sided
  alpha_use <- if (lcb_only || ucb_only) alpha else alpha / 2

  # Compute boundary values
  if (boundary == "mixture") {
    rho <- rho_from_vopt(v_opt = v_opt, alpha = alpha_use)
    u_t <- ge_boundary(v = vs, alpha = alpha_use, rho = rho, c = c)
  } else if (boundary == "stitching") {
    u_t <- ps_boundary(v = vs, alpha = alpha_use, v_opt = v_opt,
                       c = c, s = 1.4, eta = 2)
  } else {
    # hardcoded: CR23 example formula (alpha=0.05, c=1 only)
    u_t <- cs_boundary_cr23_hardcoded(vs)
  }

  # CS radius (mean scale)
  radii <- u_t / ts

  # Build output: replace one endpoint with +/-Inf for one-sided CSs
  lower <- if (ucb_only) rep(-Inf, T_) else mus - radii
  upper <- if (lcb_only) rep( Inf, T_) else mus + radii

  data.frame(
    t        = ts,
    estimate = mus,
    lower    = lower,
    upper    = upper
  )
}

#' Asymptotic confidence sequence (Appendix C, Eq. 55, Choe & Ramdas 2023)
#'
#' **Asymptotic, not finite-sample**: coverage `>= 1 - alpha` holds only as
#' `t -> infinity`. Valid without requiring bounded score differences —
#' requires only that `hat_delta_t` has finite variance — so it's appropriate
#' for tick loss and other scoring rules where hard bounds depend on
#' unbounded realised values. Suitable for large evaluation windows.
#'
#' @param scores1 Numeric vector. Scores for forecaster 1.
#' @param scores2 Numeric vector. Scores for forecaster 2.
#' @param alpha   Numeric in (0,1). Significance level. Default: 0.05.
#' @param t_star  Numeric > 0. Sample size at which CS is tightest.
#'                Default: length(scores1) (tightest at end of sample).
#'
#' @return data.frame with columns t, estimate, lower, upper.
#'
#' @details
#' \deqn{C_t^A = \hat\Delta_t \pm \sqrt{
#'   \frac{2(t \sigma^2_t \rho^2 + 1)}{t^2 \rho^2}
#'   \log\frac{\sqrt{t \sigma^2_t \rho^2 + 1}}{\alpha}}}
#' where \eqn{\sigma^2_t = \frac{1}{t}\sum_{i=1}^t (\hat\delta_i - \hat\Delta_{i-1})^2}
#' and `rho` is tuned to be tightest at `t_star`:
#' \deqn{\rho(t_{star}) = \sqrt{\frac{2\log(1/\alpha) + \log(1 + 2\log(1/\alpha))}{t_{star}}}}
#'
#' The running variance estimator uses the predictable mean `hat_Delta_{t-1}`
#' (not the current mean `hat_Delta_t`) to maintain predictability, with
#' `hat_Delta_0 := 0`.
#'
#' @examples
#' scores1 <- c(-0.4, -0.2, -0.3, -0.1, -0.2)
#' scores2 <- c(-0.5, -0.3, -0.4, -0.2, -0.3)
#' cs_asymptotic(scores1, scores2, alpha = 0.05)
#'
#' @export
cs_asymptotic <- function(scores1, scores2,
                          alpha  = 0.05,
                          t_star = NULL) {

  stopifnot(
    length(scores1) == length(scores2),
    length(scores1) >= 1,
    alpha > 0, alpha < 1
  )

  xs  <- scores1 - scores2
  T_  <- length(xs)
  ts  <- seq_len(T_)

  # Default: tightest at end of sample
  if (is.null(t_star)) t_star <- T_

  # Tuning parameter rho
  rho <- sqrt((2 * log(1 / alpha) + log(1 + 2 * log(1 / alpha))) / t_star)

  # Running mean hat_Delta_t
  mus <- cumsum(xs) / ts

  # Predictable running mean hat_Delta_{t-1}: lag by 1, initialise at 0
  mus_lag <- c(0, mus[-T_])

  # Running variance estimator using predictable mean
  # sigma2_t = (1/t) * sum_{i=1}^t (hat_delta_i - hat_Delta_{i-1})^2
  sigma2_t <- cumsum((xs - mus_lag)^2) / ts

  # CS half-width (radius)
  # inner = t * sigma2_t * rho^2 + 1
  inner  <- ts * sigma2_t * rho^2 + 1

  # radius = sqrt( (2 * inner / (t^2 * rho^2)) * log(sqrt(inner) / alpha) )
  radius <- sqrt(
    (2 * inner / (ts^2 * rho^2)) *
      log(sqrt(inner) / alpha)
  )

  data.frame(
    t        = ts,
    estimate = mus,
    lower    = mus - radius,
    upper    = mus + radius
  )
}

# =============================================================================
# tests/testthat/test_confidence_sequences.R
#
# Tests fall into two categories:
#   A. Structural tests: correct output format, monotone width, etc.
#   B. Coverage simulation: the key anytime-valid guarantee.
#      Run 1000 paths under the null; verify that the CS never exits
#      in >= (1 - alpha) of paths, uniformly over all t.
#      These are gated behind SEQCOMP_RUN_SLOW_TESTS=true since they're
#      too heavy for ordinary devtools::check() runs.
# =============================================================================

set.seed(42)
s1 <- brier_score(runif(200), rbinom(200, 1, 0.5))
s2 <- brier_score(runif(200), rbinom(200, 1, 0.5))

# ── A. STRUCTURAL TESTS ──────────────────────────────────────────────────────

test_that("cs_hoeffding returns a well-formed data.frame", {
  cs_h <- cs_hoeffding(s1, s2, alpha = 0.05, c = 1)

  expect_true(is.data.frame(cs_h))
  expect_true(all(c("t", "estimate", "lower", "upper") %in% names(cs_h)))
  expect_equal(nrow(cs_h), 200)
  expect_equal(cs_h$t, 1:200)
})

test_that("cs_hoeffding bounds are ordered and have positive width", {
  cs_h <- cs_hoeffding(s1, s2, alpha = 0.05, c = 1)

  expect_true(all(cs_h$lower < cs_h$estimate & cs_h$estimate < cs_h$upper))
  expect_true(all(cs_h$upper - cs_h$lower > 0))
})

test_that("cs_hoeffding width shrinks over time", {
  cs_h <- cs_hoeffding(s1, s2, alpha = 0.05, c = 1)
  widths <- cs_h$upper - cs_h$lower
  # Allow for slight non-monotonicity early on (t=1 to t=2 can wobble due to rho);
  # check the broad trend instead of strict monotonicity.
  expect_lt(mean(widths[151:200]), mean(widths[1:50]))
})

test_that("cs_bernstein returns a well-formed data.frame", {
  cs_eb <- cs_bernstein(s1, s2, alpha = 0.05, c = 2)

  expect_true(is.data.frame(cs_eb))
  expect_true(all(c("t", "estimate", "lower", "upper") %in% names(cs_eb)))
  expect_equal(nrow(cs_eb), 200)
  expect_true(all(cs_eb$lower < cs_eb$estimate & cs_eb$estimate < cs_eb$upper))
  expect_true(all(cs_eb$upper - cs_eb$lower > 0))
})

test_that("EB CS is tighter than Hoeffding under low-variance score differences", {
  set.seed(7)
  p1 <- pmin(pmax(rnorm(500, 0.5, 0.05), 0.01), 0.99)
  p2 <- pmin(pmax(rnorm(500, 0.5, 0.05), 0.01), 0.99)
  y  <- rbinom(500, 1, 0.5)
  s1_lv <- brier_score(p1, y)
  s2_lv <- brier_score(p2, y)

  cs_h_lv  <- cs_hoeffding(s1_lv, s2_lv, alpha = 0.05, c = 1)
  cs_eb_lv <- cs_bernstein(s1_lv, s2_lv, alpha = 0.05, c = 2)

  width_h  <- cs_h_lv$upper  - cs_h_lv$lower
  width_eb <- cs_eb_lv$upper - cs_eb_lv$lower

  expect_lt(mean(width_eb), mean(width_h))
  expect_lt(width_eb[500], width_h[500])
})

test_that("lcb_only produces an upper bound of Inf and a finite lower bound", {
  cs_lcb <- cs_bernstein(s1, s2, alpha = 0.05, c = 2, lcb_only = TRUE)

  expect_true(all(is.infinite(cs_lcb$upper)))
  expect_true(all(is.finite(cs_lcb$lower)))
})

test_that("ucb_only produces a lower bound of -Inf and a finite upper bound", {
  cs_ucb <- cs_bernstein(s1, s2, alpha = 0.05, c = 2, ucb_only = TRUE)

  expect_true(all(is.infinite(cs_ucb$lower) & cs_ucb$lower < 0))
  expect_true(all(is.finite(cs_ucb$upper)))
})

test_that("hardcoded boundary (alpha=0.05, c=1) produces valid output", {
  cs_hc <- cs_bernstein(s1, s2, alpha = 0.05, c = 1, boundary = "hardcoded")

  expect_true(all(is.finite(cs_hc$lower)))
  expect_true(all(is.finite(cs_hc$upper)))
  expect_true(all(cs_hc$lower < cs_hc$upper))
})

test_that("hardcoded boundary warns and falls back to the mixture boundary for unsupported alpha/c", {
  expect_warning(
    cs_warn <- cs_bernstein(s1, s2, alpha = 0.10, c = 2, boundary = "hardcoded"),
    regexp = "hardcoded"
  )
  expect_true(all(is.finite(cs_warn$lower)))
})

test_that("cs_asymptotic returns a well-formed data.frame", {
  cs_asy <- cs_asymptotic(s1, s2, alpha = 0.05)

  expect_true(is.data.frame(cs_asy))
  expect_true(all(c("t", "estimate", "lower", "upper") %in% names(cs_asy)))
  expect_equal(nrow(cs_asy), 200)
  expect_true(all(cs_asy$lower < cs_asy$estimate & cs_asy$estimate < cs_asy$upper))
})

test_that("cs_asymptotic width shrinks over time and stays finite", {
  cs_asy <- cs_asymptotic(s1, s2, alpha = 0.05)

  expect_lt(
    mean(cs_asy$upper[151:200] - cs_asy$lower[151:200]),
    mean(cs_asy$upper[1:50]    - cs_asy$lower[1:50])
  )
  expect_true(all(is.finite(cs_asy$lower)))
  expect_true(all(is.finite(cs_asy$upper)))
})

test_that("cs_asymptotic produces a finite, positive-width CS under low variance", {
  set.seed(7)
  p1 <- pmin(pmax(rnorm(500, 0.5, 0.05), 0.01), 0.99)
  p2 <- pmin(pmax(rnorm(500, 0.5, 0.05), 0.01), 0.99)
  y  <- rbinom(500, 1, 0.5)
  s1_lv <- brier_score(p1, y)
  s2_lv <- brier_score(p2, y)

  cs_asy_lv <- cs_asymptotic(s1_lv, s2_lv, alpha = 0.05)
  width_asy <- cs_asy_lv$upper - cs_asy_lv$lower

  expect_true(all(width_asy > 0))
  expect_true(all(is.finite(width_asy)))
})

# ── B. COVERAGE SIMULATION ───────────────────────────────────────────────────
#
# Key property: for any stopping time tau, P(Delta_tau not in C_tau) <= alpha.
# Stronger: P(exists t: Delta_t not in C_t) <= alpha (uniform over all t).
#
# Simulation design:
#   - T = 500 observations per path
#   - N = 1000 paths
#   - Null: Delta_t = 0 (both forecasters identical, score diffs iid)
#   - Coverage = fraction of paths where CS contains 0 at ALL time points
#   - Expected: coverage >= 1 - alpha = 0.95
#
# Run locally with: SEQCOMP_RUN_SLOW_TESTS=true Rscript -e 'devtools::test()'

test_that("cs_hoeffding achieves nominal (1 - alpha) uniform coverage under the null", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow coverage simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  N   <- 1000
  T_  <- 500
  alpha_sim <- 0.05
  c_sim     <- 1   # |delta_i| <= 1 for Brier differences

  n_covered_h <- 0L
  set.seed(100)
  for (i in seq_len(N)) {
    xs <- sample(c(-1, 1), T_, replace = TRUE)  # Rademacher: saturates c=1 bound
    cs_i <- cs_hoeffding(scores1 = xs, scores2 = rep(0, T_),
                         alpha = alpha_sim, c = c_sim)
    if (all(cs_i$lower <= 0 & 0 <= cs_i$upper)) n_covered_h <- n_covered_h + 1L
  }
  coverage_h <- n_covered_h / N

  expect_gte(coverage_h, 1 - alpha_sim)
  expect_lt(coverage_h, 1.00)  # not trivially conservative
})

test_that("cs_bernstein achieves nominal (1 - alpha) uniform coverage under the null", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow coverage simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  N   <- 1000
  T_  <- 500
  alpha_sim <- 0.05

  n_covered_eb <- 0L
  set.seed(200)
  for (i in seq_len(N)) {
    xs <- sample(c(-1, 1), T_, replace = TRUE)
    cs_i <- cs_bernstein(scores1 = xs, scores2 = rep(0, T_),
                         alpha = alpha_sim, c = 2)  # c = hi - lo = 1 - (-1)
    if (all(cs_i$lower <= 0 & 0 <= cs_i$upper)) n_covered_eb <- n_covered_eb + 1L
  }
  coverage_eb <- n_covered_eb / N

  expect_gte(coverage_eb, 1 - alpha_sim)
  expect_lt(coverage_eb, 1.00)
})

test_that("EB CS is substantially tighter than Hoeffding under low-variance data", {
  T_ <- 500
  set.seed(300)
  xs_lv <- rnorm(T_, mean = 0, sd = 0.1)
  xs_lv <- pmax(pmin(xs_lv, 1), -1)

  cs_h_cmp  <- cs_hoeffding(xs_lv, rep(0, T_), alpha = 0.05, c = 1)
  cs_eb_cmp <- cs_bernstein(xs_lv, rep(0, T_), alpha = 0.05, c = 2)

  width_h_cmp  <- cs_h_cmp$upper  - cs_h_cmp$lower
  width_eb_cmp <- cs_eb_cmp$upper - cs_eb_cmp$lower

  expect_lt(mean(width_eb_cmp) / mean(width_h_cmp), 0.80)
})

test_that("cs_asymptotic achieves nominal (1 - alpha) uniform coverage under the null", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow coverage simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  N   <- 1000
  T_  <- 500
  alpha_sim <- 0.05

  n_covered_asy <- 0L
  set.seed(400)
  for (i in seq_len(N)) {
    xs <- sample(c(-1, 1), T_, replace = TRUE)
    cs_i <- cs_asymptotic(xs, rep(0, T_), alpha = alpha_sim)
    if (all(cs_i$lower <= 0 & 0 <= cs_i$upper)) n_covered_asy <- n_covered_asy + 1L
  }
  coverage_asy <- n_covered_asy / N

  expect_gte(coverage_asy, 0.95)
})

# =============================================================================
# utils.R
# Mathematical primitives for sequential forecast comparison
#
# Implements the low-level boundary functions and helper utilities called by
# confidence_sequences.R, etests.R, and all downstream modules.
#
# All formulae are referenced to their source:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4), 1368-1387
#   H21   Howard, Ramdas, McAuliffe & Sekhon (2021), Ann. Statist. 49(2)
#
# Dependencies: lamW (Lambert W function)
# Install via: install.packages("lamW")
# =============================================================================


# =============================================================================
# 1. TUNING PARAMETER: rho <-> v_opt mapping
# =============================================================================
#
# The GE mixture boundary and the CM boundary both take a tuning parameter rho.
# The user specifies v_opt (the intrinsic time at which the boundary is
# tightest), and rho is derived from it.
#
# Source: Howard et al. (2021) Proposition 3.
# Formula: rho = v_opt / (-W_{-1}(-alpha^2 / e) - 1)
# where W_{-1} is the lower branch of the Lambert W function.
#
# The Python comparecast package uses a different approximation:
#   rho_py = v_opt / (2*log(1/alpha_os) + log(1 + 2*log(1/alpha_os)))
#   where alpha_os = 2 * alpha_opt
#
# Both are implemented below. rho_from_vopt() uses the Lambert W formula
# (paper-exact). rho_from_vopt_python() mirrors the Python convention.
# Numerical agreement is checked in tests/test_utils.R.
# -----------------------------------------------------------------------------

#' Convert optimal intrinsic time to rho
#'
#' Maps the user-specified intrinsic time `v_opt` (the point at which a
#' boundary is tightest) to the `rho` tuning parameter, via the Lambert W
#' formula of Howard et al. (2021), Proposition 3 (paper-exact).
#'
#' @param v_opt  Numeric > 0. Intrinsic time at which the boundary is tightest.
#'               Recommended default from CR23: 10.
#' @param alpha  Numeric in (0,1). Significance level (one-sided).
#'               For a two-sided boundary at level alpha, pass alpha/2 here.
#'
#' @return Numeric > 0. The rho tuning parameter.
#'
#' @details
#' \deqn{\rho = \frac{v_{opt}}{-W_{-1}(-\alpha^2 / e) - 1}}
#' The lower branch \eqn{W_{-1}} is defined for `x` in `[-1/e, 0)` and returns
#' values `<= -1`. For `alpha` in `(0, 1)`, `-alpha^2/e` is always in
#' `(-1/e, 0)`, so the branch is well-defined.
#'
#' @examples
#' rho_from_vopt(v_opt = 10, alpha = 0.025)
#'
#' @export
rho_from_vopt <- function(v_opt = 10, alpha = 0.025) {
  stopifnot(v_opt > 0, alpha > 0, alpha < 1)
  arg    <- -(alpha^2) / exp(1)           # in (-1/e, 0) for alpha in (0,1)
  w_val  <- lamW::lambertWm1(arg)         # lower branch, returns value <= -1
  denom  <- -w_val - 1                    # > 0
  rho    <- v_opt / denom
  return(rho)
}

#' Convert v_opt to rho via Python comparecast approximation
#'
#' @param v_opt     Numeric > 0.
#' @param alpha_opt Numeric in (0,1). One-sided alpha used inside the e-process
#'                  (comparecast default: alpha/2 for two-sided tests).
#'
#' @return Numeric > 0.
#'
#' @details
#'   Python formula: rho = v_opt / (2*log(1/a) + log(1 + 2*log(1/a)))
#'   where a = 2 * alpha_opt.
#'   Used only for cross-validation against Python output; not used in
#'   production code.
#'
#' @keywords internal
#' @noRd
rho_from_vopt_python <- function(v_opt = 10, alpha_opt = 0.025) {
  stopifnot(v_opt > 0, alpha_opt > 0, alpha_opt < 1)
  a     <- 2 * alpha_opt
  denom <- 2 * log(1 / a) + log(1 + 2 * log(1 / a))
  rho   <- v_opt / denom
  return(rho)
}


# =============================================================================
# 2. BOUNDARY 1: Normal Mixture (Conjugate-Mixture, CM)
# =============================================================================
#
# Used for: Theorem 1 (Hoeffding CS), evaluated at v = t.
#
# Source: H21 Equation (3.7) / Proposition S1; CR23 Section 4.3.3.
#
# Formula:
#   u_CM(v; rho, alpha) = sqrt((v + rho) * log((v + rho) / (alpha^2 * rho)))
#
# Notes:
#   - alpha here is the ONE-SIDED alpha (pass alpha/2 for a two-sided CS).
#   - The scale c is implicit: the formula assumes the process has been scaled
#     so that deviations are 1-sub-Gaussian. In practice for CR23 Theorem 1
#     with |delta_i| <= c, the intrinsic time is v = c^2 * t. See Section 4.3.
#   - The formula is valid at v = 0: (0 + rho)*log(rho/(alpha^2*rho)) =
#     rho*log(1/alpha^2) > 0. No guard needed.
# -----------------------------------------------------------------------------

#' Normal mixture (CM) boundary
#'
#' @param v      Numeric vector >= 0. Intrinsic time values (V_t or t).
#' @param alpha  Numeric in (0,1). ONE-SIDED significance level.
#' @param rho    Numeric > 0. Tuning parameter. Obtain via rho_from_vopt().
#'
#' @return Numeric vector of boundary values (cumulative-sum scale, before /t).
#'
#' @examples
#'   rho <- rho_from_vopt(v_opt = 10, alpha = 0.025)
#'   u   <- cm_boundary(v = 1:500, alpha = 0.025, rho = rho)
#'
#' @seealso [rho_from_vopt()] to compute `rho` from a target `v_opt`.
#'
#' @export
cm_boundary <- function(v, alpha, rho) {
  stopifnot(all(v >= 0), alpha > 0, alpha < 1, rho > 0)
  sqrt((v + rho) * log((v + rho) / (alpha^2 * rho)))
}


# =============================================================================
# 3. BOUNDARY 2: Gamma-Exponential Mixture (GE)
# =============================================================================
#
# Used for: Theorem 2 (EB CS) boundary; Theorem 3 e-process value.
#
# Source: H21 Proposition S5; CR23 Proposition 3 / Appendix B.1.
#
# The function m(s, v) serves dual purpose:
#   (a) E-process value:    E_t^mix = m(S_t, V_hat_t)
#   (b) CS boundary:        u_GE(v; alpha) = sup{s : m(s,v) < 1/alpha}
#       solved numerically via ge_boundary() using uniroot().
#
# Formula (m(s, v)):
#   Let a = rho / c^2   (shape parameter)
#   Let x = (c*s + v + rho) / c^2
#
#   m(s, v) = [a^a / (Gamma(a) * gamma_reg(a, a))]
#             * Gamma(v/c^2 + a)                        [note: see (*) below]
#             * gamma_reg(v/c^2 + a,  x)
#             * x^(-(v/c^2 + a))
#             * exp((c*s + v) / c^2)
#
# (*) IMPORTANT: The second Gamma() call takes argument (v/c^2 + a),
#     not (v + rho/c^2). These are the same when a = rho/c^2, since
#     v/c^2 + rho/c^2 = (v + rho)/c^2. We use the fully expanded form
#     for clarity.
#
# gamma_reg(a, x) = pgamma(x, shape=a, rate=1) in R  [regularised lower IGF]
#
# Note on the denominator normalisation constant:
#   Gamma(a) * gamma_reg(a, a) = the UNREGULARISED lower incomplete gamma
#   evaluated at (a, a). This equals pgamma(a, shape=a, rate=1) * gamma(a).
#   In R: gamma(a) * pgamma(a, shape=a, rate=1).
# -----------------------------------------------------------------------------

#' Gamma-Exponential mixture function m(s, v)
#'
#' Core function for both the GE boundary (via uniroot) and the e-process
#' value (direct evaluation).
#'
#' @param s    Numeric. The "wealth" argument (cumulative sum S_t for e-process;
#'             boundary search variable for CS).
#' @param v    Numeric >= 0. Intrinsic time (V_hat_t).
#' @param rho  Numeric > 0. Tuning parameter from rho_from_vopt().
#' @param c    Numeric > 0. Sub-exponential scale. For Theorems 2&3 with
#'             |delta_i| <= c/2, use c = hi - lo (e.g. c=2 for scores in
#'             `[-1,1]`). For Proposition 4 Winkler scores, use c = 2.
#'
#' @return Numeric. Value of m(s, v). Used directly as E_t^mix when
#'         s = S_t and v = V_hat_t. Used inside ge_boundary() for root finding.
#'
#' @details
#'   Computed in log space for numerical stability, then exponentiated.
#'   The formula is valid for s < v/c + rho/c (otherwise x <= 0, log undefined).
#'   For large s (well past rejection threshold), the function returns Inf
#'   which the e-process wrapper clips to 1e7.
#'
#' @keywords internal
#' @noRd
ge_mixture <- function(s, v, rho, c) {
  stopifnot(rho > 0, c > 0, v >= 0)

  a  <- rho / c^2          # shape parameter
  x  <- (c * s + v + rho) / c^2   # upper limit for incomplete gamma

  # Guard: if x <= 0, m(s,v) is not defined (s is too negative)
  if (x <= 0) return(.Machine$double.eps)

  # Log-space computation for numerical stability
  # log m(s,v) = a*log(a) - log(Gamma(a)) - log(gamma_reg(a,a))
  #            + log(Gamma(v/c^2 + a))
  #            + log(gamma_reg(v/c^2 + a, x))
  #            - (v/c^2 + a) * log(x)
  #            + (c*s + v) / c^2

  shape2 <- v / c^2 + a   # = (v + rho) / c^2

  log_norm_const <- a * log(a) -
    lgamma(a) -
    log(pgamma(a, shape = a, rate = 1))   # log(gamma_reg(a,a))

  log_m <- log_norm_const +
    lgamma(shape2) +
    log(pgamma(x, shape = shape2, rate = 1)) -
    shape2 * log(x) +
    (c * s + v) / c^2

  return(exp(log_m))
}

#' Vectorised GE mixture
#'
#' Applies \code{ge_mixture()} element-wise over parallel vectors \code{s}
#' and \code{v} via \code{mapply}.
#'
#' @param s Numeric vector. Cumulative sum values.
#' @param v Numeric vector. Intrinsic time values, same length as \code{s}.
#' @param rho Numeric > 0. GE mixture tuning parameter.
#' @param c Numeric > 0. Sub-exponential scale.
#'
#' @return Numeric vector of GE mixture values.
#'
#' @keywords internal
#' @noRd
ge_mixture_vec <- function(s, v, rho, c) {
  mapply(ge_mixture, s = s, v = v, MoreArgs = list(rho = rho, c = c))
}


#' Gamma-exponential mixture boundary
#'
#' @param v      Numeric vector >= 0. Intrinsic time values.
#' @param alpha  Numeric in (0,1). ONE-SIDED significance level.
#' @param rho    Numeric > 0. From rho_from_vopt().
#' @param c      Numeric > 0. Sub-exponential scale.
#' @param s_lo   Numeric. Lower search bound for uniroot. Default: -10.
#' @param s_hi   Numeric. Upper search bound for uniroot. Default: 500.
#'
#' @return Numeric vector of boundary values (cumulative-sum scale).
#'
#' @importFrom stats pgamma uniroot
#'
#' @details
#' Computes \eqn{u_{GE}(v; \alpha, \rho, c) = \sup\{s : m(s,v) < 1/\alpha\}} by
#' solving `m(s, v) = 1/alpha` numerically for `s` via [uniroot()], separately
#' for each `v_i`.
#'
#' Root-finding fallback: the search starts in `[s_lo, s_hi]`; if
#' `m(s_hi, v_i)` has not yet crossed the target, `s_hi` is doubled once and
#' retried. If it still fails, a warning is issued and `s_hi` is returned as
#' a conservative fallback value. Increase `s_hi` directly if this warning
#' appears often (e.g. at large `v` or small `alpha`); increase `abs(s_lo)`
#' if no root is found at small `v`.
#'
#' Computed elementwise; can be slow for long vectors — consider caching
#' boundary values when the same `(alpha, rho, c)` are reused.
#'
#' @examples
#' rho <- rho_from_vopt(v_opt = 10, alpha = 0.025)
#' ge_boundary(v = 1:3, alpha = 0.025, rho = rho, c = 2)
#'
#' @export
ge_boundary <- function(v, alpha, rho, c, s_lo = -10, s_hi = 500) {
  stopifnot(alpha > 0, alpha < 1, rho > 0, c > 0)
  target <- 1 / alpha

  vapply(v, function(vi) {
    # Check if m(s_hi, vi) > target; if not, expand s_hi
    m_hi <- ge_mixture(s_hi, vi, rho, c)
    if (is.nan(m_hi) || m_hi < target) {
      # Try extending upper bound
      s_hi_try <- s_hi * 2
      m_hi_try <- ge_mixture(s_hi_try, vi, rho, c)
      if (!is.nan(m_hi_try) && m_hi_try >= target) {
        s_hi <- s_hi_try
      } else {
        warning(
          "ge_boundary: upper search bound may be too small at v = ", vi,
          ". Returning s_hi = ", s_hi, "."
        )
        return(s_hi)
      }
    }

    tryCatch(
      uniroot(
        function(s) ge_mixture(s, vi, rho, c) - target,
        lower  = s_lo,
        upper  = s_hi,
        tol    = 1e-8
      )$root,
      error = function(e) {
        warning("ge_boundary: uniroot failed at v = ", vi, ": ", e$message)
        NA_real_
      }
    )
  }, FUN.VALUE = numeric(1))
}



# =============================================================================
# 4. BOUNDARY 3: Polynomial Stitched (PS)
# =============================================================================
#
# Used for: Alternative to CM/GE for both Theorems 1 and 2.
#           NOT the recommended primary boundary (CM/GE are tighter in CR23).
#           Included for completeness and for cross-checking CR23 Table results.
#
# Source: H21 Theorem 1 / Equation (3.3); CR23 Section 4.3.2.
#
# Formula (scalar, l0 = 1):
#   S_alpha(v) = k1 * sqrt(v * L(v)) + c * k2 * L(v)
#   where L(v) = s * log(log(eta*v/m)) + log(zeta(s) / (alpha * log(eta)^s))
#   k1 = (eta^(1/4) + eta^(-1/4)) / sqrt(2)
#   k2 = (sqrt(eta) + 1) / 2
#   zeta(s) = Riemann zeta function
#
# Evaluated as S_alpha(max(v, m)) to ensure log(log(.)) is well-defined.
# The inner argument of the outer log is eta*(v vee m)/m >= eta > 1,
# so log(eta*(v vee m)/m) >= log(eta) > 0, and log(log(.)) is finite.
#
# Recommended defaults CR23: s = 1.4, eta = 2, v_opt (= m) = 10.
#
# The Riemann zeta function at non-integer arguments requires an external
# package. We use the 'VGAM' package, or fall back to a precomputed value
# for the recommended s = 1.4.
# -----------------------------------------------------------------------------

# Precomputed zeta values for recommended fixed s values.
# Avoids VGAM dependency for the standard use case.
.ZETA_PRECOMPUTED <- c(
  "1.4" = 3.6028,  # zeta(1.4), accurate to 4 decimal places
  "1.5" = 2.6124,  # zeta(1.5)
  "2.0" = 1.6449   # zeta(2.0) = pi^2/6
)

#' Riemann zeta function
#'
#' Uses VGAM if available, otherwise uses precomputed values for common s.
#'
#' @param s Numeric > 1.
#' @return Numeric. zeta(s).
#'
#' @keywords internal
#' @noRd
.zeta <- function(s) {
  key <- as.character(round(s, 1))
  if (!is.null(.ZETA_PRECOMPUTED[[key]])) {
    return(.ZETA_PRECOMPUTED[[key]])
  }
  if (requireNamespace("VGAM", quietly = TRUE)) {
    return(VGAM::zeta(s))
  }
  stop(
    "Riemann zeta(", s, ") not precomputed and 'VGAM' package not available.\n",
    "Install VGAM with: install.packages('VGAM'), or use s = 1.4 (default)."
  )
}

#' Polynomial stitched (PS) boundary
#'
#' Alternative boundary for both Theorem 1 and Theorem 2 constructions,
#' included for completeness and for cross-checking CR23 Table results.
#'
#' @param v      Numeric vector >= 0. Intrinsic time values.
#' @param alpha  Numeric in (0,1). ONE-SIDED significance level.
#' @param v_opt  Numeric > 0. Optimal intrinsic time (= m in H21).
#'               Default: 10.
#' @param c      Numeric > 0. Sub-exponential scale.
#' @param s      Numeric > 1. Stitching parameter. Default: 1.4.
#' @param eta    Numeric > 1. Geometric spacing. Default: 2.
#'
#' @return Numeric vector of boundary values (cumulative-sum scale).
#'
#' @details
#' This is **not** the recommended primary boundary: the CM/GE mixture
#' boundaries (`cm_boundary()`, `ge_boundary()`) are tighter in CR23 and are
#' used by default throughout `seqcomp`. Use `ps_boundary()` only when you
#' specifically need the polynomial-stitched construction.
#'
#' @examples
#' ps_boundary(v = 1:5, alpha = 0.025, v_opt = 10, c = 1)
#'
#' @export
ps_boundary <- function(v, alpha, v_opt = 10, c = 1, s = 1.4, eta = 2) {
  stopifnot(all(v >= 0), alpha > 0, alpha < 1, v_opt > 0, c > 0, s > 1, eta > 1)

  m   <- v_opt
  v_  <- pmax(v, m)   # v vee m: ensures log(log(.)) well-defined

  k1  <- (eta^(1/4) + eta^(-1/4)) / sqrt(2)
  k2  <- (sqrt(eta) + 1) / 2
  z_s <- .zeta(s)

  # Inner quantity L(v)
  L   <- s * log(log(eta * v_ / m)) + log(z_s / (alpha * log(eta)^s))

  boundary <- k1 * sqrt(v_ * L) + c * k2 * L
  return(boundary)
}


#' Hardcoded 95% Empirical Bernstein boundary from CR23
#'
#' Equation provided directly in Choe & Ramdas (2023) as an example.
#' Valid only for alpha = 0.05, c = 1. For other alpha, use ps_boundary()
#' or ge_boundary().
#'
#' @param v Numeric vector >= 0. Intrinsic time (V_hat_t).
#' @return Numeric vector of boundary values (cumulative-sum scale).
#'
#' @keywords internal
#' @noRd
cs_boundary_cr23_hardcoded <- function(v) {
  v_ <- pmax(v, 1)
  1.7 * sqrt(v_) * (log(log(2 * v_)) + 3.8) +
    3.4 * log(log(2 * v_)) + 13
}

# =============================================================================
# 5. INTRINSIC TIME HELPERS
# =============================================================================

#' Predictable centering sequence (default gamma_i)
#'
#' gamma_1 = 0, gamma_t = hat_Delta_{t-1} for t >= 2.
#' This is the lagged running mean of the score differences.
#'
#' @param xs Numeric vector of score differences (hat_delta_i).
#' @param lag Integer >= 1. Lag for the centering sequence. Default 1.
#'            For lag > 1 (stream-splitting setting), the first `lag` entries
#'            are set to 0, and the running mean is lagged by `lag` steps.
#'
#' @return Numeric vector of same length as xs.
#'
#' @details
#'   For lag = 1 (standard case):
#'     gamma_t = (1/(t-1)) * sum_{i=1}^{t-1} delta_i,  t >= 2
#'     gamma_1 = 0
#'   This matches both the paper default CR23 and the comparecast convention.
#'
#' @keywords internal
#' @noRd
make_gammas <- function(xs, lag = 1) {
  T_   <- length(xs)
  mus  <- cumsum(xs) / seq_along(xs)   # running mean
  gam  <- mus                           # copy
  if (lag >= T_) {
    gam[] <- 0
  } else {
    gam[(lag + 1):T_] <- mus[1:(T_ - lag)]
    gam[1:lag]        <- 0
  }
  return(gam)
}

#' Empirical intrinsic time V_hat_t
#'
#' V_hat_t = sum_{i=1}^t (delta_i - gamma_i)^2
#'
#' @param xs     Numeric vector of score differences.
#' @param gammas Numeric vector of predictable centres (same length as xs).
#'               If NULL, constructed via make_gammas(xs).
#' @param floor  Logical. If TRUE (default), apply pmax(1, V_hat_t) to prevent
#'               log(0) in boundary formulas. Set FALSE for e-process computation
#'               (the GE mixture does not need the floor).
#'
#' @return Numeric vector of cumulative intrinsic time values.
#'
#' @keywords internal
#' @noRd
intrinsic_time <- function(xs, gammas = NULL, floor = TRUE) {
  if (is.null(gammas)) gammas <- make_gammas(xs)
  vt <- cumsum((xs - gammas)^2)
  if (floor) vt <- pmax(1, vt)
  return(vt)
}


# =============================================================================
# 6. E-PROCESS LOG-SPACE COMPUTATION AND CLIPPING
# =============================================================================

#' Sub-exponential CGF-like function psi_{E,c}(lambda)
#'
#' psi_{E,c}(lambda) = (-log(1 - c*lambda) - c*lambda) / c^2
#'
#' Source: CR23 Theorem 3 formula.
#'
#' @param lambda Numeric. Must satisfy 0 <= lambda < 1/c.
#' @param c      Numeric > 0. Sub-exponential scale.
#'
#' @return Numeric.
#'
#' @keywords internal
#' @noRd
psi_e <- function(lambda, c) {
  stopifnot(lambda >= 0, lambda < 1 / c, c > 0)
  (-log(1 - c * lambda) - c * lambda) / c^2
}

#' Log fixed-lambda e-process from a score-difference stream
#'
#' Internal helper for computing the cumulative log e-process under a fixed
#' betting parameter \eqn{\lambda}, allowing a time-varying scale sequence
#' \eqn{c_1, c_2, \ldots}.
#'
#' At each step \eqn{i}, the log increment is
#' \deqn{\lambda x_i - \psi_E(\lambda, c_i)(x_i - \gamma_i)^2.}
#'
#' The key difference from \code{log_eprocess_fixed()} is that \code{c_seq}
#' is a vector. Passing \code{c_seq = rep(c0, length(xs))} recovers the
#' constant-scale case.
#'
#' @param xs Numeric vector. Score-difference stream.
#' @param c_seq Numeric vector, same length as \code{xs}. Time-varying
#'   sub-exponential scale parameters.
#' @param lambda Numeric scalar. Fixed betting parameter satisfying
#'   \eqn{0 \le \lambda < 1 / \max(c_seq)}.
#' @param gammas Numeric vector, same length as \code{xs}, or \code{NULL}.
#'   Predictable centering sequence. If \code{NULL}, constructed using
#'   \code{make_gammas(xs)}.
#'
#' @return Numeric vector of cumulative log e-process values.
#'
#' @keywords internal
#' @noRd
log_eprocess_fixed_predictable <- function(xs, c_seq, lambda, gammas = NULL) {
  stopifnot(length(xs) == length(c_seq), all(c_seq > 0))
  if (is.null(gammas)) gammas <- make_gammas(xs)
  stopifnot(length(gammas) == length(xs))

  c_0 <- max(c_seq)
  stopifnot(lambda >= 0, lambda < 1 / c_0)

  log_increments <- lambda * xs - psi_e(lambda, c_seq) * (xs - gammas)^2
  cumsum(log_increments)
}

#' Fixed-lambda log e-process
#'
#' log E_t(lambda) = lambda * S_t - psi_{E,c}(lambda) * V_hat_t
#'
#' @param xs     Numeric vector of score differences.
#' @param lambda Numeric in `[0, 1/c)`. Betting parameter.
#' @param c      Numeric > 0. Sub-exponential scale.
#' @param gammas Numeric vector of predictable centres, or NULL for default.
#'
#' @return Numeric vector of log e-process values.
#'
#' @keywords internal
#' @noRd
log_eprocess_fixed <- function(xs, lambda, c, gammas = NULL) {
  if (is.null(gammas)) gammas <- make_gammas(xs)
  S_t  <- cumsum(xs)
  V_t  <- intrinsic_time(xs, gammas, floor = FALSE)   # no floor for e-process
  lambda * S_t - psi_e(lambda, c) * V_t
}

#' Log GE mixture from cumulative sum and intrinsic time
#'
#' Internal helper for computing
#' `log(m(S_t, V_t))` from precomputed cumulative sums and intrinsic times.
#'
#' This is used when the caller has already constructed the signed cumulative
#' sum process and the matching intrinsic time process. It centralises the
#' numerical guard used throughout the package: non-positive or `NaN` GE mixture
#' values are replaced by `log(.Machine$double.eps)`.
#'
#' @param S_t Numeric vector. Cumulative sum process.
#' @param V_t Numeric vector. Intrinsic time process, same length as `S_t`.
#' @param rho Numeric > 0. GE mixture tuning parameter.
#' @param c Numeric > 0. Sub-exponential scale.
#'
#' @return Numeric vector of log GE mixture values.
#'
#' @keywords internal
#' @noRd
log_ge_mixture_from_sv <- function(S_t, V_t, rho, c) {
  stopifnot(length(S_t) == length(V_t), rho > 0, c > 0)

  mapply(
    function(s, v) {
      val <- ge_mixture(s, v, rho, c)
      if (val <= 0 || is.nan(val)) return(log(.Machine$double.eps))
      log(val)
    },
    s = S_t,
    v = V_t
  )
}

#' Mixture log e-process via GE mixture
#'
#' log E_t^mix = log m(S_t, V_hat_t)
#'
#' @param xs      Numeric vector of score differences.
#' @param rho     Numeric > 0. Tuning parameter.
#' @param c       Numeric > 0. Sub-exponential scale.
#' @param gammas  Numeric vector or NULL.
#'
#' @return Numeric vector of log e-process values.
#'
#' @keywords internal
#' @noRd
log_eprocess_mixture <- function(xs, rho, c, gammas = NULL) {
  if (is.null(gammas)) gammas <- make_gammas(xs)
  S_t <- cumsum(xs)
  V_t <- intrinsic_time(xs, gammas, floor = FALSE)
  log_ge_mixture_from_sv(S_t, V_t, rho, c)
}

#' Clip and exponentiate log e-process
#'
#' Clips log_e to `[log(clip_min), log(clip_max)]` before exponentiating.
#' Matches Python comparecast default (clip_max = 1e7).
#'
#' @param log_e    Numeric vector of log e-process values.
#' @param clip_max Numeric > 0. Maximum e-process value. Default: 1e7.
#' @param clip_min Numeric or NULL. Minimum (optional; NULL = no lower clip).
#'
#' @return Numeric vector of e-process values.
#'
#' @keywords internal
#' @noRd
clip_eprocess <- function(log_e, clip_max = 1e7, clip_min = NULL) {
  log_e <- pmin(log_e, log(clip_max))
  if (!is.null(clip_min) && clip_min > 0) {
    log_e <- pmax(log_e, log(clip_min))
  }
  exp(log_e)
}


# =============================================================================
# 7. UTILITY HELPERS
# =============================================================================

#' Running mean
#' @param xs Numeric vector.
#' @return Numeric vector of cumulative means.
#'
#' @keywords internal
#' @noRd
running_mean <- function(xs) cumsum(xs) / seq_along(xs)

#' Score difference bounds -> sub-Gaussian / sub-exponential scale
#'
#' Given score-difference bounds `[lo, hi]`, computes the two scale constants
#' used elsewhere in `seqcomp`:
#' * `c_thm1`  = `(hi - lo) / 2` — sub-Gaussian scale for Theorem 1, where
#'   `|delta_i| <= c_thm1`.
#' * `c_thm23` = `hi - lo` — sub-exponential scale for Theorems 2 & 3, where
#'   `|delta_i| <= c_thm23 / 2`.
#'
#' Both conventions bound the same quantity: after centering, `delta_i` lies
#' in `[-(hi-lo)/2, (hi-lo)/2]`, so `max|delta_i| = (hi-lo)/2`, which equals
#' both `c_thm1` and `c_thm23 / 2`.
#'
#' @param lo Numeric. Lower bound of score difference (usually a - b).
#' @param hi Numeric. Upper bound of score difference (usually b - a).
#'
#' @return Named list with elements c_thm1 and c_thm23.
#'
#' @examples
#' score_diff_scales(lo = -1, hi = 1)
#'
#' @export
score_diff_scales <- function(lo, hi) {
  stopifnot(hi > lo)
  c_val <- hi - lo
  list(
    c_thm1  = c_val / 2,  # Theorem 1: |delta_i| <= c, so c = (hi-lo)/2
    c_thm23 = c_val       # Theorems 2 & 3: |delta_i| <= c/2, so c = hi-lo
  )
}

#' Safe log (avoids log(0) = -Inf warnings in calibration)
#' @param x Numeric.
#' @param eps Numeric. Small positive guard. Default: 1e-16.
#'
#' @keywords internal
#' @noRd
.safe_log <- function(x, eps = 1e-16) log(pmax(x, eps))


# =============================================================================
# 8. NUMERICAL VERIFICATION: rho formula cross-check
# =============================================================================
#
# Run this once to check if Lambert W and Python approximation agree.
# Called explicitly in tests/testthat/test_utils.R, not on package load.
# -----------------------------------------------------------------------------

#' Verify that the two rho formulas agree numerically
#'
#' @param v_opt     Numeric. Default: 10.
#' @param alpha_vec Numeric vector of alpha values to check. Default: standard grid.
#' @param tol       Relative tolerance. Default: 0.01 (1%).
#'
#' @return Invisibly returns a data.frame of results; prints a summary.
#'
#' @keywords internal
#' @noRd
verify_rho_formulas <- function(
    v_opt     = 10,
    alpha_vec = c(0.005, 0.010, 0.025, 0.050, 0.100),
    tol       = 0.01
) {
  results <- lapply(alpha_vec, function(a) {
    rho_lw  <- rho_from_vopt(v_opt = v_opt, alpha = a)
    rho_py  <- rho_from_vopt_python(v_opt = v_opt, alpha_opt = a)
    rel_diff <- abs(rho_lw - rho_py) / rho_lw
    data.frame(
      alpha    = a,
      rho_lambertW = rho_lw,
      rho_python   = rho_py,
      rel_diff     = rel_diff,
      agree        = rel_diff < tol
    )
  })
  out <- do.call(rbind, results)
  cat("=== rho formula cross-check (v_opt =", v_opt, ") ===\n")
  print(out, row.names = FALSE, digits = 6)
  if (all(out$agree)) {
    cat("OK: All values agree within", tol * 100, "% relative tolerance.\n")
  } else {
    cat("WARNING: Some values differ by more than", tol * 100, "%.\n")
    cat("Using Lambert W formula (paper-exact) in production code.\n")
  }
  invisible(out)
}

# =============================================================================
# tests/testthat/test_utils.R
# Numerical verification for all primitives in R/utils.R
# =============================================================================

# ------------------------------------------------------------
# 1. rho_from_vopt (Lambert W formula)
# ------------------------------------------------------------

test_that("rho_from_vopt returns a positive, plausible value", {
  rho1 <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  expect_gt(rho1, 0)
  # known numerical range: -W_{-1}(-0.025^2/e) - 1 ~= 7.3, so rho ~= 1.37
  expect_gt(rho1, 0.5)
  expect_lt(rho1, 3.0)
})

test_that("rho_from_vopt increases with v_opt", {
  rho1      <- rho_from_vopt(v_opt = 10,  alpha = 0.025)
  rho_small <- rho_from_vopt(v_opt = 1,   alpha = 0.025)
  rho_large <- rho_from_vopt(v_opt = 100, alpha = 0.025)
  expect_lt(rho_small, rho1)
  expect_lt(rho1, rho_large)
})

test_that("rho_from_vopt changes with alpha", {
  rho_strict <- rho_from_vopt(v_opt = 10, alpha = 0.005)
  rho_loose  <- rho_from_vopt(v_opt = 10, alpha = 0.100)
  expect_false(isTRUE(all.equal(rho_strict, rho_loose)))
})

# ------------------------------------------------------------
# 2. rho formula cross-check (Lambert W vs Python approximation)
# ------------------------------------------------------------

test_that("Lambert W rho is always smaller than the Python approximation", {
  capture.output({
    result <- seqcomp:::verify_rho_formulas(v_opt = 10, tol = 0.02)
  })

  expect_true(all(result$rho_lambertW < result$rho_python))
})

# ------------------------------------------------------------
# 3. cm_boundary (Normal Mixture)
# ------------------------------------------------------------

test_that("cm_boundary returns positive, increasing values", {
  rho_cm <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  v_seq  <- c(1, 10, 100, 500)
  u_cm   <- cm_boundary(v = v_seq, alpha = 0.025, rho = rho_cm)

  expect_true(all(u_cm > 0))
  expect_true(all(diff(u_cm) > 0))
})

test_that("cm_boundary matches the closed-form formula at v=10", {
  rho_cm <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  u_cm10 <- cm_boundary(v = 10, alpha = 0.025, rho = rho_cm)
  manual <- sqrt((10 + rho_cm) * log((10 + rho_cm) / (0.025^2 * rho_cm)))

  expect_equal(u_cm10, manual, tolerance = 1e-10)
})

test_that("cm_boundary is finite at v=0", {
  rho_cm <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  u_cm0 <- cm_boundary(v = 0, alpha = 0.025, rho = rho_cm)
  expect_true(is.finite(u_cm0))
})

test_that("CM radius (u/t) shrinks as t grows", {
  rho_cm <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  t_seq  <- c(1, 10, 100, 500)
  radii  <- cm_boundary(v = t_seq, alpha = 0.025, rho = rho_cm) / t_seq
  expect_true(all(diff(radii) < 0))
})

# ------------------------------------------------------------
# 4. ge_mixture (GE mixture function m(s, v)) -- internal
# ------------------------------------------------------------

test_that("ge_mixture(0, 0) equals 1 (e-process starts at 1)", {
  rho_ge <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  m00 <- seqcomp:::ge_mixture(s = 0, v = 0, rho = rho_ge, c = 2)
  expect_equal(m00, 1, tolerance = 1e-6)
})

test_that("ge_mixture is positive and finite for positive s", {
  rho_ge <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  m_pos <- seqcomp:::ge_mixture(s = 5, v = 10, rho = rho_ge, c = 2)
  expect_gt(m_pos, 0)
  expect_true(is.finite(m_pos))
})

test_that("ge_mixture increases with s for fixed v", {
  rho_ge <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  m_s1 <- seqcomp:::ge_mixture(s = 1,  v = 10, rho = rho_ge, c = 2)
  m_s2 <- seqcomp:::ge_mixture(s = 5,  v = 10, rho = rho_ge, c = 2)
  m_s3 <- seqcomp:::ge_mixture(s = 10, v = 10, rho = rho_ge, c = 2)
  expect_lt(m_s1, m_s2)
  expect_lt(m_s2, m_s3)
})

test_that("ge_mixture(0, large v) stays finite and positive under the null", {
  rho_ge <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  m_null_large_v <- seqcomp:::ge_mixture(s = 0, v = 100, rho = rho_ge, c = 2)
  expect_true(is.finite(m_null_large_v))
  expect_gt(m_null_large_v, 0)
})

test_that("ge_mixture returns 0 for very negative s (x<=0 guard)", {
  rho_ge <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  m_neg <- seqcomp:::ge_mixture(s = -1000, v = 1, rho = rho_ge, c = 2)
  expect_lt(m_neg, 1e-12)
  expect_gt(m_neg, 0)
})

# ------------------------------------------------------------
# 5. ge_boundary (GE boundary via uniroot)
# ------------------------------------------------------------

test_that("ge_boundary is finite and increasing in v", {
  rho_ge2 <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  v_test  <- c(1, 5, 10, 50, 100)
  u_ge    <- ge_boundary(v = v_test, alpha = 0.025, rho = rho_ge2, c = 2)

  expect_true(all(is.finite(u_ge)))
  expect_true(all(diff(u_ge) > 0))
})

test_that("ge_boundary inverts ge_mixture to hit 1/alpha", {
  rho_ge2 <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  v_test  <- c(1, 5, 10, 50, 100)
  u_ge    <- ge_boundary(v = v_test, alpha = 0.025, rho = rho_ge2, c = 2)

  for (i in seq_along(v_test)) {
    m_check <- seqcomp:::ge_mixture(u_ge[i], v_test[i], rho = rho_ge2, c = 2)
    expect_equal(m_check, 40, tolerance = 0.1 / 40)
  }
})

test_that("GE radius (u/t) shrinks as t grows", {
  rho_ge2  <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  t_test   <- c(1, 5, 10, 50, 100)
  u_ge2    <- ge_boundary(v = t_test, alpha = 0.025, rho = rho_ge2, c = 2)
  radii_ge <- u_ge2 / t_test
  expect_true(all(diff(radii_ge) < 0))
})

# ------------------------------------------------------------
# 6. ps_boundary (Polynomial Stitched)
# ------------------------------------------------------------

test_that("ps_boundary returns finite, positive, non-decreasing values", {
  u_ps <- ps_boundary(v = c(1, 10, 100, 500), alpha = 0.025, v_opt = 10,
                       c = 2, s = 1.4, eta = 2)

  expect_true(all(is.finite(u_ps) & u_ps > 0))
  # PS boundary is non-decreasing, not strictly increasing
  expect_true(all(diff(u_ps) >= 0))
})

test_that("hardcoded CR23 formula matches the expected value at v=10", {
  hc_10 <- seqcomp:::cs_boundary_cr23_hardcoded(10)
  expect_lt(abs(hc_10 - 43.057), 0.01)
})

# ------------------------------------------------------------
# 7. intrinsic_time and make_gammas -- internal
# ------------------------------------------------------------

test_that("make_gammas produces the correctly lagged running mean", {
  set.seed(42)
  xs_test <- rnorm(100, mean = 0.1, sd = 0.5)
  gam <- seqcomp:::make_gammas(xs_test, lag = 1)

  expect_equal(length(gam), length(xs_test))
  expect_equal(gam[1], 0)
  expect_equal(gam[2], xs_test[1], tolerance = 1e-12)

  manual_gam <- c(0, cumsum(xs_test[-length(xs_test)]) /
                       seq_along(xs_test[-length(xs_test)]))
  expect_equal(gam, manual_gam, tolerance = 1e-12)
})

test_that("intrinsic_time behaves correctly with floor = TRUE", {
  set.seed(42)
  xs_test <- rnorm(100, mean = 0.1, sd = 0.5)
  vt_floor <- seqcomp:::intrinsic_time(xs_test, floor = TRUE)

  expect_equal(length(vt_floor), length(xs_test))
  expect_true(all(vt_floor >= 1))
  expect_true(all(diff(vt_floor) >= 0))
})

test_that("intrinsic_time runs without error when floor = FALSE", {
  set.seed(42)
  xs_test <- rnorm(100, mean = 0.1, sd = 0.5)
  expect_error(seqcomp:::intrinsic_time(xs_test, floor = FALSE), NA)
})

# ------------------------------------------------------------
# 8. psi_e (sub-exponential CGF) -- internal
# ------------------------------------------------------------

test_that("psi_e is positive for positive lambda and zero at lambda=0", {
  psi_val <- seqcomp:::psi_e(lambda = 0.1, c = 2)
  expect_gt(psi_val, 0)

  psi_zero <- seqcomp:::psi_e(lambda = 0, c = 2)
  expect_equal(psi_zero, 0, tolerance = 1e-12)
})

test_that("psi_e matches the manual formula at lambda=0.1, c=2", {
  psi_val <- seqcomp:::psi_e(lambda = 0.1, c = 2)
  manual_psi <- (-log(1 - 2 * 0.1) - 2 * 0.1) / 4
  expect_equal(psi_val, manual_psi, tolerance = 1e-12)
})

test_that("psi_e is increasing in lambda", {
  psi_seq <- vapply(
    seq(0.01, 0.4, by = 0.01),
    seqcomp:::psi_e,
    c = 2,
    FUN.VALUE = numeric(1)
  )
  expect_true(all(diff(psi_seq) > 0))
})

# ------------------------------------------------------------
# 9. log_eprocess_mixture and clip_eprocess -- internal
# ------------------------------------------------------------

test_that("log_eprocess_mixture is well-behaved under the null", {
  set.seed(99)
  xs_null <- rnorm(200, mean = 0, sd = 0.5)
  rho_ep <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  log_e_null <- seqcomp:::log_eprocess_mixture(xs_null, rho = rho_ep, c = 2)

  expect_equal(length(log_e_null), length(xs_null))
  expect_false(any(is.nan(log_e_null)))
})

test_that("log_eprocess_mixture grows more under the alternative than under the null", {
  set.seed(99)
  xs_null <- rnorm(200, mean = 0,   sd = 0.5)
  xs_alt  <- rnorm(200, mean = 0.3, sd = 0.5)
  rho_ep <- rho_from_vopt(v_opt = 10, alpha = 0.025)

  log_e_null <- seqcomp:::log_eprocess_mixture(xs_null, rho = rho_ep, c = 2)
  log_e_alt  <- seqcomp:::log_eprocess_mixture(xs_alt,  rho = rho_ep, c = 2)

  expect_gt(mean(log_e_alt), mean(log_e_null))
})

test_that("clip_eprocess caps values from above and keeps them positive", {
  set.seed(99)
  xs_null <- rnorm(200, mean = 0,   sd = 0.5)
  xs_alt  <- rnorm(200, mean = 0.3, sd = 0.5)
  rho_ep <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  log_e_alt <- seqcomp:::log_eprocess_mixture(xs_alt, rho = rho_ep, c = 2)

  e_clipped <- seqcomp:::clip_eprocess(log_e_alt, clip_max = 1e7)
  expect_true(all(e_clipped <= 1e7))
  expect_true(all(e_clipped > 0))
})

# ------------------------------------------------------------
# 10. score_diff_scales
# ------------------------------------------------------------

test_that("score_diff_scales computes correct constants for Brier score diffs", {
  # Brier score: bounds [-1,0], so score diff is in [-1,1]
  scales_brier <- score_diff_scales(lo = -1, hi = 1)
  expect_equal(scales_brier$c_thm1, 1)
  expect_equal(scales_brier$c_thm23, 2)
})

test_that("score_diff_scales computes correct constants for tick loss diffs", {
  # Tick loss at quantile 0.99: diff in [-0.99, 0.99]
  scales_tick <- score_diff_scales(lo = -0.99, hi = 0.99)
  expect_equal(scales_tick$c_thm23, 1.98, tolerance = 1e-10)
})

# =============================================================================
# tests/testthat/test_internal_eprocess_helpers.R
# =============================================================================

test_that("log_ge_mixture_from_sv matches direct GE mixture loop", {
  set.seed(123)
  xs <- runif(100, -0.5, 0.5)
  gammas <- make_gammas(xs)
  V_t <- intrinsic_time(xs, gammas, floor = FALSE)
  S_t <- cumsum(xs)

  rho <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  c <- 2

  direct <- mapply(
    function(s, v) {
      val <- ge_mixture(s, v, rho, c)
      if (val <= 0 || is.nan(val)) return(log(.Machine$double.eps))
      log(val)
    },
    s = S_t,
    v = V_t
  )

  helper <- log_ge_mixture_from_sv(S_t, V_t, rho, c)

  expect_equal(helper, direct, tolerance = 1e-12)
})

test_that("log_eprocess_fixed_predictable matches direct predictable loop", {
  set.seed(124)
  xs <- runif(100, -0.4, 0.4)
  c_seq <- rep(2, length(xs))
  lambda <- 0.2
  gammas <- make_gammas(xs)

  direct_increments <- numeric(length(xs))
  for (i in seq_along(xs)) {
    direct_increments[i] <-
      lambda * xs[i] - psi_e(lambda, c_seq[i]) * (xs[i] - gammas[i])^2
  }

  direct <- cumsum(direct_increments)
  helper <- log_eprocess_fixed_predictable(xs, c_seq, lambda, gammas)

  expect_equal(helper, direct, tolerance = 1e-12)
})

# =============================================================================
# lag_handling.R
# Propositions 5 & 6: sequential forecast comparison for lag-h forecasts
#
# When forecasts are made h steps ahead (h >= 2), the score difference
# process is no longer a martingale difference sequence with respect to
# the natural filtration. Stream splitting partitions the data into h
# interleaved subsequences, each of which IS a valid martingale sequence,
# then combines the resulting e-processes.
#
#   This module is implemented for theoretical completeness and to cover
#   Propositions 5 & 6 in the thesis.
#
# Two combination rules:
#   "pw"  — period-wise weak null: average e-processes across streams
#   "w"   — standard weak null: minimum e-process across streams
#
# Both use a p-to-e calibration step.
#
# References:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4), Section 4.4
# =============================================================================

#' Split a sequence into h interleaved lag streams
#'
#' For lag `h`, the k-th stream (`k = 1, ..., h`) contains indices
#' \eqn{\{k,\, k+h,\, k+2h,\, \ldots\}}, following the CR23 convention.
#'
#' @param xs  Numeric vector. Score differences hat_delta_t.
#' @param h   Integer >= 1. Lag (number of steps ahead).
#'
#' @return List of length h. Each element is a numeric vector
#'         containing the score differences for that stream.
#'
#' @examples
#'   split_streams(1:10, h = 3)
#'   # stream 1: indices 1, 4, 7, 10
#'   # stream 2: indices 2, 5, 8
#'   # stream 3: indices 3, 6, 9
#'
#' @export
split_streams <- function(xs, h) {
  stopifnot(h >= 1, length(xs) >= h)
  T_ <- length(xs)
  lapply(seq_len(h), function(k) xs[seq(k, T_, by = h)])
}

#' Unroll a stream-wise quantity back to the original time scale
#'
#' After computing a per-stream cumulative quantity (e.g. e-process values),
#' restores them to the original length `T` by zero-padding the first
#' `k - 1` positions, repeating each stream value `h` times, then truncating
#' to length `T`.
#'
#' @param stream_vals Numeric vector. Values for stream k (length ~ T/h).
#' @param k           Integer. Stream index (1-based).
#' @param h           Integer. Lag.
#' @param T_          Integer. Total original sequence length.
#'
#' @return Numeric vector of length T_.
#'
#' @section Alignment only, not theoretical updating:
#' For lagged forecasts (`h >= 2`), the returned series is aligned to the
#' evaluated score-difference index after stream splitting. It should
#' **not** be interpreted as a process that updates at the original
#' forecast-issuance time. The unrolled process is for visualization and
#' alignment only; the theoretical validity argument relies strictly on the
#' streamwise sub-filtrations, not on this unrolled representation.
#'
#' @examples
#' unroll_stream(c(1, 2, 3), k = 2, h = 2, T_ = 6)
#'
#' @export
unroll_stream <- function(stream_vals, k, h, T_) {
  padded    <- c(rep(0, k - 1), rep(stream_vals, each = h))
  length(padded) <- T_   # truncate or extend with NA then replace
  padded[is.na(padded)] <- stream_vals[length(stream_vals)]
  padded[seq_len(T_)]
}

#' P-to-e calibrator
#'
#' Converts anytime-valid p-values to e-values using the mixture or simple
#' calibrator, as used in CR23 Section 4.4.
#'
#' @param p        Numeric vector of p-values in (0, 1].
#' @param strategy Character. `"mixture"` (default, from Vovk & Wang 2021) or
#'                 `"simple"`. See Details for the formulas.
#' @param eps      Numeric. Numerical guard for log(0). Default: 1e-16.
#'
#' @return Numeric vector of e-values >= 0.
#'
#' @details
#' Mixture calibrator (default, matches Python comparecast behaviour):
#' \deqn{f(p) = \frac{1 - p + p\log(p)}{p\,(\log p)^2}}
#'
#' Simple calibrator (`strategy = "simple"`):
#' \deqn{f(p) = \frac{1}{2\sqrt{p}}}
#'
#' @examples
#' p <- c(0.5, 0.1, 0.01)
#' calibrate_p_to_e(p)
#' calibrate_p_to_e(p, strategy = "simple")
#'
#' @export
calibrate_p_to_e <- function(p, strategy = "mixture", eps = 1e-16) {
  stopifnot(all(p >= 0), all(p <= 1 + 1e-10))
  p <- pmin(pmax(p, eps), 1)

  if (strategy == "mixture") {
    num   <- 1 - p + p * log(p)
    denom <- p * log(p)^2 + eps
    pmax(num / denom, 0)
  } else if (strategy == "simple") {
    1 / (2 * sqrt(p) + eps)
  } else {
    stop("Unknown strategy: '", strategy, "'. Use 'mixture' or 'simple'.")
  }
}

#' Lag-h e-process for sequential forecast comparison (Propositions 5 & 6)
#'
#' For h-step-ahead forecasts, constructs an anytime-valid e-process by
#' stream splitting and combining h individual e-processes.
#'
#' @param scores1     Numeric vector. Scores for forecaster 1.
#' @param scores2     Numeric vector. Scores for forecaster 2.
#' @param h           Integer >= 1. Forecast lag.
#'                    For h=1, reduces to standard eprocess() — no splitting.
#' @param alpha       Numeric in (0,1). Significance level. Default: 0.05.
#' @param c           Numeric > 0. Sub-exponential scale. Default: 2.
#' @param v_opt       Numeric > 0. Default: 10.
#' @param null        Character. Null hypothesis type:
#' * `"pw"` — period-wise weak null (average combination).
#' * `"w"` — standard weak null (minimum combination).
#' @param calibrate   Logical. Apply p-to-e calibration. Default: TRUE.
#' @param cal_strategy Character. "mixture" (default) or "simple".
#'
#' @return data.frame with columns t, e_pq, e_qp, log_e_pq, log_e_qp.
#'
#' @details
#'   For h = 1: calls eprocess() directly and returns its output unchanged.
#'
#'   For h >= 2:
#'     1. Split xs into h streams
#'     2. Compute e-process on each stream independently
#'     3. Combine using the appropriate null rule
#'     4. Convert to p-process, combine, calibrate back to e-process
#'     5. Unroll to original time scale
#'
#'   The period-wise ("pw") null is less conservative than the standard ("w")
#'   null but tests a different (weaker) hypothesis. See CR23 Section 4.4.
#'
#' @examples
#' scores1 <- c(-0.04, -0.09, -0.01, -0.16, -0.04, -0.09)
#' scores2 <- c(-0.09, -0.16, -0.04, -0.25, -0.09, -0.16)
#' ep <- eprocess_lag(scores1, scores2, h = 2, alpha = 0.05)
#' head(ep)
#'
#' @export
eprocess_lag <- function(scores1, scores2,
                         h            = 1,
                         alpha        = 0.05,
                         c            = 2,
                         v_opt        = 10,
                         null         = "pw",
                         calibrate    = TRUE,
                         cal_strategy = "mixture") {

  stopifnot(
    length(scores1) == length(scores2),
    h >= 1,
    null %in% c("pw", "w")
  )

  # h = 1: trivial case, full martingale property holds
  if (h == 1L) {
    return(eprocess(scores1, scores2, alpha = alpha, c = c, v_opt = v_opt))
  }

  xs <- scores1 - scores2
  T_ <- length(xs)

  alpha_opt <- alpha / 2   # two-sided
  rho       <- rho_from_vopt(v_opt = v_opt, alpha = alpha_opt)

  # Split into h streams
  streams <- split_streams(xs, h)

  # Compute e-process on each stream
  # Each stream is a valid martingale difference sequence
  stream_epq <- vector("list", h)
  stream_eqp <- vector("list", h)

  for (k in seq_len(h)) {
    xs_k   <- streams[[k]]
    gam_k  <- make_gammas(xs_k, lag = 1)
    V_k    <- intrinsic_time(xs_k, gam_k, floor = FALSE)
    S_pq_k <- cumsum(xs_k)
    S_qp_k <- -S_pq_k

    stream_epq[[k]] <- clip_eprocess(
      log_ge_mixture_from_sv(S_pq_k, V_k, rho, c)
    )

    stream_eqp[[k]] <- clip_eprocess(
      log_ge_mixture_from_sv(S_qp_k, V_k, rho, c)
    ) }

  # Combine streams
  combine_streams <- function(stream_evals) {
    # Running maximum within each stream (anytime-valid p-value construction)
    run_max_list <- lapply(stream_evals, cummax)

    if (null == "pw") {
      # Period-wise: average running maxima, then convert to p
      run_max_mat <- do.call(cbind, lapply(seq_len(h), function(k) {
        unroll_stream(run_max_list[[k]], k = k, h = h, T_ = T_)
      }))
      mean_max_e <- pmax(rowMeans(run_max_mat), .Machine$double.eps)
      combined_p    <- pmin(1, exp(1) * log(max(h, 2)) / pmax(mean_max_e, .Machine$double.eps))

    } else {
      # Standard weak null: max p-value across streams (most conservative)
      p_per_stream <- lapply(seq_len(h), function(k) {
        pmin(1, 1 / pmax(
          unroll_stream(run_max_list[[k]], k = k, h = h, T_ = T_),
          .Machine$double.eps
        ))
      })
      p_mat         <- do.call(cbind, p_per_stream)
      combined_p    <- apply(p_mat, 1, max)
    }

    if (calibrate) {
      pmax(calibrate_p_to_e(combined_p, strategy = cal_strategy), .Machine$double.eps)
    } else {
      combined_p
    }
  }

  e_pq_combined <- combine_streams(stream_epq)
  e_qp_combined <- combine_streams(stream_eqp)

  # Clip final values
  e_pq_combined <- pmin(e_pq_combined, 1e7)
  e_qp_combined <- pmin(e_qp_combined, 1e7)

  data.frame(
    t        = seq_len(T_),
    e_pq     = e_pq_combined,
    e_qp     = e_qp_combined,
    log_e_pq = log(pmax(e_pq_combined, .Machine$double.eps)),
    log_e_qp = log(pmax(e_qp_combined, .Machine$double.eps))
  )
}

# =============================================================================
# tests/testthat/test_lag_handling.R
# =============================================================================

# --- A. split_streams ---

test_that("split_streams partitions a sequence into h interleaved streams", {
  xs <- 1:12
  streams <- split_streams(xs, h = 3)

  expect_equal(length(streams), 3)
  expect_equal(streams[[1]], c(1, 4, 7, 10))
  expect_equal(streams[[2]], c(2, 5, 8, 11))
  expect_equal(streams[[3]], c(3, 6, 9, 12))
})

test_that("split_streams with h=1 returns a single stream equal to the input", {
  xs <- 1:12
  streams1 <- split_streams(xs, h = 1)

  expect_equal(length(streams1), 1)
  expect_equal(streams1[[1]], xs)
})

# --- B. calibrate_p_to_e ---

test_that("calibrate_p_to_e produces non-negative, decreasing e-values", {
  p_vals <- c(0.001, 0.01, 0.05, 0.1, 0.5, 1.0)
  e_vals <- calibrate_p_to_e(p_vals)

  expect_true(all(e_vals >= 0))
  expect_true(all(diff(e_vals) <= 0))
  expect_lt(e_vals[length(e_vals)], 1e-3)  # p=1 gives e near 0
})

test_that("calibrate_p_to_e with strategy='simple' produces non-negative, decreasing e-values", {
  p_vals <- c(0.001, 0.01, 0.05, 0.1, 0.5, 1.0)
  e_simple <- calibrate_p_to_e(p_vals, strategy = "simple")

  expect_true(all(e_simple >= 0))
  expect_true(all(diff(e_simple) <= 0))
})

# --- C. eprocess_lag h=1 matches eprocess ---

test_that("eprocess_lag with h=1 matches eprocess() exactly", {
  set.seed(1)
  T_ <- 200
  xs <- sample(c(-1, 1), T_, replace = TRUE)

  ep_standard <- eprocess(xs, rep(0, T_), alpha = 0.05, c = 2)
  ep_lag1     <- eprocess_lag(xs, rep(0, T_), h = 1, alpha = 0.05, c = 2)

  expect_equal(ep_lag1$e_pq, ep_standard$e_pq, tolerance = 1e-8)
  expect_equal(ep_lag1$e_qp, ep_standard$e_qp, tolerance = 1e-8)
})

# --- D. eprocess_lag h=3 structure ---

test_that("eprocess_lag with h=3 returns correctly shaped, positive e-values under both null types", {
  set.seed(1)
  T_ <- 200
  xs <- sample(c(-1, 1), T_, replace = TRUE)

  ep_lag3_pw <- eprocess_lag(xs, rep(0, T_), h = 3, alpha = 0.05, c = 2, null = "pw")
  ep_lag3_w  <- eprocess_lag(xs, rep(0, T_), h = 3, alpha = 0.05, c = 2, null = "w")

  expect_equal(nrow(ep_lag3_pw), T_)
  expect_equal(nrow(ep_lag3_w),  T_)
  expect_true(all(ep_lag3_pw$e_pq > 0))
  expect_true(all(ep_lag3_w$e_pq  > 0))
})

test_that("pw-null e-values are on average at least as large as w-null e-values (less conservative)", {
  set.seed(1)
  T_ <- 200
  xs <- sample(c(-1, 1), T_, replace = TRUE)

  ep_lag3_pw <- eprocess_lag(xs, rep(0, T_), h = 3, alpha = 0.05, c = 2, null = "pw")
  ep_lag3_w  <- eprocess_lag(xs, rep(0, T_), h = 3, alpha = 0.05, c = 2, null = "w")

  expect_gte(mean(ep_lag3_pw$e_pq), mean(ep_lag3_w$e_pq))
})

# --- E. Type I error h=3 simulation ---
#
# Gated: N=500 paths x two eprocess_lag() calls each is too slow for routine
# devtools::check(). Run locally with:
#   SEQCOMP_RUN_SLOW_TESTS=true Rscript -e 'devtools::test()'

test_that("eprocess_lag (h=3) controls Type I error under both null types", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow Type I error simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  N_sim <- 500
  T_sim <- 300
  threshold <- 2 / 0.05
  n_rej_pw <- 0L
  n_rej_w  <- 0L

  set.seed(77)
  for (i in seq_len(N_sim)) {
    xs_i <- sample(c(-1, 1), T_sim, replace = TRUE)

    ep_pw <- eprocess_lag(xs_i, rep(0, T_sim), h = 3, alpha = 0.05, c = 2, null = "pw")
    ep_w  <- eprocess_lag(xs_i, rep(0, T_sim), h = 3, alpha = 0.05, c = 2, null = "w")

    if (any(ep_pw$e_pq >= threshold)) n_rej_pw <- n_rej_pw + 1L
    if (any(ep_w$e_pq  >= threshold)) n_rej_w  <- n_rej_w  + 1L
  }

  t1_pw <- n_rej_pw / N_sim
  t1_w  <- n_rej_w  / N_sim

  expect_lte(t1_pw, 0.05)
  expect_lte(t1_w,  0.05)
})

# --- F. pw combination grows under alternative (T=2000) ---

test_that("pw e_pq grows beyond 1 under a sustained alternative at T=2000", {
  set.seed(55)
  T_long <- 2000
  xs_alt_long <- pmax(pmin(runif(T_long, -0.7, 1.3), 1), -1)  # mean difference ~0.3

  ep_alt_long <- eprocess_lag(xs_alt_long, rep(0, T_long),
                              h = 3, alpha = 0.05, c = 2, null = "pw")

  expect_gt(max(ep_alt_long$e_pq), 1)
})

# =============================================================================
# predictable_bounds.R
# Proposition 7: e-process with time-varying predictable bounds
#
# When score differences have bounds that vary over time but are known
# one step ahead (F_{t-1}-measurable), the fixed-c assumption of Theorems
# 2 & 3 can be replaced by a predictable sequence (c_i).
#
# Proposition 7 CR23:
#   If |hat_delta_i| <= c_i/2 and (c_i) is strictly positive and
#   F_{i-1}-measurable, then for fixed lambda in [0, 1/c_0):
#
#   E_t(lambda) = prod_{i=1}^t exp{lambda*hat_delta_i - psi_{E,c_i}(lambda)
#                                  * (hat_delta_i - gamma_i)^2}
#
#   is a valid e-process for H_0^w(p, q).
#
# IMPORTANT LIMITATIONS:
#   - No closed-form mixture exists when c_i varies over time.
#   - Lambda must be fixed (no mixture over lambda).
#   - c_0 = sup_i(c_i) governs the domain: lambda in [0, 1/c_0).
#   - If c_0 = Inf, lambda must be 0 (trivial test). In practice,
#     c_i must be bounded from above for the test to have power.
#
# References:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4), Proposition 7
# =============================================================================

#' Fixed-lambda e-process with predictable bounds (Proposition 7)
#'
#' Constructs a valid e-process when score difference bounds vary over time
#' but are predictable (known at time i-1 before observing hat_delta_i).
#'
#' @param scores1  Numeric vector. Scores for forecaster 1.
#' @param scores2  Numeric vector. Scores for forecaster 2.
#' @param c_seq    Numeric vector. Predictable bound sequence (c_i), same
#'                 length as scores1. Must satisfy `|scores1[i]-scores2[i]| <= c_i/2`
#'                 and c_i > 0 for all i.
#' @param lambda   Numeric in `[0, 1/c_0)`. Betting parameter. Must be
#'                 strictly less than 1/c_0 where c_0 = max(c_seq).
#'                 If NULL, uses the recommended default lambda = 0.5/c_0.
#' @param alpha    Numeric in (0,1). Significance level for rejection rule.
#'                 Default: 0.05. Not used in computation, only for API
#'                 consistency. Pass the same value to [predictable_rejections()]
#'                 when evaluating rejection.
#' @param gammas   Numeric vector or NULL. Predictable centering sequence.
#'                 If NULL, constructed as lagged running mean.
#' @param clip_max Numeric. Maximum e-process value. Default: 1e7.
#' @param strict   Logical. If TRUE, any violation of the bound condition at
#'                 any time point will stop execution with an error. If FALSE
#'                 (default), a warning is issued but the e-process is still computed.
#'                 Note that violations invalidate the e-process guarantee,
#'                 so strict = TRUE is recommended for formal inference.
#'
#' @return data.frame with columns:
#'   t, e_pq, e_qp, log_e_pq, log_e_qp, c_seq, lambda_used
#'
#' @section Predictability:
#' The bound sequence `c_seq` (and the centering sequence `gammas`) must be
#' predictable: `c_i` is fixed and known at time `i - 1`, before
#' `scores1[i]`/`scores2[i]` (and hence `hat_delta_i`) are observed —
#' formally, \eqn{c_i} is \eqn{\mathcal{F}_{i-1}}-measurable. A bound chosen
#' after seeing `hat_delta_i` (e.g. derived from the realised data range)
#' invalidates the e-process guarantee, even if it numerically satisfies
#' `|hat_delta_i| <= c_i/2`.
#'
#' @details
#'   The e-process is computed as:
#'   \deqn{\log E_t(\lambda) = \sum_{i=1}^t \Bigl[\lambda\,\hat{\delta}_i
#'         - \psi_{E,c_i}(\lambda)\,(\hat{\delta}_i - \gamma_i)^2\Bigr]}
#'
#'   where
#'   \deqn{\psi_{E,c}(\lambda) = \frac{-\log(1 - c\lambda) - c\lambda}{c^2}}
#'   is evaluated at each step with the current \eqn{c_i}.
#'
#'   LAMBDA CHOICE: lambda = 0.5/c_0 is a conservative default that stays
#'   well within the valid domain `[0, 1/c_0)`. For better power, lambda can
#'   be tuned to the expected signal size, but must never reach 1/c_0.
#'
#'   VALIDITY CHECK: The function verifies |hat_delta_i| <= c_i/2 at each
#'   step and warns if violated. Violations invalidate the e-process guarantee.
#'
#' @examples
#' scores1 <- c(0.10, 0.20, 0.15, 0.25)
#' scores2 <- c(0.05, 0.10, 0.10, 0.20)
#' c_seq <- rep(1, length(scores1))
#' ep <- eprocess_predictable(scores1, scores2, c_seq = c_seq)
#' head(ep)
#'
#' @export
eprocess_predictable <- function(scores1, scores2,
                                 c_seq,
                                 lambda    = NULL,
                                 alpha     = 0.05,
                                 gammas    = NULL,
                                 clip_max  = 1e7,
                                 strict    = FALSE) {

  T_ <- length(scores1)
  stopifnot(
    length(scores2) == T_,
    length(c_seq)   == T_,
    all(c_seq > 0),
    alpha > 0, alpha < 1,
    clip_max > 0
  )

  xs  <- scores1 - scores2
  c_0 <- max(c_seq)

  # Validity check: |hat_delta_i| <= c_i/2 for all i
  violations <- which(abs(xs) > c_seq / 2 + 1e-8)
  valid_bound <- length(violations) == 0

  if (!valid_bound) {
    msg <- paste0(
      length(violations), " observation(s) violate |hat_delta_i| <= c_i/2. ",
      "First violation at t = ", violations[1], ". ",
      "The e-process validity guarantee is compromised. ",
      "Consider increasing c_seq at those time points."
    )
    if (strict) stop(msg) else warning(msg)
  }

  if (is.null(lambda)) {
    if (is.infinite(c_0))
      stop("c_0 = sup(c_seq) is infinite. Cannot construct a non-trivial e-process.",
           " Ensure c_seq is bounded from above.")
    lambda <- 0.5 / c_0
  }

  stopifnot(lambda >= 0, lambda < 1 / c_0)

  if (is.null(gammas)) {
    gammas <- make_gammas(xs, lag = 1)
  } else {
    stopifnot(length(gammas) == T_)
  }

  log_e_pq <- log_eprocess_fixed_predictable(xs, c_seq, lambda, gammas)
  log_e_qp <- log_eprocess_fixed_predictable(-xs, c_seq, lambda, -gammas)

  log_e_pq_clipped <- pmin(log_e_pq, log(clip_max))
  log_e_qp_clipped <- pmin(log_e_qp, log(clip_max))

  e_pq <- pmax(exp(log_e_pq_clipped), .Machine$double.eps)
  e_qp <- pmax(exp(log_e_qp_clipped), .Machine$double.eps)

  data.frame(
    t           = seq_len(T_),
    e_pq        = e_pq,
    e_qp        = e_qp,
    log_e_pq    = log(e_pq),
    log_e_qp    = log(e_qp),
    c_seq       = c_seq,
    lambda_used = lambda
  )
}

#' Summarise predictable bounds e-process
#'
#' @param ep    data.frame. Output of eprocess_predictable().
#' @param alpha Numeric. Significance level.
#'
#' @return Named list matching the `eprocess_rejections()` format
#' (`threshold`, `tau_pq`, `tau_qp`, `reject_pq`, `reject_qp`), plus:
#' * `c_range` — range of `c_seq` used.
#' * `lambda` — `lambda` value used.
#'
#' @examples
#' scores1 <- c(0.10, 0.20, 0.15, 0.25)
#' scores2 <- c(0.05, 0.10, 0.10, 0.20)
#' c_seq <- rep(1, length(scores1))
#' ep <- eprocess_predictable(scores1, scores2, c_seq = c_seq)
#' predictable_rejections(ep, alpha = 0.05)
#'
#' @export
predictable_rejections <- function(ep, alpha = 0.05) {
  threshold <- 2 / alpha
  tau_pq    <- which(ep$e_pq >= threshold)
  tau_qp    <- which(ep$e_qp >= threshold)

  list(
    threshold  = threshold,
    tau_pq     = if (length(tau_pq) > 0) tau_pq[1] else NA_integer_,
    tau_qp     = if (length(tau_qp) > 0) tau_qp[1] else NA_integer_,
    reject_pq  = length(tau_pq) > 0,
    reject_qp  = length(tau_qp) > 0,
    c_range    = range(ep$c_seq),
    lambda     = ep$lambda_used[1]
  )
}

# =============================================================================
# tests/testthat/test_predictable_bounds.R
# =============================================================================

set.seed(42)
T_ <- 300

# ── A. STRUCTURAL TESTS ──────────────────────────────────────────────────────

xs       <- sample(c(-1, 1), T_, replace = TRUE)
c_const  <- rep(2, T_)    # |xs| <= c/2 = 1
lambda_c <- 0.5 / 2       # = 0.25, default for c_0 = 2

ep_pred <- eprocess_predictable(
  scores1 = xs,
  scores2 = rep(0, T_),
  c_seq   = c_const,
  lambda  = lambda_c
)

test_that("eprocess_predictable returns a well-formed data.frame", {
  expect_true(is.data.frame(ep_pred))
  expect_true(all(c("t", "e_pq", "e_qp", "log_e_pq", "log_e_qp",
                    "c_seq", "lambda_used") %in% names(ep_pred)))
  expect_equal(nrow(ep_pred), T_)
})

test_that("eprocess_predictable e-values are strictly positive and respect the clip ceiling", {
  expect_true(all(ep_pred$e_pq > 0) && all(ep_pred$e_qp > 0))
  expect_true(all(ep_pred$e_pq <= 1e7) && all(ep_pred$e_qp <= 1e7))
  expect_true(all(is.finite(ep_pred$log_e_pq)) && all(is.finite(ep_pred$log_e_qp)))
})

test_that("eprocess_predictable records lambda_used and c_seq correctly", {
  expect_true(all(ep_pred$lambda_used == lambda_c))
  expect_true(all(ep_pred$c_seq == 2))
})

test_that("eprocess_predictable satisfies sign symmetry", {
  ep_rev <- eprocess_predictable(
    scores1 = -xs,
    scores2 = rep(0, T_),
    c_seq   = c_const,
    lambda  = lambda_c
  )

  expect_equal(ep_pred$e_pq, ep_rev$e_qp, tolerance = 1e-8)
  expect_equal(ep_pred$e_qp, ep_rev$e_pq, tolerance = 1e-8)
})

test_that("NULL lambda defaults to 0.5 / c_0", {
  ep_default <- eprocess_predictable(
    scores1 = xs,
    scores2 = rep(0, T_),
    c_seq   = c_const
  )
  expect_true(all(ep_default$lambda_used == 0.25))
})

test_that("eprocess_predictable warns on predictable-bound violations", {
  xs_bad <- c(1.5, rep(0.5, T_ - 1))   # first obs violates |xs| <= c/2 = 1
  c_bad  <- rep(2, T_)

  expect_warning(
    eprocess_predictable(
      scores1 = xs_bad,
      scores2 = rep(0, T_),
      c_seq   = c_bad,
      lambda  = 0.1
    )
  )
})

test_that("strict = TRUE turns a bound violation into an error", {
  xs_bad <- c(1.5, rep(0.5, T_ - 1))
  c_bad  <- rep(2, T_)

  expect_error(
    eprocess_predictable(
      scores1 = xs_bad,
      scores2 = rep(0, T_),
      c_seq   = c_bad,
      lambda  = 0.1,
      strict  = TRUE
    )
  )
})

test_that("lambda must be strictly less than 1 / max(c_seq)", {
  expect_error(
    eprocess_predictable(
      scores1 = xs,
      scores2 = rep(0, T_),
      c_seq   = c_const,
      lambda  = 0.5   # invalid: 1 / max(c_seq) = 0.5
    )
  )
})

test_that("negative lambda raises an error", {
  expect_error(
    eprocess_predictable(
      scores1 = xs,
      scores2 = rep(0, T_),
      c_seq   = c_const,
      lambda  = -0.1
    )
  )
})

test_that("different c_seq produces a different e-process path under the same lambda", {
  set.seed(10)
  xs_pos <- abs(rnorm(T_, 0.3, 0.3))
  xs_pos <- pmin(xs_pos, 0.9)
  c_tight <- rep(2,  T_)
  c_loose <- rep(10, T_)
  lam_fair <- 0.05   # must satisfy lambda < 1 / max(c_seq) for both

  ep_tight <- eprocess_predictable(
    scores1 = xs_pos, scores2 = rep(0, T_), c_seq = c_tight, lambda = lam_fair
  )
  ep_loose <- eprocess_predictable(
    scores1 = xs_pos, scores2 = rep(0, T_), c_seq = c_loose, lambda = lam_fair
  )

  expect_gt(max(abs(ep_tight$e_pq - ep_loose$e_pq)), 1e-12)
})

# ── B. REJECTION SUMMARY ─────────────────────────────────────────────────────

test_that("predictable_rejections returns a correctly structured two-sided summary", {
  rej <- predictable_rejections(ep_pred, alpha = 0.05)

  expect_true(is.list(rej))
  expect_true(all(c("threshold", "tau_pq", "tau_qp", "reject_pq", "reject_qp",
                    "c_range", "lambda") %in% names(rej)))
  expect_equal(rej$threshold, 40)
  expect_true(is.logical(rej$reject_pq) && length(rej$reject_pq) == 1)
  expect_true(is.logical(rej$reject_qp) && length(rej$reject_qp) == 1)
  expect_identical(as.numeric(rej$c_range), c(2, 2))
  expect_equal(rej$lambda, lambda_c)
})

# ── C. OPTIONAL LIGHT SANITY CHECK UNDER NULL ────────────────────────────────
#
# N=100, T=100 is two orders of magnitude cheaper than the other coverage
# simulations in this package, so this stays at normal test speed.

test_that("light null rejection rate is not obviously inflated under the predictable e-process", {
  set.seed(123)
  N_sim     <- 100
  T_sim     <- 100
  alpha_sim <- 0.05
  threshold <- 2 / alpha_sim
  n_rej     <- 0L

  for (i in seq_len(N_sim)) {
    xs_i <- sample(c(-1, 1), T_sim, replace = TRUE)
    c_i  <- rep(2, T_sim)
    ep_i <- eprocess_predictable(
      scores1 = xs_i, scores2 = rep(0, T_sim), c_seq = c_i, lambda = 0.2
    )
    if (any(pmax(ep_i$e_pq, ep_i$e_qp) >= threshold)) n_rej <- n_rej + 1L
  }
  type1_light <- n_rej / N_sim

  expect_lte(type1_light, 0.15)
})

# =============================================================================
# winkler.R
# Proposition 4 pipeline: Winkler score extension for unbounded scoring rules
#
# Choe & Ramdas (2023), Proposition 4:
#   For a proper scoring rule S on binary outcomes, the normalised Winkler
#   score w(p, q, y) is bounded above by 1 and satisfies the sub-exponential
#   condition required for Theorems 2 and 3 with scale c = 2.
#
# This file provides:
#   winkler_cs()   — one-sided EB confidence sequence for Winkler scores
#   winkler_etest() — e-process for Winkler scores
#   winkler_compare() — combined pipeline returning both
#
# IMPORTANT RESTRICTIONS (from Proposition 4 and Section G of CR23):
#   1. Binary outcomes only: y in {0, 1}, p, q in (0, 1).
#   2. The CS is ONE-SIDED (upper bound only) because the Winkler score is
#      unbounded below. Output: (-Inf, U_t].
#   3. Two-sided CS requires a finite analytical lower bound (Corollary 2),
#      which must be derived per scoring rule and forecaster pair.
#   4. NOT applicable to QLIKE or other continuous-outcome scoring rules.
#
# References:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4)
# =============================================================================

#' One-sided empirical Bernstein CS for Winkler scores (Proposition 4)
#'
#' Applies the Winkler normalisation and constructs a one-sided upper
#' confidence sequence for the mean Winkler score W_t = (1/t)*sum w_i.
#' The CS takes the form `(-Inf, U_t]`, valid uniformly over all t >= 1.
#'
#' @param p          Numeric vector in (0,1). Forecasts from model 1.
#' @param q          Numeric vector in (0,1). Forecasts from model 2.
#' @param y          Numeric vector containing only 0 and 1. Binary outcomes.
#' @param alpha      Numeric in (0,1). Significance level. Default: 0.05.
#' @param base_score Function. Underlying scoring rule. Default: log_score.
#' @param v_opt      Numeric > 0. Optimal intrinsic time. Default: 10.
#' @param lower_bound Numeric or NULL. Analytical lower bound on w_i for
#'                   two-sided CS via Corollary 2. If NULL (default), returns
#'                   one-sided CS only. If supplied, must satisfy w_i >= lower_bound
#'                   for all i almost surely.
#'
#' @return data.frame with columns t, estimate, lower, upper.
#'         lower = -Inf always (one-sided) unless lower_bound is supplied.
#'
#' @section Interpretation:
#' If `U_t < 0` for some `t`, this is time-uniform evidence that forecaster 1
#' (`p`) is worse than forecaster 2 (`q`) on average — i.e. a rejection is
#' evidence against `p`, not for it. More generally, `W_t > 0` suggests `p`
#' outperforms `q`; `W_t < 0` suggests `q` outperforms `p`.
#'
#' @details
#' Scale convention: Winkler score bounded above by 1, so c/2 = 1, c = 2.
#' This is hardcoded — do not change c without re-deriving the bound.
#'
#' @examples
#' p <- c(0.7, 0.6, 0.8, 0.65)
#' q <- c(0.5, 0.7, 0.6, 0.55)
#' y <- c(1, 1, 0, 1)
#' winkler_cs(p, q, y, alpha = 0.05)
#'
#' @export
winkler_cs <- function(p, q, y,
                       alpha       = 0.05,
                       base_score  = log_score,
                       v_opt       = 10,
                       lower_bound = NULL) {

  stopifnot(
    length(p) == length(q), length(p) == length(y),
    all(p > 0 & p < 1), all(q > 0 & q < 1),
    all(y %in% c(0, 1)),
    alpha > 0, alpha < 1
  )

  # Step 1: compute Winkler scores
  ws <- winkler_score(p, q, y, base_score = base_score)

  # Step 2: scale parameter — Winkler bounded above by 1, so c = 2
  c_wink <- 2

  if (is.null(lower_bound)) {
    # One-sided upper CS: (-Inf, U_t]
    # ucb_only = TRUE in cs_bernstein returns upper bound only
    cs <- cs_bernstein(
      scores1  = ws,
      scores2  = rep(0, length(ws)),
      alpha    = alpha,
      c        = c_wink,
      v_opt    = v_opt,
      ucb_only = TRUE
    )
  } else {
    # Two-sided CS via a finite lower bound.
    # For raw Winkler scores in [lower_bound, 1], our cs_bernstein() wrapper
    # expects a symmetric bound of the form |x_i| <= c/2.
    # A conservative compatible choice is therefore:
    #   c/2 >= max(1, -lower_bound)
    # i.e. c = 2 * max(1, -lower_bound).
    stopifnot(is.numeric(lower_bound), length(lower_bound) == 1, lower_bound < 1)

    if (any(ws < lower_bound - 1e-8)) {
      warning(
        "Some Winkler scores fall below the supplied lower_bound (",
        lower_bound, "). The two-sided CS validity may be compromised."
      )
    }

    c_twosided <- 2 * max(1, -lower_bound)

    cs <- cs_bernstein(
      scores1 = ws,
      scores2 = rep(0, length(ws)),
      alpha   = alpha,
      c       = c_twosided,
      v_opt   = v_opt
    )
  }

  cs
}

#' E-process for Winkler scores (Proposition 4 + Theorem 3)
#'
#' Tests whether the mean Winkler score W_t >= 0 for all t.
#' A rejection provides time-uniform evidence that forecaster 1 (p)
#' is worse than forecaster 2 (q) under the base scoring rule.
#'
#' @param p         Numeric vector in (0,1). Forecasts from model 1.
#' @param q         Numeric vector in (0,1). Forecasts from model 2.
#' @param y         Numeric vector containing only 0 and 1. Binary outcomes.
#' @param alpha     Numeric in (0,1). Significance level. Default: 0.05.
#' @param base_score Function. Underlying scoring rule. Default: log_score.
#' @param v_opt     Numeric > 0. Default: 10.
#' @param clip_max  Numeric. Maximum e-process value before clipping.
#'                  Default: 1e7.
#'
#' @return data.frame with columns t, e, log_e.
#'
#' @section Rejection rule:
#' Reject at level `alpha` when `e >= 1 / alpha`; this provides time-uniform
#' evidence that `p` is worse than `q`.
#'
#' @examples
#' p <- c(0.7, 0.6, 0.8, 0.65)
#' q <- c(0.5, 0.7, 0.6, 0.55)
#' y <- c(1, 1, 0, 1)
#' winkler_etest(p, q, y, alpha = 0.05)
#'
#' @export
winkler_etest <- function(p, q, y,
                          alpha      = 0.05,
                          base_score = log_score,
                          v_opt      = 10,
                          clip_max   = 1e7) {

  stopifnot(
    length(p) == length(q), length(p) == length(y),
    all(p > 0 & p < 1), all(q > 0 & q < 1),
    all(y %in% c(0, 1)),
    alpha > 0, alpha < 1,
    v_opt > 0,
    clip_max > 0
  )

  ws <- winkler_score(p, q, y, base_score = base_score)

  # Proposition 4 applies because the Winkler score is bounded above by 1.
  # To use the lower-bounded sub-exponential e-process machinery, negate it:
  #   x_i = -w_i >= -1
  xs <- -ws

  # Predictable centering sequence
  gammas <- make_gammas(xs, lag = 1)

  # Shared intrinsic time (no floor for e-processes)
  V_hat <- intrinsic_time(xs, gammas, floor = FALSE)
  S_t   <- cumsum(xs)

  # One-sided tuning parameter
  rho <- rho_from_vopt(v_opt = v_opt, alpha = alpha)

  log_e <- log_ge_mixture_from_sv(S_t, V_hat, rho, c = 2)

  e <- clip_eprocess(log_e, clip_max = clip_max)

  data.frame(
    t     = seq_along(xs),
    e     = e,
    log_e = pmin(log_e, log(clip_max))
  )
}

#' Summarise one-sided Winkler e-process rejections
#'
#' Internal helper used by `winkler_compare()`.
#'
#' @param ep Data frame returned by `winkler_etest()`, with an `e` column.
#' @param alpha Numeric in `(0, 1)`. Significance level.
#'
#' @return A list with elements `threshold`, `tau`, and `reject`.
#'
#' @keywords internal
#' @noRd
winkler_etest_rejections <- function(ep, alpha = 0.05) {
  threshold <- 1 / alpha
  tau <- which(ep$e >= threshold)

  list(
    threshold = threshold,
    tau       = if (length(tau) > 0) tau[1] else NA_integer_,
    reject    = length(tau) > 0
  )
}

#' Full Winkler comparison pipeline (Proposition 4)
#'
#' Convenience wrapper that computes Winkler scores, one-sided CS, and
#' e-process in a single call.
#'
#' @param p          Numeric vector in (0,1).
#' @param q          Numeric vector in (0,1).
#' @param y          Numeric vector containing only 0 and 1. Binary outcomes.
#' @param alpha      Numeric in (0,1). Default: 0.05.
#' @param base_score Function. Default: log_score.
#' @param v_opt      Numeric > 0. Default: 10.
#' @param lower_bound Numeric or NULL. See winkler_cs().
#'
#' @return Named list with elements:
#' * `winkler_scores` — raw Winkler score vector.
#' * `cs` — `data.frame` from `winkler_cs()`.
#' * `etest_p_worse` — one-sided e-process testing whether `p` is worse than `q`.
#' * `etest_q_worse` — one-sided e-process testing whether `q` is worse than `p`.
#' * `rejections` — list of one-sided rejection summaries.
#'
#' @examples
#' p <- c(0.7, 0.6, 0.8, 0.65)
#' q <- c(0.5, 0.7, 0.6, 0.55)
#' y <- c(1, 1, 0, 1)
#' winkler_compare(p, q, y, alpha = 0.05)
#'
#' @export
winkler_compare <- function(p, q, y,
                            alpha       = 0.05,
                            base_score  = log_score,
                            v_opt       = 10,
                            lower_bound = NULL) {

  ws_pq <- winkler_score(p, q, y, base_score = base_score)
  cs    <- winkler_cs(p, q, y, alpha = alpha, base_score = base_score,
                      v_opt = v_opt, lower_bound = lower_bound)

  ep_p_worse <- winkler_etest(
    p = p, q = q, y = y,
    alpha = alpha, base_score = base_score, v_opt = v_opt
  )

  ep_q_worse <- winkler_etest(
    p = q, q = p, y = y,
    alpha = alpha, base_score = base_score, v_opt = v_opt
  )

  rej_p_worse <- winkler_etest_rejections(ep_p_worse, alpha = alpha)
  rej_q_worse <- winkler_etest_rejections(ep_q_worse, alpha = alpha)

  list(
    winkler_scores = ws_pq,
    cs             = cs,
    etest_p_worse  = ep_p_worse,
    etest_q_worse  = ep_q_worse,
    rejections     = list(
      p_worse_than_q = rej_p_worse,
      q_worse_than_p = rej_q_worse
    )
  )
}

# =============================================================================
# tests/testthat/test_winkler.R
# =============================================================================

set.seed(42)
n <- 500
y <- rbinom(n, 1, 0.5)
# Two forecasters: p slightly better than q on average
p <- pmin(pmax(rbeta(n, 3, 2), 0.01), 0.99)   # skewed toward 1
q <- pmin(pmax(rbeta(n, 2, 2), 0.01), 0.99)   # symmetric

# --- A. winkler_cs: structure ---

test_that("winkler_cs returns a well-formed one-sided data.frame", {
  cs_w <- winkler_cs(p, q, y, alpha = 0.05)

  expect_true(is.data.frame(cs_w))
  expect_true(all(c("t", "estimate", "lower", "upper") %in% names(cs_w)))
  expect_equal(nrow(cs_w), n)
  expect_true(all(is.infinite(cs_w$lower) & cs_w$lower < 0))
  expect_true(all(is.finite(cs_w$upper)))
})

test_that("winkler_cs estimate is the running mean of the Winkler score", {
  cs_w <- winkler_cs(p, q, y, alpha = 0.05)
  ws <- winkler_score(p, q, y)
  expect_equal(cs_w$estimate, cumsum(ws) / seq_along(ws), tolerance = 1e-10)
})

test_that("winkler_cs upper bound always exceeds the estimate", {
  cs_w <- winkler_cs(p, q, y, alpha = 0.05)
  expect_true(all(cs_w$upper > cs_w$estimate))
})

# --- B. winkler_etest: one-sided structure ---

test_that("winkler_etest returns a well-formed one-sided e-process data.frame", {
  ep_w <- winkler_etest(p, q, y, alpha = 0.05)

  expect_true(is.data.frame(ep_w))
  expect_true(all(c("t", "e", "log_e") %in% names(ep_w)))
  expect_equal(nrow(ep_w), n)
  expect_true(all(ep_w$e > 0))
  expect_true(all(ep_w$e <= 1e7))
  expect_true(all(is.finite(ep_w$log_e)))
  expect_true(all(ep_w$log_e <= log(1e7) + 1e-12))
})

test_that("winkler_etest_rejections returns a correctly structured one-sided summary", {
  ep_w <- winkler_etest(p, q, y, alpha = 0.05)
  rej_w <- seqcomp:::winkler_etest_rejections(ep_w, alpha = 0.05)

  expect_equal(rej_w$threshold, 20)
  expect_true(all(c("threshold", "tau", "reject") %in% names(rej_w)))
  expect_true(is.logical(rej_w$reject) && length(rej_w$reject) == 1)
})

# --- C. winkler_compare: combined one-sided pipeline ---

test_that("winkler_compare returns the expected top-level structure", {
  wc <- winkler_compare(p, q, y, alpha = 0.05)
  expect_true(all(c("winkler_scores", "cs", "etest_p_worse", "etest_q_worse",
                    "rejections") %in% names(wc)))
})

test_that("winkler_compare winkler_scores are correctly sized and bounded above by 1", {
  wc <- winkler_compare(p, q, y, alpha = 0.05)
  expect_equal(length(wc$winkler_scores), n)
  expect_true(all(wc$winkler_scores <= 1 + 1e-10))
})

test_that("winkler_compare's cs matches a standalone winkler_cs() call", {
  wc <- winkler_compare(p, q, y, alpha = 0.05)
  cs_w <- winkler_cs(p, q, y, alpha = 0.05)
  expect_equal(wc$cs$upper, cs_w$upper, tolerance = 1e-10)
})

test_that("winkler_compare's etest components have the one-sided e-process columns", {
  wc <- winkler_compare(p, q, y, alpha = 0.05)

  expect_true(is.data.frame(wc$etest_p_worse))
  expect_true(all(c("t", "e", "log_e") %in% names(wc$etest_p_worse)))
  expect_true(is.data.frame(wc$etest_q_worse))
  expect_true(all(c("t", "e", "log_e") %in% names(wc$etest_q_worse)))
})

test_that("winkler_compare's rejections are correctly nested with threshold 20", {
  wc <- winkler_compare(p, q, y, alpha = 0.05)

  expect_true(all(c("p_worse_than_q", "q_worse_than_p") %in% names(wc$rejections)))
  expect_equal(wc$rejections$p_worse_than_q$threshold, 20)
  expect_equal(wc$rejections$q_worse_than_p$threshold, 20)
})

# --- D. one-sided guarantee: upper bound covers true mean ---
#
# Simulation: N paths under a symmetric null construction.
# The upper CS should contain 0 in at least about 1-alpha of paths.
# Gated: too slow for routine devtools::check(). Run locally with:
#   SEQCOMP_RUN_SLOW_TESTS=true Rscript -e 'devtools::test()'

test_that("winkler_cs one-sided upper bound achieves nominal coverage under a symmetric null", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow coverage simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  N_sim <- 500
  T_sim <- 300
  n_covered <- 0L
  set.seed(123)
  for (i in seq_len(N_sim)) {
    p_i <- pmin(pmax(rbeta(T_sim, 2, 2), 0.01), 0.99)
    q_i <- pmin(pmax(rbeta(T_sim, 2, 2), 0.01), 0.99)
    y_i <- rbinom(T_sim, 1, 0.5)
    cs_i <- winkler_cs(p_i, q_i, y_i, alpha = 0.05)
    if (all(0 <= cs_i$upper)) n_covered <- n_covered + 1L
  }
  coverage_w <- n_covered / N_sim

  expect_gte(coverage_w, 0.95)
})

# --- E. two-sided CS with lower_bound ---
#
# A finite lower_bound activates the two-sided CS branch. brier_score keeps
# the underlying score bounded here. The supplied lower_bound is interpreted
# as a user-supplied analytical bound on the Winkler scores; the function
# checks observed violations but cannot verify the mathematical guarantee.
# For unbounded scores such as the log score, no finite lower bound can be
# guaranteed in general, so the one-sided CS is the safer/default procedure.

test_that("winkler_cs with a finite lower_bound produces a valid two-sided CS", {
  cs_2s <- winkler_cs(
    p = p, q = q, y = y, alpha = 0.05,
    base_score = brier_score,
    lower_bound = -10
  )

  expect_true(all(is.finite(cs_2s$lower)))
  expect_true(all(is.finite(cs_2s$upper)))
  expect_true(all(cs_2s$lower < cs_2s$upper))
})

# =============================================================================
# etests.R
# Sequential e-processes for testing the weak null hypothesis
#
# Implements:
#   eprocess()  — Theorem 3 CR23: sub-exponential mixture e-process
#
# The weak null being tested:
#   H_0^w(p, q): Delta_t = (1/t) * sum_{i=1}^t E[hat_delta_i | F_{i-1}] <= 0
#   for all t = 1, 2, ...
#   i.e. forecaster 1 is no better than forecaster 2 on average.
#
# Two one-sided e-processes are run simultaneously:
#   e_pq: tests H_0^w(p, q) — is p better than q?
#   e_qp: tests H_0^w(q, p) — is q better than p?
#
# Rejection rule (two-sided, level alpha):
#   Reject H_0^w(p, q) if e_pq >= 2/alpha  [p outperforms q]
#   Reject H_0^w(q, p) if e_qp >= 2/alpha  [q outperforms p]
#
# Returns a data.frame with columns:
#   t      — time index
#   e_pq   — e-process value for H_0^w(p, q)
#   e_qp   — e-process value for H_0^w(q, p)
#   log_e_pq, log_e_qp — log-scale values (for plotting)
#
# References:
#   CR23  Choe & Ramdas (2023), Operations Research 72(4)
#   H21   Howard et al. (2021), Annals of Statistics 49(2)
# =============================================================================

#' Sub-exponential mixture e-process (Theorem 3, Choe & Ramdas 2023)
#'
#' Constructs two simultaneous one-sided e-processes for sequentially testing
#' whether forecaster 1 (p) outperforms forecaster 2 (q) or vice versa.
#'
#' The mixture e-process at time t is:
#' \deqn{E_t^{\mathrm{mix}} = m(S_t, \hat{V}_t)}
#' where \eqn{S_t = \sum_{i=1}^t \hat{\delta}_i},
#' \eqn{\hat{V}_t = \sum_{i=1}^t (\hat{\delta}_i - \gamma_i)^2},
#' and \eqn{m(s, v)} is the Gamma-Exponential mixture function (Proposition 3, CR23).
#'
#' @param scores1   Numeric vector. Scores S(p_t, y_t) for forecaster 1.
#' @param scores2   Numeric vector. Scores S(q_t, y_t) for forecaster 2.
#' @param alpha     Numeric in (0,1). Significance level. Rejection threshold
#'                  is 2/alpha for the two-sided test. Default: 0.05.
#' @param c         Numeric > 0. Sub-exponential scale. Must satisfy
#'                  |hat_delta_i| <= c/2 for all i.
#'                  For score differences in `[-(b-a), b-a]`: c = 2*(b-a).
#'                  Default: 2 (for Brier score differences in `[-1,1]`).
#' @param v_opt     Numeric > 0. Intrinsic time at which e-process grows
#'                  fastest. Default: 10 (recommended by CR23).
#' @param alpha_opt Numeric in (0,1). One-sided alpha used to compute rho.
#'                  Default: alpha/2 (matches comparecast two-sided convention).
#' @param gammas    Numeric vector or NULL. Predictable centering sequence.
#'                  If NULL, constructed as lagged running mean.
#' @param clip_max  Numeric. Maximum e-process value before clipping.
#'                  Default: 1e7 (matches Python comparecast).
#'
#' @return data.frame with columns t, e_pq, e_qp, log_e_pq, log_e_qp.
#'
#' @section Rejection rule:
#' At level `alpha`: reject \eqn{H_0^w(p, q)} (conclude `p` outperforms `q`)
#' when `e_pq >= 2/alpha`; reject \eqn{H_0^w(q, p)} (conclude `q` outperforms
#' `p`) when `e_qp >= 2/alpha`. Use `eprocess_rejections()` to extract the
#' first crossing time for each.
#'
#' @details
#'   VARIANCE PROCESS: The intrinsic time V_hat_t uses NO floor (unlike the
#'   EB CS). The GE mixture m(s, v) is well-defined at v=0 (returns 1 when
#'   s=0), so no floor is needed. Adding a floor would distort e-values.
#'
#'   SCALE CONVENTION: c is the sub-exponential scale parameter such that
#'   |hat_delta_i| <= c/2. This is the Theorems 2 & 3 convention from CR23.
#'   For Brier score differences in `[-1,1]`: c = 2.
#'   For Winkler scores (bounded above by 1): c = 2.
#'
#'   LOG-SPACE: E-process values are computed in log-space and clipped before
#'   exponentiating to avoid numerical overflow.
#'
#' @examples
#' scores1 <- c(-0.04, -0.09, -0.01, -0.16)
#' scores2 <- c(-0.09, -0.16, -0.04, -0.25)
#' ep <- eprocess(scores1, scores2, alpha = 0.05)
#' head(ep)
#'
#' @export
eprocess <- function(scores1, scores2,
                     alpha     = 0.05,
                     c         = 2,
                     v_opt     = 10,
                     alpha_opt = NULL,
                     gammas    = NULL,
                     clip_max  = 1e7) {

  stopifnot(
    length(scores1) == length(scores2),
    length(scores1) >= 1,
    alpha > 0, alpha < 1,
    c > 0,
    v_opt > 0,
    clip_max > 0
  )

  # Default alpha_opt: half of alpha for two-sided test
  if (is.null(alpha_opt)) alpha_opt <- alpha / 2

  xs <- scores1 - scores2
  T_ <- length(xs)

  # Predictable centering sequence
  if (is.null(gammas)) {
    gammas <- make_gammas(xs, lag = 1)
  } else {
    stopifnot(length(gammas) == T_)
  }

  # Tuning parameter rho from v_opt and alpha_opt
  rho <- rho_from_vopt(v_opt = v_opt, alpha = alpha_opt)

  # Compute shared variance process (sign-invariant)
  V_shared <- intrinsic_time(xs, gammas, floor = FALSE)
  S_pq     <- cumsum(xs)
  S_qp     <- -S_pq

  log_e_pq <- log_ge_mixture_from_sv(S_pq, V_shared, rho, c)
  log_e_qp <- log_ge_mixture_from_sv(S_qp, V_shared, rho, c)

  # Clip and exponentiate
  e_pq <- clip_eprocess(log_e_pq, clip_max = clip_max)
  e_qp <- clip_eprocess(log_e_qp, clip_max = clip_max)

  data.frame(
    t        = seq_len(T_),
    e_pq     = e_pq,
    e_qp     = e_qp,
    log_e_pq = pmin(log_e_pq, log(clip_max)),
    log_e_qp = pmin(log_e_qp, log(clip_max))
  )
}

#' Determine rejection times for an e-process output
#'
#' @param ep      data.frame. Output of eprocess().
#' @param alpha   Numeric. Significance level. Threshold is 2/alpha.
#'
#' @return Named list with elements:
#' * `threshold` — rejection threshold (`2 / alpha`).
#' * `tau_pq` — first `t` where `e_pq >= threshold` (`NA` if never crossed).
#' * `tau_qp` — first `t` where `e_qp >= threshold` (`NA` if never crossed).
#' * `reject_pq` — logical: was \eqn{H_0^w(p,q)} ever rejected?
#' * `reject_qp` — logical: was \eqn{H_0^w(q,p)} ever rejected?
#'
#' @examples
#' scores1 <- c(-0.04, -0.09, -0.01, -0.16)
#' scores2 <- c(-0.09, -0.16, -0.04, -0.25)
#' ep <- eprocess(scores1, scores2, alpha = 0.05)
#' eprocess_rejections(ep, alpha = 0.05)
#'
#' @export
eprocess_rejections <- function(ep, alpha = 0.05) {
  threshold <- 2 / alpha

  tau_pq <- which(ep$e_pq >= threshold)
  tau_qp <- which(ep$e_qp >= threshold)

  list(
    threshold  = threshold,
    tau_pq     = if (length(tau_pq) > 0) tau_pq[1] else NA_integer_,
    tau_qp     = if (length(tau_qp) > 0) tau_qp[1] else NA_integer_,
    reject_pq  = length(tau_pq) > 0,
    reject_qp  = length(tau_qp) > 0
  )
}

# =============================================================================
# tests/testthat/test_etests.R
#
# Tests:
#   A. Structural: correct output, initialisation at 1, monotonicity under alt
#   B. Type I error simulation: e-process under null never exceeds 2/alpha
#      in more than alpha fraction of paths (Markov inequality guarantee)
#   C. Power: e-process under alternative reaches threshold faster than chance
#
# B and C are gated behind SEQCOMP_RUN_SLOW_TESTS=true; run locally with:
#   SEQCOMP_RUN_SLOW_TESTS=true Rscript -e 'devtools::test()'
# =============================================================================

set.seed(42)
s1 <- brier_score(runif(300), rbinom(300, 1, 0.5))
s2 <- brier_score(runif(300), rbinom(300, 1, 0.5))

# ── A. STRUCTURAL TESTS ──────────────────────────────────────────────────────

test_that("eprocess returns a well-formed data.frame", {
  ep <- eprocess(s1, s2, alpha = 0.05, c = 2)

  expect_true(is.data.frame(ep))
  expect_true(all(c("t", "e_pq", "e_qp", "log_e_pq", "log_e_qp") %in% names(ep)))
  expect_equal(nrow(ep), 300)
  expect_equal(ep$t, 1:300)
})

test_that("eprocess e-values are positive and respect the clip ceiling", {
  ep <- eprocess(s1, s2, alpha = 0.05, c = 2)

  expect_true(all(ep$e_pq > 0) && all(ep$e_qp > 0))
  expect_true(all(ep$e_pq <= 1e7) && all(ep$e_qp <= 1e7))
})

test_that("log_e_pq matches log(e_pq)", {
  ep <- eprocess(s1, s2, alpha = 0.05, c = 2)
  expect_equal(ep$log_e_pq, log(ep$e_pq), tolerance = 1e-8)
})

test_that("e-process initialises near 1 for a near-zero first observation", {
  xs_small <- c(0.01, rep(0, 299))
  ep_small <- eprocess(xs_small, rep(0, 300), alpha = 0.05, c = 2)
  expect_equal(ep_small$e_pq[1], 1, tolerance = 0.1)
})

test_that("eprocess satisfies sign symmetry: e_pq(p,q) == e_qp(q,p)", {
  ep_fwd <- eprocess(s1, s2, alpha = 0.05, c = 2)
  ep_rev <- eprocess(s2, s1, alpha = 0.05, c = 2)

  expect_equal(ep_fwd$e_pq, ep_rev$e_qp, tolerance = 1e-8)
  expect_equal(ep_fwd$e_qp, ep_rev$e_pq, tolerance = 1e-8)
})

test_that("eprocess_rejections detects a clear alternative", {
  xs_alt <- rep(0.5, 300)   # constant positive differences: strong evidence for p
  ep_alt <- eprocess(xs_alt, rep(0, 300), alpha = 0.05, c = 2)
  rej    <- eprocess_rejections(ep_alt, alpha = 0.05)

  expect_equal(rej$threshold, 40)
  expect_true(rej$reject_pq)
  expect_false(is.na(rej$tau_pq))
  expect_false(rej$reject_qp)
})

test_that("eprocess_rejections does not reject on a null path", {
  xs_null_path <- rep(0, 300)
  ep_null_path <- eprocess(xs_null_path, rep(0, 300), alpha = 0.05, c = 2)
  rej_null     <- eprocess_rejections(ep_null_path, alpha = 0.05)

  expect_false(rej_null$reject_pq)
})

# ── B + C1. TYPE I ERROR AND POWER SIMULATION ────────────────────────────────
#
# Under H_0^w(p,q): Delta_t <= 0 for all t.
# By Markov's inequality: P(exists t: E_t >= 1/alpha_os) <= alpha_os = alpha/2
# Two-sided: P(max(e_pq, e_qp) >= 2/alpha) <= alpha.
#
# Power is then checked against the Type I error rate measured here, so both
# simulations live in one test (power_est > type1_pq mirrors the original
# script's direct comparison).

test_that("eprocess controls Type I error under the sharp null and has power exceeding it under the alternative", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow Type I error / power simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  alpha_sim <- 0.05
  threshold <- 2 / alpha_sim   # = 40

  # --- Type I error under the sharp null (Rademacher, mean 0) ---
  N     <- 1000
  T_sim <- 500
  n_reject_pq <- 0L
  n_reject_qp <- 0L
  set.seed(42)
  for (i in seq_len(N)) {
    xs <- sample(c(-1, 1), T_sim, replace = TRUE)
    ep_i <- eprocess(scores1 = xs, scores2 = rep(0, T_sim),
                     alpha = alpha_sim, c = 2)
    if (any(ep_i$e_pq >= threshold)) n_reject_pq <- n_reject_pq + 1L
    if (any(ep_i$e_qp >= threshold)) n_reject_qp <- n_reject_qp + 1L
  }
  type1_pq <- n_reject_pq / N
  type1_qp <- n_reject_qp / N

  expect_lte(type1_pq, alpha_sim)
  expect_lte(type1_qp, alpha_sim)

  # --- Power under the alternative (delta = 0.3) ---
  N_pow <- 500
  T_pow <- 1000
  delta <- 0.3
  n_power <- 0L
  set.seed(99)
  for (i in seq_len(N_pow)) {
    xs <- runif(T_pow, -1 + delta, 1 + delta)
    xs <- pmax(pmin(xs, 1), -1)
    ep_i <- eprocess(scores1 = xs, scores2 = rep(0, T_pow),
                     alpha = alpha_sim, c = 2)
    if (any(ep_i$e_pq >= threshold)) n_power <- n_power + 1L
  }
  power_est <- n_power / N_pow

  expect_gt(power_est, 0.5)
  expect_gt(power_est, type1_pq)  # test is informative: power beats the null rejection rate
})

# ── C2. POWER CURVE ACROSS DELTA VALUES ──────────────────────────────────────

test_that("eprocess power increases with the effect size delta", {
  skip_on_cran()
  skip_if(Sys.getenv("SEQCOMP_RUN_SLOW_TESTS") != "true",
          "slow power curve simulation; set SEQCOMP_RUN_SLOW_TESTS=true to run")

  alpha_sim <- 0.05
  threshold <- 2 / alpha_sim
  deltas    <- c(0.1, 0.2, 0.3, 0.5)
  N_curve   <- 200
  T_curve   <- 1000

  power_curve <- numeric(length(deltas))
  set.seed(77)
  for (j in seq_along(deltas)) {
    d <- deltas[j]
    n_rej <- 0L
    for (i in seq_len(N_curve)) {
      xs   <- pmax(pmin(runif(T_curve, -1 + d, 1 + d), 1), -1)
      ep_i <- eprocess(xs, rep(0, T_curve), alpha = alpha_sim, c = 2)
      if (any(ep_i$e_pq >= threshold)) n_rej <- n_rej + 1L
    }
    power_curve[j] <- n_rej / N_curve
  }

  expect_true(all(power_curve >= 0 & power_curve <= 1))
  expect_gt(power_curve[length(power_curve)], power_curve[1])
})

---
title: "Getting started with seqcomp"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with seqcomp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)
```

```{r setup}
library(seqcomp)
```

## Overview

`seqcomp` compares two sequential probabilistic forecasters.

The basic question is:

> As outcomes arrive over time, is one forecaster consistently scoring better
> than the other?

The package follows the convention that all scores are positively oriented:
larger scores are better.

For two forecasters `p` and `q`, the pointwise score difference is

\[
\hat{\delta}_t = S(p_t, y_t) - S(q_t, y_t).
\]

Positive values favour forecaster `p`; negative values favour forecaster `q`.

## A simple binary forecasting example

Suppose we observe a sequence of binary outcomes.

```{r}
set.seed(1)

n <- 300
y <- rbinom(n, size = 1, prob = 0.55)
```

Now create two forecasters.

Forecaster `p` is mildly informative: it tends to assign higher probability
when the event occurs and lower probability when it does not.

Forecaster `q` is less informative and stays closer to 0.5.

```{r}
p <- ifelse(y == 1, 0.62, 0.38)
q <- rep(0.50, n)

head(data.frame(y = y, p = p, q = q))
```

This toy example is deliberately simple. In a real application, `p` and `q`
would come from two forecasting models, analysts, or institutions.

## Compare the forecasts

The easiest workflow is to use `compare_forecasts()`.

```{r}
cmp <- compare_forecasts(
  p = p,
  q = q,
  y = y,
  scoring_rule = "brier"
)

head(cmp)
```

The output contains one row per time point.

The most important columns are:

- `score_p`: the score of forecaster `p`,
- `score_q`: the score of forecaster `q`,
- `delta`: the score difference `score_p - score_q`,
- `estimate`: the running mean score difference,
- `lower` and `upper`: the confidence sequence,
- `e_pq` and `e_qp`: the two one-sided e-processes.

## Interpreting the confidence sequence

The confidence sequence tracks the running average score advantage.

```{r}
plot(
  cmp$t, cmp$estimate,
  type = "l",
  ylim = range(c(cmp$lower, cmp$upper, 0), finite = TRUE),
  xlab = "Time",
  ylab = "Mean score difference",
  main = "Sequential comparison using the Brier score"
)
lines(cmp$t, cmp$lower, lty = 2)
lines(cmp$t, cmp$upper, lty = 2)
abline(h = 0, col = "gray50")
```

If the whole interval lies above zero, the data favour `p`.

If the whole interval lies below zero, the data favour `q`.

If the interval still contains zero, the comparison is not yet decisive at the
chosen level.

## Interpreting the e-process

The e-process is an evidence process. Larger values mean stronger evidence
against the corresponding null hypothesis.

For a two-sided comparison at level `alpha`, the rejection threshold used by
`eprocess()` is `2 / alpha`.

```{r}
alpha <- 0.05
threshold <- 2 / alpha

threshold
```

The column `e_pq` gives evidence that `p` outperforms `q`.

The column `e_qp` gives evidence that `q` outperforms `p`.

```{r}
plot(
  cmp$t, cmp$e_pq,
  type = "l",
  log = "y",
  xlab = "Time",
  ylab = "e-process value",
  main = "Evidence that p outperforms q"
)
abline(h = threshold, lty = 2, col = "gray50")
```

We can summarize rejection times with `eprocess_rejections()`.

```{r}
eprocess_rejections(cmp, alpha = alpha)
```

## Using lower-level functions directly

The wrapper is convenient, but the package also exposes the individual pieces.

First compute the scores.

```{r}
score_p <- brier_score(p, y)
score_q <- brier_score(q, y)

head(score_p)
head(score_q)
```

Then construct a confidence sequence.

```{r}
cs <- cs_bernstein(
  scores1 = score_p,
  scores2 = score_q,
  alpha = 0.05,
  c = 2
)

head(cs)
```

And construct an e-process.

```{r}
ep <- eprocess(
  scores1 = score_p,
  scores2 = score_q,
  alpha = 0.05,
  c = 2
)

head(ep)
```

This gives the same core information as `compare_forecasts()`, but with more
manual control.

## Choosing a scoring rule

For binary probability forecasts, `"brier"` is the safest starting point.

```{r}
cmp_brier <- compare_forecasts(p, q, y, scoring_rule = "brier")
tail(cmp_brier)
```

The spherical score is also bounded and can be used with finite-sample
confidence sequences and e-processes.

```{r}
cmp_spherical <- compare_forecasts(p, q, y, scoring_rule = "spherical")
tail(cmp_spherical)
```

The logarithmic score is unbounded. Therefore `compare_forecasts()` uses an
asymptotic confidence sequence by default and does not compute an e-process.

```{r}
cmp_log <- compare_forecasts(
  p = p,
  q = q,
  y = y,
  scoring_rule = "log",
  compute_e = FALSE
)

tail(cmp_log)
```

For binary log-score comparisons where the Winkler construction is appropriate,
use `winkler_compare()`.

```{r}
wcmp <- winkler_compare(p, q, y)

names(wcmp)
```

## Practical guidance

Use `compare_forecasts()` for the common workflow.

Use the lower-level functions when you want more control over the scoring rule,
confidence sequence, e-process, lag handling, or predictable bounds.

For bounded binary or categorical scores such as Brier and spherical scores,
finite-sample confidence sequences and e-processes are available.

For unbounded scores such as the log score, QLIKE, tick loss, and CRPS, be more
careful. Use asymptotic confidence sequences, Winkler scores where applicable,
or problem-specific predictable bounds.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confidence_sequences.R
\name{cs_hoeffding}
\alias{cs_hoeffding}
\title{Hoeffding-style confidence sequence (Theorem 1, Choe & Ramdas 2023)}
\usage{
cs_hoeffding(
  scores1,
  scores2,
  alpha = 0.05,
  c = 1,
  v_opt = 10,
  boundary = "mixture"
)
}
\arguments{
\item{scores1}{Numeric vector. Scores S(p_t, y_t) for forecaster 1.}

\item{scores2}{Numeric vector. Scores S(q_t, y_t) for forecaster 2.}

\item{alpha}{Numeric in (0,1). Significance level. The CS has coverage
1 - alpha uniformly over all t. Default: 0.05.}

\item{c}{Numeric > 0. Sub-Gaussian scale. The process must satisfy
|hat_delta_i| <= c for all i. For scores in \verb{[a,b]}, the
difference is in \verb{[a-b, b-a]}, so c = b - a.
Default: 1 (appropriate for Brier score differences in \verb{[-1,1]}).}

\item{v_opt}{Numeric > 0. Intrinsic time at which the CS is tightest.
Default: 10 (recommended by CR23).}

\item{boundary}{Character. "mixture" (default, recommended) or "stitching".}
}
\value{
data.frame with columns t, estimate, lower, upper.
}
\description{
Constructs a time-uniform confidence sequence for the mean score difference
\eqn{\Delta_t = \frac{1}{t} \sum_{i=1}^t E[\hat{\delta_i} \mid \mathcal{F}_{i-1}]}.
}
\section{Assumption}{

Requires \code{hat_delta_i} to be c-sub-Gaussian given \eqn{\mathcal{F}_{i-1}},
i.e. \verb{|hat_delta_i| <= c} for all \code{i}.
}

\section{Boundary}{

\deqn{C_t^H = \hat\Delta_t \pm u^{CM}_{\alpha/2}(c^2 t; \rho) / t}
where \eqn{u^{CM}} is the normal mixture boundary and \eqn{c^2 t} is the
intrinsic time for a c-sub-Gaussian process with deterministic variance
proxy. The intrinsic time for Theorem 1 is \code{v_t = c^2 * t}, not \code{v_t = t}:
the CM boundary implicitly assumes 1-sub-Gaussian inputs, so the \code{c^2}
scaling must be applied explicitly. This matches the H21 convention,
where the boundary absorbs the sub-Gaussian parameter via the variance
process definition.

Relation to Python comparecast: Python uses \code{v_t = sigma * t} where
\code{sigma = (hi - lo)/2 = c}. This is equivalent to our \code{c^2 * t} only when
\code{c = 1}. For \code{c != 1} the parametrisations differ; we follow the paper.
}

\section{Output}{

Returns a \code{data.frame} with one row per \code{t} and columns \code{t}, \code{estimate}
(the running mean \code{hat_Delta_t}), \code{lower}, and \code{upper}, with coverage
guarantee \eqn{P(\forall t \geq 1 : \Delta_t \in [\text{lower}_t,
\text{upper}_t]) \geq 1 - \alpha}.
}

\examples{
scores1 <- c(-0.04, -0.09, -0.01, -0.16)
scores2 <- c(-0.09, -0.16, -0.04, -0.25)
cs_hoeffding(scores1, scores2, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqcomp-package.R
\name{seqcomp-package}
\alias{seqcomp-package}
\alias{seqcomp}
\title{seqcomp: Sequential Comparison of Probabilistic Forecasts}
\description{
\code{seqcomp} provides tools for comparing probabilistic forecasters
sequentially, following the anytime-valid framework of Choe and Ramdas
(2024).
}
\details{
The package is built around the score difference

\deqn{\hat{\delta}_t = S(p_t, y_t) - S(q_t, y_t),}

where scores are positively oriented, so larger values are better. Positive
score differences favour forecaster \code{p}; negative score differences favour
forecaster \code{q}.
}
\section{Main workflow}{

For most applications, start with \code{\link[=compare_forecasts]{compare_forecasts()}}. It computes
pointwise scores, running mean score differences, confidence sequences, and
e-processes in one call.
}

\section{Scoring rules}{

The package includes positively oriented scoring rules such as
\code{\link[=brier_score]{brier_score()}}, \code{\link[=log_score]{log_score()}}, \code{\link[=spherical_score]{spherical_score()}}, \code{\link[=tick_loss]{tick_loss()}},
\code{\link[=qlike_score]{qlike_score()}}, \code{\link[=winkler_score]{winkler_score()}}, \code{\link[=crps_normal]{crps_normal()}}, \code{\link[=crps_empirical]{crps_empirical()}},
and \code{\link[=crps_std]{crps_std()}}.
}

\section{Confidence sequences}{

Use \code{\link[=cs_hoeffding]{cs_hoeffding()}} for Hoeffding-style confidence sequences,
\code{\link[=cs_bernstein]{cs_bernstein()}} for empirical Bernstein confidence sequences, and
\code{\link[=cs_asymptotic]{cs_asymptotic()}} for asymptotic confidence sequences when finite-sample
boundedness is not available.
}

\section{E-processes}{

Use \code{\link[=eprocess]{eprocess()}} for the main sub-exponential mixture e-process and
\code{\link[=eprocess_rejections]{eprocess_rejections()}} to extract first rejection times. For multi-step
forecasts, see \code{\link[=eprocess_lag]{eprocess_lag()}}. For predictable time-varying bounds, see
\code{\link[=eprocess_predictable]{eprocess_predictable()}}.
}

\section{Winkler scores}{

For binary probability forecasts with unbounded base scores, use
\code{\link[=winkler_score]{winkler_score()}}, \code{\link[=winkler_cs]{winkler_cs()}}, \code{\link[=winkler_etest]{winkler_etest()}}, or
\code{\link[=winkler_compare]{winkler_compare()}}.
}

\references{
Choe, Y. J. and Ramdas, A. (2024). Comparing Sequential Forecasters.
Operations Research, 72(4), 1368-1387.

Howard, S. R., Ramdas, A., McAuliffe, J. and Sekhon, J. (2021).
Time-uniform, nonparametric, nonasymptotic confidence sequences.
The Annals of Statistics, 49(2).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{score_bounds}
\alias{score_bounds}
\title{Score difference bounds for a named scoring rule}
\usage{
score_bounds(scoring_rule)
}
\arguments{
\item{scoring_rule}{Character. One of:
\itemize{
\item \code{"brier"}, \code{"spherical"} — bounded, exact finite-sample \code{c}.
\item \code{"winkler"} — descriptive helper for the one-sided CS on the log score.
\item \code{"tick"} — unbounded; returns \code{NULL} with guidance.
\item \code{"crps"}, \code{"crps_normal"}, \code{"crps_empirical"}, \code{"crps_std"} — unbounded;
returns \code{NULL} with guidance.
\item \code{"log"}, \code{"qlike"} — unbounded; returns \code{NULL} with guidance.
}}
}
\value{
Named list with elements lo, hi, c_thm1, c_thm23 for bounded
rules, or NULL for unbounded rules (with an informative message).
}
\description{
Returns lo, hi and the derived scale parameters c_thm1, c_thm23 for the
score difference process
hat_delta_t = S(p, y) - S(q, y),
in those cases where a genuine, theorem-valid bound is available.
}
\details{
Convention (utils.R::score_diff_scales):
c_thm1  = (hi - lo) / 2   # Theorem 1:      |delta_i| <= c
c_thm23 = hi - lo        # Theorems 2 & 3: |delta_i| <= c/2
}
\section{Per-rule notes}{

\itemize{
\item \strong{Brier / Spherical} — individual scores lie in \verb{[-1, 0]} (Brier) or
\verb{[0, 1]} (Spherical), so score differences lie in \verb{[-1, 1]} either way.
This bound is exact and yields finite-sample anytime-valid CS via
Hoeffding/Bernstein.
\item \strong{Winkler} — bounded above by 1; the lower bound is problem-dependent,
so \code{lo = -Inf} and only \code{hi = 1} is used, as a descriptive helper for
the one-sided CS wrapper \code{winkler_cs()}. Not intended for generic
Hoeffding/Bernstein use (Theorem 1 requires a finite symmetric interval).
\item \strong{Tick loss} — unbounded on general financial returns. Any bound
derived from an empirical data range is ex-post and not
filtration-respecting, so it cannot justify finite-sample anytime
validity. Use \code{cs_asymptotic()} for tick comparisons.
\item \strong{CRPS} (normal, t, empirical) — unbounded, since both the predictive
distributions and the realised outcomes are unbounded. A historical
data range is again an ex-post surrogate and does not provide a
theorem-valid \code{c} for Hoeffding/Bernstein. Use \code{cs_asymptotic()}, or
supply genuine ex ante bounds in problem-specific code if available.
\item \strong{Log / QLIKE} — both unbounded. For binary log-score comparisons, use
\code{winkler_score()} + \code{winkler_cs()} when the Winkler construction is
appropriate. For categorical log-score, QLIKE, and other unbounded
score differences, use \code{cs_asymptotic()}, or \code{eprocess_predictable()}
only with genuine ex ante predictable bounds.
}
}

\examples{
score_bounds("brier")
score_bounds("winkler")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{cm_boundary}
\alias{cm_boundary}
\title{Normal mixture (CM) boundary}
\usage{
cm_boundary(v, alpha, rho)
}
\arguments{
\item{v}{Numeric vector >= 0. Intrinsic time values (V_t or t).}

\item{alpha}{Numeric in (0,1). ONE-SIDED significance level.}

\item{rho}{Numeric > 0. Tuning parameter. Obtain via rho_from_vopt().}
}
\value{
Numeric vector of boundary values (cumulative-sum scale, before /t).
}
\description{
Normal mixture (CM) boundary
}
\examples{
  rho <- rho_from_vopt(v_opt = 10, alpha = 0.025)
  u   <- cm_boundary(v = 1:500, alpha = 0.025, rho = rho)

}
\seealso{
\code{\link[=rho_from_vopt]{rho_from_vopt()}} to compute \code{rho} from a target \code{v_opt}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/winkler.R
\name{winkler_compare}
\alias{winkler_compare}
\title{Full Winkler comparison pipeline (Proposition 4)}
\usage{
winkler_compare(
  p,
  q,
  y,
  alpha = 0.05,
  base_score = log_score,
  v_opt = 10,
  lower_bound = NULL
)
}
\arguments{
\item{p}{Numeric vector in (0,1).}

\item{q}{Numeric vector in (0,1).}

\item{y}{Numeric vector containing only 0 and 1. Binary outcomes.}

\item{alpha}{Numeric in (0,1). Default: 0.05.}

\item{base_score}{Function. Default: log_score.}

\item{v_opt}{Numeric > 0. Default: 10.}

\item{lower_bound}{Numeric or NULL. See winkler_cs().}
}
\value{
Named list with elements:
\itemize{
\item \code{winkler_scores} — raw Winkler score vector.
\item \code{cs} — \code{data.frame} from \code{winkler_cs()}.
\item \code{etest_p_worse} — one-sided e-process testing whether \code{p} is worse than \code{q}.
\item \code{etest_q_worse} — one-sided e-process testing whether \code{q} is worse than \code{p}.
\item \code{rejections} — list of one-sided rejection summaries.
}
}
\description{
Convenience wrapper that computes Winkler scores, one-sided CS, and
e-process in a single call.
}
\examples{
p <- c(0.7, 0.6, 0.8, 0.65)
q <- c(0.5, 0.7, 0.6, 0.55)
y <- c(1, 1, 0, 1)
winkler_compare(p, q, y, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictable_bounds.R
\name{eprocess_predictable}
\alias{eprocess_predictable}
\title{Fixed-lambda e-process with predictable bounds (Proposition 7)}
\usage{
eprocess_predictable(
  scores1,
  scores2,
  c_seq,
  lambda = NULL,
  alpha = 0.05,
  gammas = NULL,
  clip_max = 1e+07,
  strict = FALSE
)
}
\arguments{
\item{scores1}{Numeric vector. Scores for forecaster 1.}

\item{scores2}{Numeric vector. Scores for forecaster 2.}

\item{c_seq}{Numeric vector. Predictable bound sequence (c_i), same
length as scores1. Must satisfy \verb{|scores1[i]-scores2[i]| <= c_i/2}
and c_i > 0 for all i.}

\item{lambda}{Numeric in \verb{[0, 1/c_0)}. Betting parameter. Must be
strictly less than 1/c_0 where c_0 = max(c_seq).
If NULL, uses the recommended default lambda = 0.5/c_0.}

\item{alpha}{Numeric in (0,1). Significance level for rejection rule.
Default: 0.05. Not used in computation, only for API
consistency. Pass the same value to \code{\link[=predictable_rejections]{predictable_rejections()}}
when evaluating rejection.}

\item{gammas}{Numeric vector or NULL. Predictable centering sequence.
If NULL, constructed as lagged running mean.}

\item{clip_max}{Numeric. Maximum e-process value. Default: 1e7.}

\item{strict}{Logical. If TRUE, any violation of the bound condition at
any time point will stop execution with an error. If FALSE
(default), a warning is issued but the e-process is still computed.
Note that violations invalidate the e-process guarantee,
so strict = TRUE is recommended for formal inference.}
}
\value{
data.frame with columns:
t, e_pq, e_qp, log_e_pq, log_e_qp, c_seq, lambda_used
}
\description{
Constructs a valid e-process when score difference bounds vary over time
but are predictable (known at time i-1 before observing hat_delta_i).
}
\details{
The e-process is computed as:
\deqn{\log E_t(\lambda) = \sum_{i=1}^t \Bigl[\lambda\,\hat{\delta}_i
        - \psi_{E,c_i}(\lambda)\,(\hat{\delta}_i - \gamma_i)^2\Bigr]}

where
\deqn{\psi_{E,c}(\lambda) = \frac{-\log(1 - c\lambda) - c\lambda}{c^2}}
is evaluated at each step with the current \eqn{c_i}.

LAMBDA CHOICE: lambda = 0.5/c_0 is a conservative default that stays
well within the valid domain \verb{[0, 1/c_0)}. For better power, lambda can
be tuned to the expected signal size, but must never reach 1/c_0.

VALIDITY CHECK: The function verifies |hat_delta_i| <= c_i/2 at each
step and warns if violated. Violations invalidate the e-process guarantee.
}
\section{Predictability}{

The bound sequence \code{c_seq} (and the centering sequence \code{gammas}) must be
predictable: \code{c_i} is fixed and known at time \code{i - 1}, before
\code{scores1[i]}/\code{scores2[i]} (and hence \code{hat_delta_i}) are observed —
formally, \eqn{c_i} is \eqn{\mathcal{F}_{i-1}}-measurable. A bound chosen
after seeing \code{hat_delta_i} (e.g. derived from the realised data range)
invalidates the e-process guarantee, even if it numerically satisfies
\verb{|hat_delta_i| <= c_i/2}.
}

\examples{
scores1 <- c(0.10, 0.20, 0.15, 0.25)
scores2 <- c(0.05, 0.10, 0.10, 0.20)
c_seq <- rep(1, length(scores1))
ep <- eprocess_predictable(scores1, scores2, c_seq = c_seq)
head(ep)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lag_handling.R
\name{unroll_stream}
\alias{unroll_stream}
\title{Unroll a stream-wise quantity back to the original time scale}
\usage{
unroll_stream(stream_vals, k, h, T_)
}
\arguments{
\item{stream_vals}{Numeric vector. Values for stream k (length ~ T/h).}

\item{k}{Integer. Stream index (1-based).}

\item{h}{Integer. Lag.}

\item{T_}{Integer. Total original sequence length.}
}
\value{
Numeric vector of length T_.
}
\description{
After computing a per-stream cumulative quantity (e.g. e-process values),
restores them to the original length \code{T} by zero-padding the first
\code{k - 1} positions, repeating each stream value \code{h} times, then truncating
to length \code{T}.
}
\section{Alignment only, not theoretical updating}{

For lagged forecasts (\code{h >= 2}), the returned series is aligned to the
evaluated score-difference index after stream splitting. It should
\strong{not} be interpreted as a process that updates at the original
forecast-issuance time. The unrolled process is for visualization and
alignment only; the theoretical validity argument relies strictly on the
streamwise sub-filtrations, not on this unrolled representation.
}

\examples{
unroll_stream(c(1, 2, 3), k = 2, h = 2, T_ = 6)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/etests.R
\name{eprocess}
\alias{eprocess}
\title{Sub-exponential mixture e-process (Theorem 3, Choe & Ramdas 2023)}
\usage{
eprocess(
  scores1,
  scores2,
  alpha = 0.05,
  c = 2,
  v_opt = 10,
  alpha_opt = NULL,
  gammas = NULL,
  clip_max = 1e+07
)
}
\arguments{
\item{scores1}{Numeric vector. Scores S(p_t, y_t) for forecaster 1.}

\item{scores2}{Numeric vector. Scores S(q_t, y_t) for forecaster 2.}

\item{alpha}{Numeric in (0,1). Significance level. Rejection threshold
is 2/alpha for the two-sided test. Default: 0.05.}

\item{c}{Numeric > 0. Sub-exponential scale. Must satisfy
|hat_delta_i| <= c/2 for all i.
For score differences in \verb{[-(b-a), b-a]}: c = 2*(b-a).
Default: 2 (for Brier score differences in \verb{[-1,1]}).}

\item{v_opt}{Numeric > 0. Intrinsic time at which e-process grows
fastest. Default: 10 (recommended by CR23).}

\item{alpha_opt}{Numeric in (0,1). One-sided alpha used to compute rho.
Default: alpha/2 (matches comparecast two-sided convention).}

\item{gammas}{Numeric vector or NULL. Predictable centering sequence.
If NULL, constructed as lagged running mean.}

\item{clip_max}{Numeric. Maximum e-process value before clipping.
Default: 1e7 (matches Python comparecast).}
}
\value{
data.frame with columns t, e_pq, e_qp, log_e_pq, log_e_qp.
}
\description{
Constructs two simultaneous one-sided e-processes for sequentially testing
whether forecaster 1 (p) outperforms forecaster 2 (q) or vice versa.
}
\details{
The mixture e-process at time t is:
\deqn{E_t^{\mathrm{mix}} = m(S_t, \hat{V}_t)}
where \eqn{S_t = \sum_{i=1}^t \hat{\delta}_i},
\eqn{\hat{V}_t = \sum_{i=1}^t (\hat{\delta}_i - \gamma_i)^2},
and \eqn{m(s, v)} is the Gamma-Exponential mixture function (Proposition 3, CR23).

VARIANCE PROCESS: The intrinsic time V_hat_t uses NO floor (unlike the
EB CS). The GE mixture m(s, v) is well-defined at v=0 (returns 1 when
s=0), so no floor is needed. Adding a floor would distort e-values.

SCALE CONVENTION: c is the sub-exponential scale parameter such that
|hat_delta_i| <= c/2. This is the Theorems 2 & 3 convention from CR23.
For Brier score differences in \verb{[-1,1]}: c = 2.
For Winkler scores (bounded above by 1): c = 2.

LOG-SPACE: E-process values are computed in log-space and clipped before
exponentiating to avoid numerical overflow.
}
\section{Rejection rule}{

At level \code{alpha}: reject \eqn{H_0^w(p, q)} (conclude \code{p} outperforms \code{q})
when \code{e_pq >= 2/alpha}; reject \eqn{H_0^w(q, p)} (conclude \code{q} outperforms
\code{p}) when \code{e_qp >= 2/alpha}. Use \code{eprocess_rejections()} to extract the
first crossing time for each.
}

\examples{
scores1 <- c(-0.04, -0.09, -0.01, -0.16)
scores2 <- c(-0.09, -0.16, -0.04, -0.25)
ep <- eprocess(scores1, scores2, alpha = 0.05)
head(ep)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{ge_boundary}
\alias{ge_boundary}
\title{Gamma-exponential mixture boundary}
\usage{
ge_boundary(v, alpha, rho, c, s_lo = -10, s_hi = 500)
}
\arguments{
\item{v}{Numeric vector >= 0. Intrinsic time values.}

\item{alpha}{Numeric in (0,1). ONE-SIDED significance level.}

\item{rho}{Numeric > 0. From rho_from_vopt().}

\item{c}{Numeric > 0. Sub-exponential scale.}

\item{s_lo}{Numeric. Lower search bound for uniroot. Default: -10.}

\item{s_hi}{Numeric. Upper search bound for uniroot. Default: 500.}
}
\value{
Numeric vector of boundary values (cumulative-sum scale).
}
\description{
Gamma-exponential mixture boundary
}
\details{
Computes \eqn{u_{GE}(v; \alpha, \rho, c) = \sup\{s : m(s,v) < 1/\alpha\}} by
solving \code{m(s, v) = 1/alpha} numerically for \code{s} via \code{\link[=uniroot]{uniroot()}}, separately
for each \code{v_i}.

Root-finding fallback: the search starts in \verb{[s_lo, s_hi]}; if
\code{m(s_hi, v_i)} has not yet crossed the target, \code{s_hi} is doubled once and
retried. If it still fails, a warning is issued and \code{s_hi} is returned as
a conservative fallback value. Increase \code{s_hi} directly if this warning
appears often (e.g. at large \code{v} or small \code{alpha}); increase \code{abs(s_lo)}
if no root is found at small \code{v}.

Computed elementwise; can be slow for long vectors — consider caching
boundary values when the same \verb{(alpha, rho, c)} are reused.
}
\examples{
rho <- rho_from_vopt(v_opt = 10, alpha = 0.025)
ge_boundary(v = 1:3, alpha = 0.025, rho = rho, c = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lag_handling.R
\name{eprocess_lag}
\alias{eprocess_lag}
\title{Lag-h e-process for sequential forecast comparison (Propositions 5 & 6)}
\usage{
eprocess_lag(
  scores1,
  scores2,
  h = 1,
  alpha = 0.05,
  c = 2,
  v_opt = 10,
  null = "pw",
  calibrate = TRUE,
  cal_strategy = "mixture"
)
}
\arguments{
\item{scores1}{Numeric vector. Scores for forecaster 1.}

\item{scores2}{Numeric vector. Scores for forecaster 2.}

\item{h}{Integer >= 1. Forecast lag.
For h=1, reduces to standard eprocess() — no splitting.}

\item{alpha}{Numeric in (0,1). Significance level. Default: 0.05.}

\item{c}{Numeric > 0. Sub-exponential scale. Default: 2.}

\item{v_opt}{Numeric > 0. Default: 10.}

\item{null}{Character. Null hypothesis type:
\itemize{
\item \code{"pw"} — period-wise weak null (average combination).
\item \code{"w"} — standard weak null (minimum combination).
}}

\item{calibrate}{Logical. Apply p-to-e calibration. Default: TRUE.}

\item{cal_strategy}{Character. "mixture" (default) or "simple".}
}
\value{
data.frame with columns t, e_pq, e_qp, log_e_pq, log_e_qp.
}
\description{
For h-step-ahead forecasts, constructs an anytime-valid e-process by
stream splitting and combining h individual e-processes.
}
\details{
For h = 1: calls eprocess() directly and returns its output unchanged.

For h >= 2:
1. Split xs into h streams
2. Compute e-process on each stream independently
3. Combine using the appropriate null rule
4. Convert to p-process, combine, calibrate back to e-process
5. Unroll to original time scale

The period-wise ("pw") null is less conservative than the standard ("w")
null but tests a different (weaker) hypothesis. See CR23 Section 4.4.
}
\examples{
scores1 <- c(-0.04, -0.09, -0.01, -0.16, -0.04, -0.09)
scores2 <- c(-0.09, -0.16, -0.04, -0.25, -0.09, -0.16)
ep <- eprocess_lag(scores1, scores2, h = 2, alpha = 0.05)
head(ep)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{crps_std}
\alias{crps_std}
\title{Negated CRPS for Student-t predictive distributions}
\usage{
crps_std(mu, sigma, dof, x)
}
\arguments{
\item{mu}{Numeric vector. Location parameters (conditional means).}

\item{sigma}{Numeric vector. Scale parameters (conditional SDs, > 0).}

\item{dof}{Numeric vector or scalar. Degrees of freedom (> 2).
May be scalar if constant across all observations (e.g.
estimated once per rolling window).}

\item{x}{Numeric vector. Realised observations.}
}
\value{
Numeric vector of CRPS values in \verb{(-Inf, 0]} (negated loss).
}
\description{
Wrapper over \code{scoringRules::crps_t}. Positively oriented:
higher = better. The dof > 2 constraint ensures finite variance,
which is required for the CRPS to be well-defined for the t-distribution.
}
\details{
Calls \code{scoringRules::crps_t(y = x, df = dof, location = mu,
  scale = sigma)} and negates. Use for GARCH(1,1)-std forecasts where
dof is the estimated degrees-of-freedom parameter from ugarchroll.
}
\examples{
if (requireNamespace("scoringRules", quietly = TRUE)) {
  crps_std(mu = c(0, 1), sigma = c(1, 2), dof = 5, x = c(0.2, 1.3))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{crps_normal}
\alias{crps_normal}
\title{Negated CRPS for normal predictive distributions}
\usage{
crps_normal(mu, sigma, x)
}
\arguments{
\item{mu}{Numeric vector. Location parameters (conditional means).}

\item{sigma}{Numeric vector. Scale parameters (conditional SDs, > 0).}

\item{x}{Numeric vector. Realised observations.}
}
\value{
Numeric vector of CRPS values in \verb{(-Inf, 0]} (negated loss).
}
\description{
Computes the Continuous Ranked Probability Score for a normal predictive
distribution using \code{scoringRules::crps_norm()} and negates it so that higher
values are better.
}
\details{
Calls \code{scoringRules::crps_norm(y = x, mean = mu, sd = sigma)} and
negates. Use for GARCH(1,1)-norm forecasts where mu is the conditional
mean and sigma is the conditional standard deviation.
}
\examples{
if (requireNamespace("scoringRules", quietly = TRUE)) {
  crps_normal(mu = c(0, 1), sigma = c(1, 2), x = c(0.2, 1.3))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lag_handling.R
\name{calibrate_p_to_e}
\alias{calibrate_p_to_e}
\title{P-to-e calibrator}
\usage{
calibrate_p_to_e(p, strategy = "mixture", eps = 1e-16)
}
\arguments{
\item{p}{Numeric vector of p-values in (0, 1].}

\item{strategy}{Character. \code{"mixture"} (default, from Vovk & Wang 2021) or
\code{"simple"}. See Details for the formulas.}

\item{eps}{Numeric. Numerical guard for log(0). Default: 1e-16.}
}
\value{
Numeric vector of e-values >= 0.
}
\description{
Converts anytime-valid p-values to e-values using the mixture or simple
calibrator, as used in CR23 Section 4.4.
}
\details{
Mixture calibrator (default, matches Python comparecast behaviour):
\deqn{f(p) = \frac{1 - p + p\log(p)}{p\,(\log p)^2}}

Simple calibrator (\code{strategy = "simple"}):
\deqn{f(p) = \frac{1}{2\sqrt{p}}}
}
\examples{
p <- c(0.5, 0.1, 0.01)
calibrate_p_to_e(p)
calibrate_p_to_e(p, strategy = "simple")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictable_bounds.R
\name{predictable_rejections}
\alias{predictable_rejections}
\title{Summarise predictable bounds e-process}
\usage{
predictable_rejections(ep, alpha = 0.05)
}
\arguments{
\item{ep}{data.frame. Output of eprocess_predictable().}

\item{alpha}{Numeric. Significance level.}
}
\value{
Named list matching the \code{eprocess_rejections()} format
(\code{threshold}, \code{tau_pq}, \code{tau_qp}, \code{reject_pq}, \code{reject_qp}), plus:
\itemize{
\item \code{c_range} — range of \code{c_seq} used.
\item \code{lambda} — \code{lambda} value used.
}
}
\description{
Summarise predictable bounds e-process
}
\examples{
scores1 <- c(0.10, 0.20, 0.15, 0.25)
scores2 <- c(0.05, 0.10, 0.10, 0.20)
c_seq <- rep(1, length(scores1))
ep <- eprocess_predictable(scores1, scores2, c_seq = c_seq)
predictable_rejections(ep, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rho_from_vopt}
\alias{rho_from_vopt}
\title{Convert optimal intrinsic time to rho}
\usage{
rho_from_vopt(v_opt = 10, alpha = 0.025)
}
\arguments{
\item{v_opt}{Numeric > 0. Intrinsic time at which the boundary is tightest.
Recommended default from CR23: 10.}

\item{alpha}{Numeric in (0,1). Significance level (one-sided).
For a two-sided boundary at level alpha, pass alpha/2 here.}
}
\value{
Numeric > 0. The rho tuning parameter.
}
\description{
Maps the user-specified intrinsic time \code{v_opt} (the point at which a
boundary is tightest) to the \code{rho} tuning parameter, via the Lambert W
formula of Howard et al. (2021), Proposition 3 (paper-exact).
}
\details{
\deqn{\rho = \frac{v_{opt}}{-W_{-1}(-\alpha^2 / e) - 1}}
The lower branch \eqn{W_{-1}} is defined for \code{x} in \verb{[-1/e, 0)} and returns
values \verb{<= -1}. For \code{alpha} in \verb{(0, 1)}, \code{-alpha^2/e} is always in
\verb{(-1/e, 0)}, so the branch is well-defined.
}
\examples{
rho_from_vopt(v_opt = 10, alpha = 0.025)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confidence_sequences.R
\name{cs_bernstein}
\alias{cs_bernstein}
\title{Empirical Bernstein confidence sequence (Theorem 2, Choe & Ramdas 2023)}
\usage{
cs_bernstein(
  scores1,
  scores2,
  alpha = 0.05,
  c = 2,
  v_opt = 10,
  boundary = "mixture",
  gammas = NULL,
  lcb_only = FALSE,
  ucb_only = FALSE
)
}
\arguments{
\item{scores1}{Numeric vector. Scores for forecaster 1.}

\item{scores2}{Numeric vector. Scores for forecaster 2.}

\item{alpha}{Numeric in (0,1). Significance level. Default: 0.05.}

\item{c}{Numeric > 0. Sub-exponential scale. The process must satisfy
|hat_delta_i| <= c/2. For score differences in \verb{[a-b, b-a]},
c = b - a (e.g. c = 2 for Brier score differences in \verb{[-1,1]}).
Default: 2.}

\item{v_opt}{Numeric > 0. Optimal intrinsic time. Default: 10.}

\item{boundary}{Character. "mixture" (default, GE mixture) or "stitching"
(polynomial stitched) or "hardcoded" (CR23 example formula,
only valid for alpha=0.05, c=1).}

\item{gammas}{Numeric vector or NULL. Predictable centering sequence.
If NULL, constructed as lagged running mean (default).}

\item{lcb_only}{Logical. If TRUE, return lower CS only: \verb{[lower, +Inf)}.
Requires finite lower bound on hat_delta_i; provide c.}

\item{ucb_only}{Logical. If TRUE, return upper CS only: \verb{(-Inf, upper]}.}
}
\value{
data.frame with columns t, estimate, lower, upper.
lower = -Inf if ucb_only = TRUE; upper = Inf if lcb_only = TRUE.
}
\description{
Constructs a variance-adaptive time-uniform CS using empirical intrinsic
time \eqn{\hat{V}_t = \sum_{i=1}^t (\hat{\delta}_i - \gamma_i)^2}.
Tighter than the Hoeffding CS when score differences have low variance.
}
\details{
The CS is:
\deqn{C_t^{EB} = \hat{\Delta}_t \pm u_{\alpha/2}^{GE}(\hat{V}_t;\, \rho, c) \;/\; t}
}
\examples{
scores1 <- c(-0.04, -0.09, -0.01, -0.16)
scores2 <- c(-0.09, -0.16, -0.04, -0.25)
cs_bernstein(scores1, scores2, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{crps_empirical}
\alias{crps_empirical}
\title{Negated CRPS for empirical predictive distributions}
\usage{
crps_empirical(ensemble, y)
}
\arguments{
\item{ensemble}{Matrix. T x n matrix of forecast draws. Each row
corresponds to one observation in y and comprises
n simulation draws from the predictive distribution.
For Historical Simulation: each row is the past
WINDOW returns.}

\item{y}{Numeric vector of length T. Realised observations.}
}
\value{
Numeric vector of length T of CRPS values in \verb{(-Inf, 0]}
(negated loss).
}
\description{
Wrapper over \code{scoringRules::crps_sample} using method = "edf"
(empirical distribution function, O(n log n) via quantile decomposition
of Laio & Tamea, 2007). Positively oriented: higher = better.
}
\details{
Requires \code{nrow(ensemble) == length(y)}. Passes
\code{dat = ensemble} directly to \code{crps_sample} which handles
vectorisation over rows natively. \code{show_messages} is suppressed
as the "edf" method requires no bandwidth selection messages.
}
\examples{
if (requireNamespace("scoringRules", quietly = TRUE)) {
  ensemble <- matrix(c(0.1, 0.2, 0.3, 1.0, 1.1, 1.2), nrow = 2, byrow = TRUE)
  crps_empirical(ensemble, y = c(0.25, 1.05))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/winkler.R
\name{winkler_cs}
\alias{winkler_cs}
\title{One-sided empirical Bernstein CS for Winkler scores (Proposition 4)}
\usage{
winkler_cs(
  p,
  q,
  y,
  alpha = 0.05,
  base_score = log_score,
  v_opt = 10,
  lower_bound = NULL
)
}
\arguments{
\item{p}{Numeric vector in (0,1). Forecasts from model 1.}

\item{q}{Numeric vector in (0,1). Forecasts from model 2.}

\item{y}{Numeric vector containing only 0 and 1. Binary outcomes.}

\item{alpha}{Numeric in (0,1). Significance level. Default: 0.05.}

\item{base_score}{Function. Underlying scoring rule. Default: log_score.}

\item{v_opt}{Numeric > 0. Optimal intrinsic time. Default: 10.}

\item{lower_bound}{Numeric or NULL. Analytical lower bound on w_i for
two-sided CS via Corollary 2. If NULL (default), returns
one-sided CS only. If supplied, must satisfy w_i >= lower_bound
for all i almost surely.}
}
\value{
data.frame with columns t, estimate, lower, upper.
lower = -Inf always (one-sided) unless lower_bound is supplied.
}
\description{
Applies the Winkler normalisation and constructs a one-sided upper
confidence sequence for the mean Winkler score W_t = (1/t)*sum w_i.
The CS takes the form \verb{(-Inf, U_t]}, valid uniformly over all t >= 1.
}
\details{
Scale convention: Winkler score bounded above by 1, so c/2 = 1, c = 2.
This is hardcoded — do not change c without re-deriving the bound.
}
\section{Interpretation}{

If \code{U_t < 0} for some \code{t}, this is time-uniform evidence that forecaster 1
(\code{p}) is worse than forecaster 2 (\code{q}) on average — i.e. a rejection is
evidence against \code{p}, not for it. More generally, \code{W_t > 0} suggests \code{p}
outperforms \code{q}; \code{W_t < 0} suggests \code{q} outperforms \code{p}.
}

\examples{
p <- c(0.7, 0.6, 0.8, 0.65)
q <- c(0.5, 0.7, 0.6, 0.55)
y <- c(1, 1, 0, 1)
winkler_cs(p, q, y, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lag_handling.R
\name{split_streams}
\alias{split_streams}
\title{Split a sequence into h interleaved lag streams}
\usage{
split_streams(xs, h)
}
\arguments{
\item{xs}{Numeric vector. Score differences hat_delta_t.}

\item{h}{Integer >= 1. Lag (number of steps ahead).}
}
\value{
List of length h. Each element is a numeric vector
containing the score differences for that stream.
}
\description{
For lag \code{h}, the k-th stream (\verb{k = 1, ..., h}) contains indices
\eqn{\{k,\, k+h,\, k+2h,\, \ldots\}}, following the CR23 convention.
}
\examples{
  split_streams(1:10, h = 3)
  # stream 1: indices 1, 4, 7, 10
  # stream 2: indices 2, 5, 8
  # stream 3: indices 3, 6, 9

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{brier_score}
\alias{brier_score}
\title{Brier score for binary and categorical forecasts}
\usage{
brier_score(p, y)
}
\arguments{
\item{p}{Numeric vector in \verb{[0, 1]} for binary forecasts, or a numeric
matrix whose rows are probability vectors for categorical forecasts.}

\item{y}{For binary vector input, numeric vector in \verb{\{0, 1\}}. For
categorical matrix input, integer vector in \verb{\{1, ..., K\}}, where
\code{K = ncol(p)}.}
}
\value{
Numeric vector of scores in \verb{[-1, 0]}. Higher is better.
}
\description{
Computes the positively oriented Brier/quadratic score. Vector probability
input is treated as binary; matrix probability input is treated as
categorical.
}
\details{
For binary forecasts, this computes
\deqn{S(p, y) = -(p-y)^2.}

For categorical forecasts, this computes
\deqn{S(\mathbf{p}, y) = -\frac{1}{2}\|\mathbf{p} - e_y\|_2^2,}
where \code{e_y} is the one-hot vector of the realised category.

With the convention that category 2 corresponds to the binary event
\code{y = 1}, the categorical formula recovers the binary formula exactly
when \code{K = 2}.
}
\section{Bounds}{

Score differences lie in \verb{[-1, 1]}, so use \code{c = 1} for Theorem 1 and
\code{c = 2} for Theorems 2 and 3.
}

\examples{
p <- c(0.2, 0.7, 0.9)
y <- c(0, 1, 1)
brier_score(p, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_forecasts.R
\name{compare_forecasts}
\alias{compare_forecasts}
\title{Compare Two Sequential Forecasters}
\usage{
compare_forecasts(
  p,
  q,
  y,
  scoring_rule = c("brier", "spherical", "log"),
  alpha = 0.05,
  cs_type = NULL,
  compute_cs = TRUE,
  compute_e = TRUE,
  v_opt = 10,
  boundary = "mixture",
  lcb_only = FALSE,
  ucb_only = FALSE,
  eps = 1e-15,
  clip_max = 1e+07
)
}
\arguments{
\item{p}{Forecasts from forecaster 1. For binary outcomes, a numeric vector
of probabilities for event \code{y = 1}. For categorical outcomes, a numeric
matrix whose rows are probability vectors.}

\item{q}{Forecasts from forecaster 2, in the same format as \code{p}.}

\item{y}{Outcomes. For binary vector forecasts, a numeric vector in \verb{\{0, 1\}}.
For categorical matrix forecasts, integer class labels in \verb{\{1, ..., K\}}.}

\item{scoring_rule}{Character. Scoring rule used to compare forecasts.
Currently supports \code{"brier"}, \code{"spherical"}, and \code{"log"}.}

\item{alpha}{Numeric in \verb{(0, 1)}. Significance level. Default is \code{0.05}.}

\item{cs_type}{Character or \code{NULL}. Confidence sequence type:
\code{"bernstein"}, \code{"hoeffding"}, \code{"asymptotic"}, or \code{"none"}.
If \code{NULL}, the wrapper uses \code{"bernstein"} for bounded scoring rules
(\code{"brier"} and \code{"spherical"}) and \code{"asymptotic"} for \code{"log"}.}

\item{compute_cs}{Logical. If \code{TRUE}, compute a confidence sequence.
Default is \code{TRUE}.}

\item{compute_e}{Logical. If \code{TRUE}, compute two one-sided e-processes.
Default is \code{TRUE}. This is only allowed for bounded score differences
under the current wrapper, namely \code{"brier"} and \code{"spherical"}.}

\item{v_opt}{Numeric > 0. Intrinsic time at which the mixture boundary or
e-process is tuned to be tightest. Default is \code{10}.}

\item{boundary}{Character. Boundary type passed to \code{\link[=cs_hoeffding]{cs_hoeffding()}} or
\code{\link[=cs_bernstein]{cs_bernstein()}}. Default is \code{"mixture"}.}

\item{lcb_only}{Logical. If \code{TRUE}, compute a lower one-sided empirical
Bernstein CS. Only used when \code{cs_type = "bernstein"}.}

\item{ucb_only}{Logical. If \code{TRUE}, compute an upper one-sided empirical
Bernstein CS. Only used when \code{cs_type = "bernstein"}.}

\item{eps}{Numeric. Probability floor passed to \code{\link[=log_score]{log_score()}} when
\code{scoring_rule = "log"}. Default is \code{1e-15}.}

\item{clip_max}{Numeric. Maximum e-process value before clipping. Passed to
\code{\link[=eprocess]{eprocess()}}. Default is \code{1e7}.}
}
\value{
A \code{data.frame} with one row per time point and columns:
\describe{
\item{\code{t}}{Time index.}
\item{\code{score_p}}{Pointwise score of forecaster \code{p}.}
\item{\code{score_q}}{Pointwise score of forecaster \code{q}.}
\item{\code{delta}}{Pointwise score difference, \code{score_p - score_q}.}
\item{\code{estimate}}{Running mean score difference. Positive values favour
forecaster \code{p}; negative values favour forecaster \code{q}.}
\item{\code{lower}, \code{upper}}{Confidence sequence bounds. These are \code{NA} if
\code{compute_cs = FALSE} or \code{cs_type = "none"}.}
\item{\code{e_pq}, \code{e_qp}}{One-sided e-processes. \code{e_pq} tests whether
forecaster \code{p} outperforms \code{q}; \code{e_qp} tests the reverse direction.
These are \code{NA} if \code{compute_e = FALSE}.}
}
}
\description{
Computes pointwise scores for two probabilistic forecasters and compares
them sequentially using confidence sequences and, when valid finite-sample
bounds are available, e-processes.
}
\details{
This is a convenience wrapper around \code{\link[=brier_score]{brier_score()}}, \code{\link[=spherical_score]{spherical_score()}},
\code{\link[=log_score]{log_score()}}, \code{\link[=cs_hoeffding]{cs_hoeffding()}}, \code{\link[=cs_bernstein]{cs_bernstein()}}, \code{\link[=cs_asymptotic]{cs_asymptotic()}}, and
\code{\link[=eprocess]{eprocess()}}. It is designed for the common workflow where the user has two
forecast streams \code{p} and \code{q}, an outcome stream \code{y}, and wants a single tidy
output object.

All scoring rules in \code{seqcomp} are positively oriented: higher scores are
better. Therefore
\deqn{\hat{\delta}_t = S(p_t, y_t) - S(q_t, y_t)}
is positive when forecaster \code{p} performs better than forecaster \code{q} at time
\code{t}.

For \code{"brier"} and \code{"spherical"}, score differences are bounded in \verb{[-1, 1]}.
The wrapper therefore uses \code{c = 1} for Hoeffding-style confidence sequences
and \code{c = 2} for empirical Bernstein confidence sequences and e-processes.

For \code{"log"}, score differences are unbounded. The wrapper therefore defaults
to \code{\link[=cs_asymptotic]{cs_asymptotic()}} and refuses to compute finite-sample e-processes. For
binary log-score comparisons where the Winkler construction is appropriate,
use \code{\link[=winkler_compare]{winkler_compare()}} instead.
}
\section{Interpretation}{

The confidence sequence estimates the running average score advantage of
\code{p} over \code{q}. If the whole interval lies above zero, the data favour \code{p};
if the whole interval lies below zero, the data favour \code{q}.

The e-processes are evidence processes for one-sided null hypotheses. At
level \code{alpha}, the two-sided rejection threshold used by \code{\link[=eprocess]{eprocess()}} is
\code{2 / alpha}.
}

\examples{
set.seed(1)
y <- rbinom(200, 1, 0.5)
p <- rep(0.5, 200)
q <- runif(200)

out <- compare_forecasts(p, q, y, scoring_rule = "brier")
tail(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{winkler_score}
\alias{winkler_score}
\title{Winkler-normalized binary score}
\usage{
winkler_score(p, q, y, base_score = log_score, eps = 1e-08)
}
\arguments{
\item{p}{Numeric vector in (0,1). Forecasts from model 1.}

\item{q}{Numeric vector in (0,1). Forecasts from model 2.}

\item{y}{Numeric vector containing only 0 and 1. Binary outcomes.}

\item{base_score}{Function. The underlying scoring rule S(p, y).
Must accept two arguments: forecast probability and outcome.
Default: log_score (with eps clipping).}

\item{eps}{Numeric. Zero-protection for the normaliser denominator.
Default: 1e-8 (matches Python comparecast convention).}
}
\value{
Numeric vector. Winkler scores in \verb{(-Inf, 1]}.
Upper bound of 1 is tight: w = 1 when y = 1(p > q).
}
\description{
Normalises the score difference S(p,y) - S(q,y) by the maximum possible
score difference given the forecaster ordering, mapping the result to
\verb{(-Inf, 1]} (Proposition 4, Choe & Ramdas 2023). Used to apply Theorems 2 & 3
to unbounded scoring rules on binary outcomes.
}
\details{
\deqn{w(p, q, y) = \frac{S(p,y) - S(q,y)}{S(p, \mathbb{1}(p>q)) - S(q, \mathbb{1}(p>q))}}
with the convention 0/0 := 0.

The lower bound is problem-dependent (depends on how extreme p and q can
be). For a two-sided CS via Corollary 2, the user must establish a finite
lower bound analytically. If no finite lower bound can be guaranteed, use
the one-sided (upper) CS only, as in the CR23 MLB experiments.
}
\section{When to use}{

Strictly limited to binary outcomes \code{y} in \verb{\{0, 1\}} and probability
forecasts \code{p}, \code{q} in \verb{(0, 1)}. Not applicable to QLIKE or other
continuous-outcome scoring rules. See CR23 Section G for discussion.

For use in Theorems 2 & 3: upper bound = 1 implies c/2 = 1, so use \code{c = 2}
in all GE boundary and e-process calls.
}

\examples{
p <- c(0.7, 0.6, 0.8, 0.65)
q <- c(0.5, 0.7, 0.6, 0.55)
y <- c(1, 1, 0, 1)
winkler_score(p, q, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{ps_boundary}
\alias{ps_boundary}
\title{Polynomial stitched (PS) boundary}
\usage{
ps_boundary(v, alpha, v_opt = 10, c = 1, s = 1.4, eta = 2)
}
\arguments{
\item{v}{Numeric vector >= 0. Intrinsic time values.}

\item{alpha}{Numeric in (0,1). ONE-SIDED significance level.}

\item{v_opt}{Numeric > 0. Optimal intrinsic time (= m in H21).
Default: 10.}

\item{c}{Numeric > 0. Sub-exponential scale.}

\item{s}{Numeric > 1. Stitching parameter. Default: 1.4.}

\item{eta}{Numeric > 1. Geometric spacing. Default: 2.}
}
\value{
Numeric vector of boundary values (cumulative-sum scale).
}
\description{
Alternative boundary for both Theorem 1 and Theorem 2 constructions,
included for completeness and for cross-checking CR23 Table results.
}
\details{
This is \strong{not} the recommended primary boundary: the CM/GE mixture
boundaries (\code{cm_boundary()}, \code{ge_boundary()}) are tighter in CR23 and are
used by default throughout \code{seqcomp}. Use \code{ps_boundary()} only when you
specifically need the polynomial-stitched construction.
}
\examples{
ps_boundary(v = 1:5, alpha = 0.025, v_opt = 10, c = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{tick_loss}
\alias{tick_loss}
\title{Negated tick loss for quantile forecasts}
\usage{
tick_loss(q, y, alpha)
}
\arguments{
\item{q}{Numeric vector. Quantile forecasts at level alpha.}

\item{y}{Numeric vector. Realised outcomes.}

\item{alpha}{Numeric in (0,1). Quantile level.}
}
\value{
Numeric vector of negated tick loss scores. Higher = better.
}
\description{
Computes the positively oriented (negated) tick/quantile loss
(Koenker & Bassett, 1978).
}
\details{
The standard tick loss is
\deqn{\rho_\alpha(u) = u \left(\alpha - \mathbb{1}(u < 0)\right),}
where \eqn{u = y - q_\alpha} is the forecast error. This is loss-oriented
(lower = better), so the function negates it:
\deqn{S_T(q, y; \alpha) = -(y - q)\left(\alpha - \mathbb{1}(y < q)\right).}

Tick loss is unbounded on general real-valued outcomes. Bounds derived from
an empirical data range are ex-post and do not provide theorem-valid
constants for finite-sample Hoeffding/Bernstein confidence sequences or
e-processes.

Sign convention: the negation means \code{hat_delta_t > 0} when forecaster \code{p}
has smaller tick loss, hence a better quantile forecast, than forecaster \code{q}.
}
\examples{
q <- c(1.0, 1.5, 2.0)
y <- c(1.2, 1.4, 2.3)
tick_loss(q, y, alpha = 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/etests.R
\name{eprocess_rejections}
\alias{eprocess_rejections}
\title{Determine rejection times for an e-process output}
\usage{
eprocess_rejections(ep, alpha = 0.05)
}
\arguments{
\item{ep}{data.frame. Output of eprocess().}

\item{alpha}{Numeric. Significance level. Threshold is 2/alpha.}
}
\value{
Named list with elements:
\itemize{
\item \code{threshold} — rejection threshold (\code{2 / alpha}).
\item \code{tau_pq} — first \code{t} where \code{e_pq >= threshold} (\code{NA} if never crossed).
\item \code{tau_qp} — first \code{t} where \code{e_qp >= threshold} (\code{NA} if never crossed).
\item \code{reject_pq} — logical: was \eqn{H_0^w(p,q)} ever rejected?
\item \code{reject_qp} — logical: was \eqn{H_0^w(q,p)} ever rejected?
}
}
\description{
Determine rejection times for an e-process output
}
\examples{
scores1 <- c(-0.04, -0.09, -0.01, -0.16)
scores2 <- c(-0.09, -0.16, -0.04, -0.25)
ep <- eprocess(scores1, scores2, alpha = 0.05)
eprocess_rejections(ep, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{score_diff_scales}
\alias{score_diff_scales}
\title{Score difference bounds -> sub-Gaussian / sub-exponential scale}
\usage{
score_diff_scales(lo, hi)
}
\arguments{
\item{lo}{Numeric. Lower bound of score difference (usually a - b).}

\item{hi}{Numeric. Upper bound of score difference (usually b - a).}
}
\value{
Named list with elements c_thm1 and c_thm23.
}
\description{
Given score-difference bounds \verb{[lo, hi]}, computes the two scale constants
used elsewhere in \code{seqcomp}:
\itemize{
\item \code{c_thm1}  = \code{(hi - lo) / 2} — sub-Gaussian scale for Theorem 1, where
\verb{|delta_i| <= c_thm1}.
\item \code{c_thm23} = \code{hi - lo} — sub-exponential scale for Theorems 2 & 3, where
\verb{|delta_i| <= c_thm23 / 2}.
}
}
\details{
Both conventions bound the same quantity: after centering, \code{delta_i} lies
in \verb{[-(hi-lo)/2, (hi-lo)/2]}, so \verb{max|delta_i| = (hi-lo)/2}, which equals
both \code{c_thm1} and \code{c_thm23 / 2}.
}
\examples{
score_diff_scales(lo = -1, hi = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{qlike_score}
\alias{qlike_score}
\title{Negated QLIKE score for variance forecasts}
\usage{
qlike_score(sigma2_hat, sigma2)
}
\arguments{
\item{sigma2_hat}{Numeric vector. Forecast variance (strictly positive).}

\item{sigma2}{Numeric vector. Realised variance (strictly positive).}
}
\value{
Numeric vector of negated QLIKE scores. Higher is better.
Maximum value is 0, achieved at a perfect forecast \code{sigma2_hat = sigma2}.
Unbounded below.
}
\description{
Computes the positively oriented (negated) QLIKE quasi-likelihood loss for
variance forecasts.
}
\details{
Standard QLIKE loss is
\deqn{L_{QL}(\hat\sigma^2, \sigma^2) = \frac{\sigma^2}{\hat\sigma^2} -
  \log\frac{\sigma^2}{\hat\sigma^2} - 1.}
This is loss-oriented (lower = better, minimum 0 at a perfect forecast), so
the function negates it: \eqn{S_{QL} = -L_{QL}}.

Literature note: some sources define QLIKE as
\code{log(sigma2_hat) + sigma2 / sigma2_hat}, which differs by constants from
the form above. Here the loss is normalised to have minimum 0 and is then
negated for positive orientation.
}
\section{Unbounded below}{

QLIKE is unbounded below. It should not be used directly with the
finite-sample bounded-difference confidence sequences or e-processes.
Use \code{cs_asymptotic()} for QLIKE-based confidence sequences, or use
\code{eprocess_predictable()} only when genuine ex ante predictable bounds are
available. QLIKE is not compatible with the Winkler construction because
Winkler scores are restricted to binary outcomes and probability forecasts.
}

\examples{
sigma2_hat <- c(1.0, 1.5, 2.0)
sigma2 <- c(1.1, 1.4, 2.2)
qlike_score(sigma2_hat, sigma2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confidence_sequences.R
\name{cs_asymptotic}
\alias{cs_asymptotic}
\title{Asymptotic confidence sequence (Appendix C, Eq. 55, Choe & Ramdas 2023)}
\usage{
cs_asymptotic(scores1, scores2, alpha = 0.05, t_star = NULL)
}
\arguments{
\item{scores1}{Numeric vector. Scores for forecaster 1.}

\item{scores2}{Numeric vector. Scores for forecaster 2.}

\item{alpha}{Numeric in (0,1). Significance level. Default: 0.05.}

\item{t_star}{Numeric > 0. Sample size at which CS is tightest.
Default: length(scores1) (tightest at end of sample).}
}
\value{
data.frame with columns t, estimate, lower, upper.
}
\description{
\strong{Asymptotic, not finite-sample}: coverage \verb{>= 1 - alpha} holds only as
\code{t -> infinity}. Valid without requiring bounded score differences —
requires only that \code{hat_delta_t} has finite variance — so it's appropriate
for tick loss and other scoring rules where hard bounds depend on
unbounded realised values. Suitable for large evaluation windows.
}
\details{
\deqn{C_t^A = \hat\Delta_t \pm \sqrt{
  \frac{2(t \sigma^2_t \rho^2 + 1)}{t^2 \rho^2}
  \log\frac{\sqrt{t \sigma^2_t \rho^2 + 1}}{\alpha}}}
where \eqn{\sigma^2_t = \frac{1}{t}\sum_{i=1}^t (\hat\delta_i - \hat\Delta_{i-1})^2}
and \code{rho} is tuned to be tightest at \code{t_star}:
\deqn{\rho(t_{star}) = \sqrt{\frac{2\log(1/\alpha) + \log(1 + 2\log(1/\alpha))}{t_{star}}}}

The running variance estimator uses the predictable mean \verb{hat_Delta_\{t-1\}}
(not the current mean \code{hat_Delta_t}) to maintain predictability, with
\code{hat_Delta_0 := 0}.
}
\examples{
scores1 <- c(-0.4, -0.2, -0.3, -0.1, -0.2)
scores2 <- c(-0.5, -0.3, -0.4, -0.2, -0.3)
cs_asymptotic(scores1, scores2, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/winkler.R
\name{winkler_etest}
\alias{winkler_etest}
\title{E-process for Winkler scores (Proposition 4 + Theorem 3)}
\usage{
winkler_etest(
  p,
  q,
  y,
  alpha = 0.05,
  base_score = log_score,
  v_opt = 10,
  clip_max = 1e+07
)
}
\arguments{
\item{p}{Numeric vector in (0,1). Forecasts from model 1.}

\item{q}{Numeric vector in (0,1). Forecasts from model 2.}

\item{y}{Numeric vector containing only 0 and 1. Binary outcomes.}

\item{alpha}{Numeric in (0,1). Significance level. Default: 0.05.}

\item{base_score}{Function. Underlying scoring rule. Default: log_score.}

\item{v_opt}{Numeric > 0. Default: 10.}

\item{clip_max}{Numeric. Maximum e-process value before clipping.
Default: 1e7.}
}
\value{
data.frame with columns t, e, log_e.
}
\description{
Tests whether the mean Winkler score W_t >= 0 for all t.
A rejection provides time-uniform evidence that forecaster 1 (p)
is worse than forecaster 2 (q) under the base scoring rule.
}
\section{Rejection rule}{

Reject at level \code{alpha} when \code{e >= 1 / alpha}; this provides time-uniform
evidence that \code{p} is worse than \code{q}.
}

\examples{
p <- c(0.7, 0.6, 0.8, 0.65)
q <- c(0.5, 0.7, 0.6, 0.55)
y <- c(1, 1, 0, 1)
winkler_etest(p, q, y, alpha = 0.05)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{log_score}
\alias{log_score}
\title{Logarithmic score for binary and categorical forecasts}
\usage{
log_score(p, y, eps = 1e-15)
}
\arguments{
\item{p}{Numeric vector in \verb{[0, 1]} for binary forecasts, or a numeric
matrix whose rows are probability vectors for categorical forecasts.}

\item{y}{For binary vector input, numeric vector in \verb{\{0, 1\}}. For
categorical matrix input, integer vector in \verb{\{1, ..., K\}}, where
\code{K = ncol(p)}.}

\item{eps}{Numeric. Probability floor used before taking logarithms.
Default is \code{1e-15}. Set to \code{0} to disable clipping.}
}
\value{
Numeric vector of scores in \verb{(-Inf, 0]}. Higher is better.
}
\description{
Computes the positively oriented logarithmic score. Vector probability input
is treated as binary; matrix probability input is treated as categorical.
}
\details{
For binary forecasts, this computes
\deqn{S(p, y) = y\log(p) + (1-y)\log(1-p).}

For categorical forecasts, this computes
\deqn{S(\mathbf{p}, y) = \log(p_y),}
where \code{p_y} is the forecast probability assigned to the realised category.
}
\section{Use with seqcomp}{

The logarithmic score is unbounded below. It should not be used directly
with the finite-sample bounded-difference confidence sequences or
e-processes. For binary outcomes, use \code{winkler_score()} and \code{winkler_cs()}
when the Winkler construction is appropriate. For unbounded score
differences, use \code{cs_asymptotic()} or supply genuine predictable bounds
to \code{eprocess_predictable()}.
}

\examples{
p <- c(0.2, 0.7, 0.9)
y <- c(0, 1, 1)
log_score(p, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scores.R
\name{spherical_score}
\alias{spherical_score}
\title{Spherical score for binary and categorical forecasts}
\usage{
spherical_score(p, y)
}
\arguments{
\item{p}{Numeric vector in \verb{[0, 1]} for binary forecasts, or a numeric
matrix whose rows are probability vectors for categorical forecasts.}

\item{y}{For binary vector input, numeric vector in \verb{\{0, 1\}}. For
categorical matrix input, integer vector in \verb{\{1, ..., K\}}, where
\code{K = ncol(p)}.}
}
\value{
Numeric vector of scores in \verb{[0, 1]}. Higher is better.
}
\description{
Computes the positively oriented spherical score. Vector probability input
is treated as binary; matrix probability input is treated as categorical.
}
\details{
For binary forecasts, this computes \deqn{S(p, y) =
\frac{py + (1-p)(1-y)}{\sqrt{p^2 + (1-p)^2}}.}
For categorical forecasts, this computes \deqn{S(\mathbf{p}, y) =
\frac{p_y}{\|\mathbf{p}\|_2},}
where \code{p_y} is the forecast probability assigned to the realised category.

Score differences lie in \verb{[-1, 1]}, so use \code{c = 1} for Theorem 1 and
\code{c = 2} for Theorems 2 and 3.
}
\examples{
p <- c(0.2, 0.7, 0.9)
y <- c(0, 1, 1)
spherical_score(p, y)

}
